/*
 * RageIRCd, an advanced Internet Relay Chat daemon (IRCd)
 * contrib/aup_notice.c: send AUP notice on connect
 *
 * Copyright (C) 2002 the past and present IRCd developers, and others.
 * Refer to doc/authors/ for full credits and copyright details.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA.
 *
 * $Id: aup_notice.c,v 1.2 2004/04/03 19:54:44 rageircd Exp $
 */

/* This module creates a new conf variable within the network {} conf
 * block called acceptable_use_policy. This should be set to the
 * location of your network Acceptable Use Policy (rules). When set,
 * a notice will be sent to connecting clients, regarding this AUP.
 *
 * network {
 *	acceptable_use_policy = "http://www.your-irc-network.com/aup.html";
 * };
 */

#include "config.h"
#include "struct.h"
#include "common.h"
#include "sys.h"
#include "numeric.h"
#include "msg.h"
#include "channel.h"
#include "h.h"
#include "memory.h"
#include "modules.h"
#include "hook.h"
#include "conf2.h"
#include "fd.h"
#include <time.h>
#include <sys/stat.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>

static int parse_aup(HookData *);
static int complete_aup(HookData *);
static int rehash_aup(HookData *);
static int show_aup_notice(HookData *);

static void free_aup();

static char *aup_url = NULL;
static char aup_notice[256];
static int aup_length = 0;

Module MOD_HEADER(aup_notice) = {
	"modules/aup_notice.so",
	"On-connect AUP notice",
	5, "$Revision: 1.2 $"
};

int MOD_LOAD(aup_notice)()
{
	register_hook_event(&MOD_HEADER(aup_notice), h_conf_parse, parse_aup);
	register_hook_event(&MOD_HEADER(aup_notice), h_conf_complete, complete_aup);
	register_hook_event(&MOD_HEADER(aup_notice), h_conf_rehash, rehash_aup);
	register_hook_event(&MOD_HEADER(aup_notice), h_pre_connauth, show_aup_notice);
	return MOD_SUCCESS;
}

int MOD_UNLOAD(aup_notice)()
{
	free_aup();
	return MOD_SUCCESS;
}

static void free_aup()
{
	if (!BadPtr(aup_url)) {
		cMyFree(aup_url);
	}
	aup_notice[0] = '\0';
	aup_length = 0;
}

static int parse_aup(HookData *hdata)
{
	ConfigUnknown *unknown = (ConfigUnknown *)hdata->v;
	ConfigEntry *entry = unknown->entry;

	if (unknown->type == CONF_NETWORK) {
		if (!mycmp(entry->varname, "acceptable_use_policy")) {
			cDupString(aup_url, entry->vardata);
			return CONF_SUCCESS;
		}
	}
	return CONF_NOMATCH;
}

static int complete_aup(HookData *hdata)
{
	if (!BadPtr(aup_url)) {
		aup_length = ircsnprintf(aup_notice, 256, "NOTICE AUTH :*** Connecting to this "
			"server constitutes agreement with our Acceptable Use Policy, available "
			"at %s. If you do not agree with this policy, you must disconnect "
			"immediately.\r\n", aup_url);
	}
	return 0;
}

static int rehash_aup(HookData *hdata)
{
	free_aup();
	return CONF_SUCCESS;
}

static int show_aup_notice(HookData *hdata)
{
	aClient *cptr = (aClient *)hdata->cptr;
	ASSERT(cptr != NULL);

	if (aup_length) {
		ircsend(cptr, aup_notice, aup_length);
	}
	return 0;
}
