/*
 * RageIRCd, an advanced Internet Relay Chat daemon (IRCd)
 * modules/m_accept.c: ACCEPT command module
 *
 * Copyright (C) 2002 the past and present IRCd developers, and others.
 * Refer to doc/authors/ for full credits and copyright details.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA.
 *
 * $Id: m_accept.c,v 1.32 2004/05/13 22:26:56 rageircd Exp $
 */

#include "config.h"
#include "struct.h"
#include "common.h"
#include "sys.h"
#include "numeric.h"
#include "msg.h"
#include "channel.h"
#include "h.h"
#include "memory.h"
#include "hook.h"
#include "dlink.h"
#include "modules.h"
#include "xmode.h"
#include <time.h>
#include <sys/stat.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>

static int accept_msg(HookData *);
static int del_accepts(HookData *);

Module MOD_HEADER(m_accept) = {
	"modules/commands/m_accept.so",
	"/ACCEPT command",
	6, "$Revision: 1.32 $"
};

int MOD_LOAD(m_accept)()
{
	Hook *hook[3];

	if ((hook[0] = hook_find_from_module("h_privmsg", "modules/commands/m_message.so")) == NULL) {
		ircdlog(LOG_ERROR, "m_accept.so: couldn't locate hook h_privmsg");
		return MOD_FAILURE;
	}
	if ((hook[1] = hook_find_from_module("h_nick_change_local", "modules/commands/m_nick.so")) == NULL) {
		ircdlog(LOG_ERROR, "m_accept.so: couldn't locate hook h_nick_change_local");
		return MOD_FAILURE;
	}
	if ((hook[2] = hook_find_from_module("h_nick_change_remote", "modules/commands/m_nick.so")) == NULL) {
		ircdlog(LOG_ERROR, "m_accept.so: couldn't locate hook h_nick_change_remote");
		return MOD_FAILURE;
	}

	register_command(&MOD_HEADER(m_accept), &CMD_ACCEPT, m_accept);

	register_hook_event(&MOD_HEADER(m_accept), hook[0], accept_msg);
	register_hook_event(&MOD_HEADER(m_accept), h_exit_user_local, del_accepts);
	register_hook_event(&MOD_HEADER(m_accept), hook[1], del_accepts);
	register_hook_event(&MOD_HEADER(m_accept), hook[2], del_accepts);

	return MOD_SUCCESS;
}

int MOD_UNLOAD(m_accept)()
{
	return MOD_SUCCESS;
}

static int accept_msg(HookData *hdata)
{
	aClient *from = hdata->sptr;
	aClient *to = hdata->acptr;

	if (to == from || !IsPerson(from) || IsULine(from) || !MyClient(to) || !HasMode(to, UMODE_PRIVACY)) {
		return 0;
	}
	if (!HasMode(to, UMODE_PRIVACY) || (GeneralConfig.no_oper_accept && HasMode(from, UMODE_OPER))) {
		return 0;
	}

	if (dlink_find(&to->user->accept_list, (void *)from) == NULL) {
		send_me_notice(from, ":*** %s is in private mode (+p), and is not accepting "
			"privmsgs from you.", to->name);
		if ((timeofday - to->user->last_accept_notice) >= FloodConfig.accept_notice_time) {
			send_me_noticeNA(from, ":*** He or she has been informed of your privmsg attempt.");
			send_me_notice(to, ":*** %s has attempted to privmsg you, but it was blocked "
				"because you are in privacy mode (+p).", from->name);
			send_me_notice(to, ":*** Type /ACCEPT ? for info on how to accept messages "
				"from %s.", from->name);
			to->user->last_accept_notice = timeofday;
		}
		return 1;
	}
	return 0;
}

static void show_accept_list(aClient *cptr)
{
	dlink_node *node;
	aClient *acptr;
	char buf[BUFSIZE], *s;
	int len = 0, count = 0;
	int plen = strlen(cptr->name) + 10;

	DLINK_FOREACH_DATA(cptr->user->accept_list.head, node, acptr, aClient) {
		if ((plen + len + strlen(acptr->name) + 1) > BUFSIZE || count > 10) {
			send_me_numeric(cptr, RPL_ACCEPTLIST, buf);
			buf[0] = '\0';
			len = count = 0;
		}
		if (len > 0) {
			buf[len++] = ' ';
		}
		for (s = acptr->name; *s != '\0'; s++) {
			buf[len++] = *s;
		}
		buf[len] = '\0';
		count++;
	}
	if (count > 0) {
		send_me_numeric(cptr, RPL_ACCEPTLIST, buf);
	}
	send_me_numericNA(cptr, RPL_ENDOFACCEPT);
	return;
}

static void show_accept_help(aClient *cptr)
{
	char **p;
	static char *helptext[] = {
		"/ACCEPT [-]nick[,[-]nick2[,[-]nick3[...]]]",
		"/ACCEPT <*|?>",
		" ",
		"If you are in privacy mode (usermode +p), you can still receive private",
		"messages by accepting users. This is done via the ACCEPT command (syntax",
		"above).",
		" ",
		"For example; if you are in privacy mode (+p), but want to accept any",
		"privmsgs from client Bob, you would issue the following command.",
		" ",
		"/ACCEPT Bob",
		" ",
		"However, you then decide you want to block privmsgs from Bob again. You",
		"simply remove him from your accept list. Syntax is similar to that of",
		"adding clients, but you prefix the nickname with a - (subtract) char.",
		" ",
		"/ACCEPT -Bob",
		" ",
		"You can add and remove more than one client to and from your accept list",
		"at once. Just separate your client list with , (commas), as per the",
		"following example.",
		" ",
		"/ACCEPT Bob,-Joe,-Lisa,Homer",
		" ",
		"NB. If a client on your accept list quits, is lost in a netsplit, or",
		"    changes his or her nickname, all references within your accept list",
		"    to that client will be removed. This functionality prevents users",
		"    being tracked.",
		" ",
		"To view all clients on your accept list, issue an ACCEPT command with",
		"an * (asterisk) as parameter. To view this help, issue an ACCEPT command",
		"with a ? (question mark). You cannot mix *, ? and/or nicknames together.",
		NULL
	};
	for (p = helptext; *p != '\0'; p++) {
		send_me_numeric(cptr, RPL_ACCEPTLIST, *p);
	}
	send_me_numericNA(cptr, RPL_ENDOFACCEPT);
	return;
}

static void add_accept(aClient *sptr, aClient *acptr)
{
	dlink_add(&sptr->user->accept_list, acptr);
	dlink_add(&acptr->user->on_accept_list, sptr);
	return;
}

static void del_accept(aClient *sptr, aClient *acptr)
{
	dlink_node *anode, *anext = NULL;
	dlink_node *node, *next = NULL;
	aClient *cptr;

	DLINK_FOREACH_SAFE_DATA(acptr->user->accept_list.head, anode, anext, cptr, aClient) {
		if (sptr == cptr) {
			dlink_del(&acptr->user->accept_list, NULL, anode);
			DLINK_FOREACH_SAFE_DATA(sptr->user->on_accept_list.head, node, next, cptr,
			  aClient) {
				if (acptr == cptr) {
					dlink_del(&sptr->user->on_accept_list, NULL, node);
				}
			}
		}
	}
	return;
}

static int del_accepts(HookData *hdata)
{
	aClient *sptr = hdata->sptr;
	aClient *acptr = NULL;
	dlink_node *node, *next = NULL;

	DLINK_FOREACH_SAFE_DATA(sptr->user->accept_list.head, node, next, acptr, aClient) {
		if (acptr != NULL) {
			del_accept(acptr, sptr);
		}
	}
	DLINK_FOREACH_SAFE_DATA(sptr->user->on_accept_list.head, node, next, acptr, aClient) {
		if (acptr != NULL) {
			del_accept(sptr, acptr);
		}
	}
	return 0;
}

/*
 * m_accept
 *	parv[0] = sender prefix
 *	parv[1] = parameter
 */
int m_accept(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
	char *s, *p;
	aClient *acptr;
	int del, count = 0;

	if (!MyClient(sptr)) {
		return 0;
	}
	if (parc < 2 || BadPtr(parv[1])) {
		send_me_numeric(sptr, ERR_NEEDMOREPARAMS, "ACCEPT");
		return 0;
	}
	if (*parv[1] == '*') {
		show_accept_list(sptr);
		return 0;
	}
	else if (*parv[1] == '?') {
		show_accept_help(sptr);
		return 0;
	}

	count = dlink_length(&sptr->user->accept_list);
	for (s = strtoken(&p, parv[1], ","); s != NULL; s = strtoken(&p, NULL, ",")) {
		while (*s == ' ') {
			s++;
		}
		if (*s == '\0') {
			continue;
		}
		else if (*s == '-') {
			del = 1;
			s++;
		}
		else {
			del = 0;
			if (*s == '+') {
				s++;
			}
		}

		if ((acptr = find_person(s, NULL)) == NULL) {
			send_me_numeric(sptr, ERR_NOSUCHNICK, s);
			continue;
		}
		else if (acptr == sptr) {
			continue;
		}

		if (del) {
			if (dlink_find(&sptr->user->accept_list, acptr) == NULL) {
				send_me_numeric(sptr, ERR_ACCEPTNOT, acptr->name);
				continue;
			}
			del_accept(acptr, sptr);
			count--;
		}
		else {
			if (dlink_find(&sptr->user->accept_list, acptr) != NULL) {
				send_me_numeric(sptr, ERR_ACCEPTEXISTS, acptr->name);
				continue;
			}
			if (count >= MAXACCEPT) {
				send_me_numeric(sptr, ERR_ACCEPTFULL, acptr->name, MAXACCEPT);
				continue;
			}
			add_accept(sptr, acptr);
			count++;
		}
	}
	return 0;
}
