/*
 * RageIRCd, an advanced Internet Relay Chat daemon (IRCd)
 * modules/m_dccallow.c: DCCALLOW command module
 *
 * Copyright (C) 2002 the past and present IRCd developers, and others.
 * Refer to doc/authors/ for full credits and copyright details.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA.
 *
 * $Id: m_dccallow.c,v 1.36 2004/05/13 16:12:53 rageircd Exp $
 */

#include "config.h"
#include "struct.h"
#include "common.h"
#include "sys.h"
#include "numeric.h"
#include "msg.h"
#include "channel.h"
#include "h.h"
#include "memory.h"
#include "hook.h"
#include "dlink.h"
#include "modules.h"
#include "xmode.h"
#include <time.h>
#include <sys/stat.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>

static int allow_dcc(HookData *);
static int del_dccallows(HookData *);

Module MOD_HEADER(m_dccallow) = {
	"modules/commands/m_dccallow.so",
	"/DCCALLOW command",
	6, "$Revision: 1.36 $"
};

int MOD_LOAD(m_dccallow)()
{
	Hook *hook[3];

	if ((hook[0] = hook_find_from_module("h_privmsg", "modules/commands/m_message.so")) == NULL) {
		ircdlog(LOG_ERROR, "m_dccallow.so: couldn't find hook h_privmsg");
		return MOD_FAILURE;
	}
	if ((hook[1] = hook_find_from_module("h_nick_change_local", "modules/commands/m_nick.so")) == NULL) {
		ircdlog(LOG_ERROR, "m_dccallow.so: couldn't find hook h_nick_change_local");
		return MOD_FAILURE;
	}
	if ((hook[2] = hook_find_from_module("h_nick_change_remote", "modules/commands/m_nick.so")) == NULL) {
		ircdlog(LOG_ERROR, "m_dccallow.so: couldn't find hook h_nick_change_remote");
		return MOD_FAILURE;
	}

	register_command(&MOD_HEADER(m_dccallow), &CMD_DCCALLOW, m_dccallow);

	register_hook_event(&MOD_HEADER(m_dccallow), hook[0], allow_dcc);
	register_hook_event(&MOD_HEADER(m_dccallow), h_exit_user_local, del_dccallows);
	register_hook_event(&MOD_HEADER(m_dccallow), hook[1], del_dccallows);
	register_hook_event(&MOD_HEADER(m_dccallow), hook[2], del_dccallows);

	return MOD_SUCCESS;
}

int MOD_UNLOAD(m_dccallow)()
{
	return MOD_SUCCESS;
}

static int allow_dcc(HookData *hdata)
{
	aClient *from = hdata->sptr;
	aClient *to = hdata->acptr;
	char filebuf[BUFSIZE], *s;
	int len = 0;
	ConfigItem_ban *ban;

	if (to == from || !IsPerson(from) || IsULine(from) || HasMode(from, UMODE_OPER) ||
	  !MyClient(to)) {
		return 0;
	}

	strncpyzt(filebuf, hdata->c, BUFSIZE);
	s = filebuf + 8;
	while (*s == ' ') {
		s++;
	}
	if (*s == '\0') {
		return 0;
	}
	while (*(s + len) != ' ') {
		len++;
		if (*(s + len) == '\0') {
			break;
		}
	}
	s[len] = '\0';

	if ((ban = find_ban(BAN_FILE, s)) == NULL) {
		return 0;
	}
	if (dlink_find(&to->user->dccallow_list, from) == NULL) {
		send_me_notice(from, ":*** The transfer (DCC SEND) of file %s is not permitted "
			"on this server (%s).", s, BadPtr(ban->reason) ? "no reason" : ban->reason);
		send_me_notice(to, ":*** %s has attempted to transfer (DCC SEND) file %s to you, "
			"but it was blocked because the filetype (%s) is banned (%s).", to->name,
			s, ban->mask, BadPtr(ban->reason) ? "no reason" : ban->reason);
		if (!SeenDCCNotice(to)) {
			send_me_noticeNA(to, ":*** The majority of files sent of this type contain "
				"malicious virri and trojans. In order to protect clients on this "
				"server, such transfers are blocked.");
			send_me_notice(to, ":*** If you trust %s and wish to receive the file "
				"from him or her, regardless of the block in effect, add him or her "
				"to your dcc allow list. Type /DCCALLOW ? for more information.",
				from->name);
			send_me_noticeNA(to, ":*** You will not receive this message again.");
			SetDCCNotice(to);
		}
		ircdlog(LOG_BAN, "%s (%s@%s) attempted to transfer banned file %s (mask %s) to "
			"%s (%s@%s) [%s]", from->name, from->username, from->host, s,
			ban->mask, to->name, to->username, to->host,
			BadPtr(ban->reason) ? "no reason" : ban->reason);
		sendto_realops_lev(DCCSEND_LEV, "%s!%s@%s attempted to transfer banned file %s "
			"(mask %s) to %s!%s@%s (%s)", from->name, from->username, MaskedHost(from),
			s, ban->mask, to->name, to->username, MaskedHost(to),
			BadPtr(ban->reason) ? "no reason" : ban->reason);
		return 1;
	}
	return 0;
}

static void show_dccallow_list(aClient *cptr)
{
	dlink_node *node;
	aClient *acptr;
	char buf[BUFSIZE], *s;
	int len = 0, count = 0;
	int plen = strlen(cptr->name) + 10;

	DLINK_FOREACH_DATA(cptr->user->dccallow_list.head, node, acptr, aClient) {
		if ((plen + len + strlen(acptr->name) + 1) > BUFSIZE || count > 10) {
			send_me_numeric(cptr, RPL_DCCALLOWLIST, buf);
			buf[0] = '\0';
			len = count = 0;
		}
		if (len > 0) {
			buf[len++] = ' ';
		}
		for (s = acptr->name; *s != '\0'; s++) {
			buf[len++] = *s;
		}
		buf[len] = '\0';
		count++;
	}
	if (count > 0) {
		send_me_numeric(cptr, RPL_DCCALLOWLIST, buf);
	}
	send_me_numericNA(cptr, RPL_ENDOFDCCALLOW);
	return;
}

static void show_dccallow_help(aClient *cptr)
{
	char **p;
	static char *helptext[] = {
		"/DCCALLOW [-]nick[,[-]nick2[,[-]nick3[...]]]",
		"/DCCALLOW <*|?>",
		" ",
		"Transmission (DCC SEND) of some filetypes are banned in order to prevent",
		"the spread of trojans and viruses. However, you can choose to bypass this",
		"block. This is done by adding clients to your dcc allow list.",
		" ",
		"For example; client Bob is attempting to send you mirc.zip. However,",
		"zip files (*.zip) are banned from transfer. So, you would issue the",
		"following command.",
		" ",
		"/DCCALLOW Bob",
		" ",
		"After receiving the file, you may then want to remove Bob from your dcc",
		"allow list. Syntax is similar to that of adding clients, but you prefix",
		"the nickname with a - (subtract) char.",
		" ",
		"/DCCALLOW -Bob",
		" ",
		"You can add and remove more than one client to and from your dcc allow",
		"list at once. Just separate your client list with, (commas), as per the",
		"following example.",
		" ",
		"/DCCALLOW Bob,-Joe,-Lisa,Homer",
		" ",
		"To view all clients on your dcc allow list, issue a DCCALLOW command with",
		"an * (asterisk) as parameter. To view this help, issue a DCCALLOW command",
		"with a ? (question mark). You cannot mix *, ? and/or nicknames together.",
		NULL
	};
	for (p = helptext; *p != '\0'; p++) {
		send_me_numeric(cptr, RPL_DCCALLOWLIST, *p);
	}
	send_me_numericNA(cptr, RPL_ENDOFDCCALLOW);
	return;
}

static void add_dccallow(aClient *cptr, aClient *acptr)
{
	dlink_add(&cptr->user->dccallow_list, acptr);
	dlink_add(&acptr->user->on_dccallow_list, cptr);
	return;
}

static void del_dccallow(aClient *sptr, aClient *acptr)
{
	dlink_node *anode, *anext = NULL;
	dlink_node *node, *next = NULL;
	aClient *cptr;

	DLINK_FOREACH_SAFE_DATA(acptr->user->dccallow_list.head, anode, anext, cptr, aClient) {
		if (sptr == cptr) {
			dlink_del(&acptr->user->dccallow_list, NULL, anode);
			DLINK_FOREACH_SAFE_DATA(sptr->user->on_dccallow_list.head, node, next, cptr,
			  aClient) {
				if (acptr == cptr) {
					dlink_del(&sptr->user->on_dccallow_list, NULL, node);
				}
			}
		}
	}
	return;
}

static int del_dccallows(HookData *hdata)
{
	aClient *sptr = hdata->sptr;
	aClient *acptr;
	dlink_node *node, *next = NULL;

	DLINK_FOREACH_SAFE_DATA(sptr->user->dccallow_list.head, node, next, acptr, aClient) {
		if (acptr != NULL) {
			del_dccallow(acptr, sptr);
		}
	}
	DLINK_FOREACH_SAFE_DATA(sptr->user->on_dccallow_list.head, node, next, acptr, aClient) {
		if (acptr != NULL) {
			del_dccallow(sptr, acptr);
		}
	}
	return 0;
}

/*
 * m_dccallow
 *	parv[0] = sender prefix
 *	parv[1] = parameter
 */
int m_dccallow(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
	char *s, *p;
	aClient *acptr;
	int del, count = 0;

	if (!MyClient(sptr)) {
		return 0;
	}
	if (parc < 2 || BadPtr(parv[1])) {
		send_me_numeric(sptr, ERR_NEEDMOREPARAMS, "DCCALLOW");
		return 0;
	}
	if (*parv[1] == '*') {
		show_dccallow_list(sptr);
		return 0;
	}
	else if (*parv[1] == '?') {
		show_dccallow_help(sptr);
		return 0;
	}

	count = dlink_length(&sptr->user->dccallow_list);
	for (s = strtoken(&p, parv[1], ","); s != NULL; s = strtoken(&p, NULL, ",")) {
		while (*s == ' ') {
			s++;
		}
		if (*s == '\0') {
			continue;
		}
		else if (*s == '-') {
			del = 1;
			s++;
		}
		else {
			del = 0;
			if (*s == '+') {
				s++;
			}
		}

		if ((acptr = find_person(s, NULL)) == NULL) {
			send_me_numeric(sptr, ERR_NOSUCHNICK, s);
			continue;
		}
		else if (acptr == sptr) {
			continue;
		}

		if (del) {
			if (dlink_find(&sptr->user->dccallow_list, acptr) == NULL) {
				char buf[NICKLEN + 31];
				ircsprintf(buf, "%s is not on your dcc allow list", acptr->name);
				send_me_numeric(sptr, RPL_DCCALLOWINFO, buf);
				continue;
			}
			del_dccallow(acptr, sptr);
			count--;
		}
		else {
			if (dlink_find(&sptr->user->dccallow_list, acptr) != NULL) {
				char buf[NICKLEN + 35];
				ircsprintf(buf, "%s is already on your dcc allow list", acptr->name);
				send_me_numeric(sptr, RPL_DCCALLOWINFO, buf);
				continue;
			}
			if (count >= MAXDCCALLOW) {
				send_me_numeric(sptr, ERR_DCCALLOWFULL, acptr->name);
				continue;
			}
			add_dccallow(sptr, acptr);
			count++;
		}
		send_me_numeric(sptr, RPL_DCCALLOWSTATUS, acptr->name,
			(del ? "removed from" : "added to"));
	}
	return 0;
}
