/*
 * RageIRCd, an advanced Internet Relay Chat daemon (IRCd)
 * modules/m_join.c: JOIN command module
 *
 * Copyright (C) 2002 the past and present IRCd developers, and others.
 * Refer to doc/authors/ for full credits and copyright details.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA.
 *
 * $Id: m_join.c,v 1.70 2004/05/08 16:12:02 rageircd Exp $
 */

#include "config.h"
#include "struct.h"
#include "common.h"
#include "sys.h"
#include "numeric.h"
#include "msg.h"
#include "channel.h"
#include "h.h"
#include "memory.h"
#include "send.h"
#include "hook.h"
#include "modules.h"
#include "xmode.h"
#include <time.h>
#include <sys/stat.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>

Hook *h_post_join = NULL;

static int auto_join_connect(HookData *);
static int auto_join_oper(HookData *);

Module MOD_HEADER(m_join) = {
	"modules/commands/m_join.so",
	"/JOIN command",
	6, "$Revision: 1.70 $"
};

int MOD_LOAD(m_join)()
{
	h_post_join = register_hook(&MOD_HEADER(m_join), "h_post_join");

	register_hook_event(&MOD_HEADER(m_join), h_post_register_user, auto_join_connect);
	register_hook_event(&MOD_HEADER(m_join), h_post_oper, auto_join_oper);

	register_command(&MOD_HEADER(m_join), &CMD_JOIN, m_join);

	MOD_SET_FLAG(&MOD_HEADER(m_join), MOD_FLAG_PERM);

	return MOD_SUCCESS;
}

int MOD_UNLOAD(m_join)()
{
	return MOD_SUCCESS;
}

static int auto_join_connect(HookData *hdata)
{
	if (!BadPtr(GeneralConfig.join_on_connect)) {
		char *parv[3] = { hdata->sptr->name, GeneralConfig.join_on_connect, NULL };
		m_join(hdata->cptr, hdata->sptr, 2, parv);
	}
	return 0;
}

static int auto_join_oper(HookData *hdata)
{
	ConfigItem_oper *oper = (ConfigItem_oper *)hdata->v;

	if (!BadPtr(oper->join_on_oper)) {
		char *parv[3] = { hdata->sptr->name, oper->join_on_oper, NULL };
		m_join(hdata->cptr, hdata->sptr, 2, parv);
	}
	return 0;
}

/*
 * m_join
 *	parv[0] = sender prefix
 *	parv[1] = channel name
 *	parv[2] = channel password (key)
 */
int m_join(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
	static char jbuf[BUFSIZE];
	char *name, *key = NULL, *p, *p2;
	aChannel *chptr;
	chanMember *cm;
	ConfigItem_ban *ban = NULL;
	int i, flags = 0, chanlen = 0, exists = 0;
#ifdef ANTI_SPAMBOT
	int successful_join_count = 0;
#endif

	if (sptr->user == NULL) {
		return 0;
	}
	if (parc < 2 || *parv[1] == '\0') {
		send_me_notice(sptr, ":Something's wrong! parc %d", parc);
		send_me_numeric(sptr, ERR_NEEDMOREPARAMS, "JOIN");
		return 0;
	}

	*jbuf = '\0';
	for (i = 0, name = strtoken(&p, parv[1], ","); name != NULL; name = strtoken(&p, NULL, ",")) {
		if (!check_channelname(sptr, (unsigned char *)name)) {
			continue;
		}
		if ((chanlen = strlen(name)) > CHANNELLEN) {
			name[CHANNELLEN] = '\0';
			chanlen = CHANNELLEN;
		}
		if (*name == '&' && !MyConnect(sptr)) {
			continue;
		}
		if (*name == '0' && *(name + 1) == '\0') {
			*jbuf = '\0';
		}
		else if (!IsChannelName(name)) {
			if (MyClient(sptr)) {
				send_me_numeric(sptr, ERR_NOSUCHCHANNEL, name);
			}
			continue;
		}
		if (*jbuf != '\0') {
			strcat(jbuf, ",");
		}
		strncat(jbuf, name, sizeof(jbuf) - i - 1);
		i += chanlen + 1;
	}

	p = NULL;
	if (parv[2] != NULL) {
		key = strtoken(&p2, parv[2], ",");
	}
	parv[2] = NULL;

	for (name = strtoken(&p, jbuf, ","); name != NULL; key = (key != NULL) ? 
	  strtoken(&p2, NULL, ",") : NULL, name = strtoken(&p, NULL, ",")) {
		if (*name == '0') {
			if (sptr->user->channel == NULL) {
				continue;
			}
			while ((cm = sptr->user->channel) != NULL) {
				send_channel_msg_local_copy(sptr, cm->chptr, &CMD_PART, PREFIX_NONE, "");
				remove_user_from_channel(sptr, cm->chptr);
			}
#ifdef ANTI_SPAMBOT
			if (MyConnect(sptr) && !HasMode(sptr, UMODE_OPER)) {
				if (sptr->localUser->join_part_count >= spam_num) {
					sendto_realops_lev(SPAM_LEV, "User %s (%s@%s) is a possible spambot",
						sptr->name, sptr->username, MaskedHost(sptr));
					sptr->localUser->oper_warn_countdown = OPER_SPAM_COUNTDOWN;
				}
				else {
					int t_delta = timeofday - sptr->localUser->last_part_time;
					if (t_delta > JOIN_LEAVE_COUNT_EXPIRE_TIME) {
						int decrement_count = (t_delta / JOIN_LEAVE_COUNT_EXPIRE_TIME);
						if (decrement_count > sptr->localUser->join_part_count) {
							sptr->localUser->join_part_count = 0;
						}
						else {
							sptr->localUser->join_part_count -= decrement_count;
						}
					}
					else {
						if ((timeofday - sptr->localUser->last_join_time) < spam_time) {
							sptr->localUser->join_part_count++;
						}
					}
					sptr->localUser->last_part_time = timeofday;
				}
			}
#endif
			sendto_serv_msg_butone(cptr, sptr, NULL, &CMD_JOIN, "0");
			continue;
		}

		exists = ChannelExists(name) ? 1 : 0;
		if (MyConnect(sptr)) {
			if (!GeneralConfig.custom_channels && !exists && !HasMode(sptr, UMODE_OPER)) {
				send_me_numeric(sptr, ERR_FUNCDISABLED, "JOIN");
				return 0;
			}
			if ((ban = find_ban(BAN_CHANNEL, name)) != NULL && !HasMode(sptr, UMODE_ADMIN)) {
				send_me_notice(sptr, ":*** Channel %s is banned (%s)", name,
				BadPtr(ban->reason) ? "no reason specified" : ban->reason);
				return 0;
			}
			if ((sptr->user->joined >= GeneralConfig.max_chans_per_user) && (!HasMode(sptr, UMODE_OPER)
			  || (sptr->user->joined >= (GeneralConfig.max_chans_per_user * 3)))) {
				send_me_numeric(sptr, ERR_TOOMANYCHANNELS, name);
#ifdef ANTI_SPAMBOT
				if (successful_join_count > 0) {
					sptr->localUser->last_join_time = timeofday;
				}
#endif
				return 0;
			}

#ifdef ANTI_SPAMBOT
			flags = (!exists) ? CMODE_CHANOP : 0;
			if (!flags) {
				successful_join_count++;
			}

			if (sptr->localUser->join_part_count >= spam_num) {
				if (sptr->localUser->oper_warn_countdown > 0) {
					sptr->localUser->oper_warn_countdown--;
				}
				else {
					sptr->localUser->oper_warn_countdown = 0;
				}
				if (!sptr->localUser->oper_warn_countdown) {
					sendto_realops_lev(SPAM_LEV, "User %s (%s@%s) trying to join %s is a "
						"possible spambot", sptr->name, sptr->username,
						MaskedHost(sptr), name);
					sptr->localUser->oper_warn_countdown = OPER_SPAM_COUNTDOWN;
				}
#ifndef ANTI_SPAMBOT_WARN_ONLY
				return 0;
#endif
			}
#endif
		}
		else if (!exists) {
			ts_warn("User on %s is remotely JOINing new channel", sptr->user->server);
		}

		if ((chptr = get_channel(sptr, name, CREATE, NULL)) == NULL) {
			continue;
		}

		if (MyConnect(sptr) && (i = can_join(sptr, chptr, key))) {
			send_me_numeric(sptr, i, name, (i == ERR_NEEDREGGEDNICK) ? "join" : "");
#ifdef ANTI_SPAMBOT
			if (successful_join_count > 0) {
				successful_join_count--;
			}
#endif
			continue;
		}
		if (IsMember(sptr, chptr)) {
			continue;
		}

		add_user_to_channel(chptr, sptr, flags);

		if (MyClient(sptr) && flags) {
			chptr->channelts = timeofday;

			sendto_serv_capab_msg_butone(cptr, &me, NULL, 0, CAP_SUID|CAP_TOK1,
				&CMD_SJOIN, "%ld %s + :@%s", chptr->channelts, chptr->chname,
				sptr->name);
			if (HasSUID(sptr)) {
				sendto_serv_capab_msg_butone(cptr, &me, NULL, CAP_SUID|CAP_TOK1, 0,
					&CMD_SJOIN, "%B %s + :@!%s", chptr->channelts, chptr->chname,
					sptr->id.string);
			}
			else {
				sendto_serv_capab_msg_butone(cptr, &me, NULL, CAP_SUID|CAP_TOK1, 0,
					&CMD_SJOIN, "%B %s + :@%s", chptr->channelts, chptr->chname, 
					sptr->name);
			}
		}
		else {
			sendto_serv_msg_butone(cptr, sptr, NULL, &CMD_SJOIN, "%ld %s", chptr->channelts,
				chptr->chname);
		}
		send_channel_msg_local_copy(sptr, chptr, &CMD_JOIN, PREFIX_NONE, "");

		if (MyClient(sptr)) {
			HookData hdata;

			del_invite(sptr, chptr);
			if (*chptr->topic != '\0') {
				send_me_numeric(sptr, RPL_TOPIC, name, chptr->topic);
				send_me_numeric(sptr, RPL_TOPICWHOTIME, name, chptr->topic_nick,
					chptr->topic_time);
			}

			hdata.cptr = cptr;
			hdata.sptr = sptr;
			hdata.chptr = chptr;
			hook_run(h_post_join, &hdata);
		}
	}

#ifdef ANTI_SPAMBOT
	if (MyConnect(sptr) && successful_join_count) {
		sptr->localUser->last_join_time = timeofday;
	}
#endif

	return 0;
}
