/*
 * RageIRCd, an advanced Internet Relay Chat daemon (IRCd)
 * modules/m_kill.c: KILL command module
 *
 * Copyright (C) 2002 the past and present IRCd developers, and others.
 * Refer to doc/authors/ for full credits and copyright details.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA.
 *
 * $Id: m_kill.c,v 1.24 2004/05/12 01:51:18 rageircd Exp $
 */

#include "config.h"
#include "struct.h"
#include "common.h"
#include "sys.h"
#include "numeric.h"
#include "msg.h"
#include "channel.h"
#include "h.h"
#include "memory.h"
#include "modules.h"
#include "xmode.h"
#include <time.h>
#include <sys/stat.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>

Module MOD_HEADER(m_kill) = {
	"modules/commands/m_kill.so",
	"/KILL command",
	6, "$Revision: 1.24 $"
};

int MOD_LOAD(m_kill)()
{
	register_command(&MOD_HEADER(m_kill), &CMD_KILL, m_kill);
	return MOD_SUCCESS;
}

int MOD_UNLOAD(m_kill)()
{
	return MOD_SUCCESS;
}

/*
 * m_kill
 *	parv[0] = sender prefix
 *	parv[1] = target
 *	parv[2] = path
 */
int m_kill(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
	static char buf[BUFSIZE];
	aClient *acptr;
	char *user, *path, *reason, *p, *nick;
	char mypath[KILLLEN + 1], mymsg[KILLLEN + 1];
	int chasing = 0, kcount = 0;

	if (!IsPrivileged(cptr)) {
		send_me_numericNA(sptr, ERR_NOPRIVILEGES);
		return 0;
	}
	if (parc < 2 || *parv[1] == '\0') {
		send_me_numeric(sptr, ERR_NEEDMOREPARAMS, "KILL");
		return 0;
	}

	user = parv[1];
	path = parv[2];

	if (!BadPtr(path)) {
		if (strlen(path) > KILLLEN) {
			path[KILLLEN] = '\0';
		}
	}

	if (MyClient(sptr)) {
		user = canonize(user);
	}

	for (p = NULL, nick = strtoken(&p, user, ","); nick != NULL; nick = strtoken(&p, NULL, ",")) {
		chasing = 0;
		if (MyClient(sptr) || (*nick != '!')) {
			if ((acptr = find_client(nick, NULL)) == NULL) {
				if ((acptr = get_history(nick, (long)KILLCHASETIMELIMIT)) == NULL) {
					target_left(sptr, nick, MSG_KILL, NULL);
					continue;
				}
				send_me_notice(sptr, ":KILL target changed from %s to %s", nick,
					acptr->name);
				chasing = 1;
			}
		}
		else {
			if ((acptr = find_target(nick)) == NULL) {
				send_me_notice(sptr, ":KILL lost in transit - user signoff before "
					"delivery/desynch (ID: %s)", nick);
				continue;
			}
		}

		if (MyClient(cptr) && ((MyConnect(acptr) && !OPHasFlag(cptr, OFLAG_LKILL)) ||
		  (!MyConnect(acptr) && !OPHasFlag(cptr, OFLAG_GKILL)))) {
			send_me_numericNA(sptr, ERR_NOPRIVILEGES);
			continue;
		}
		if (IsServer(acptr) || IsMe(acptr) || (HasMode(acptr, UMODE_SCLIENT) && !IsServer(cptr))) {
			send_me_numericNA(sptr, ERR_CANTKILLSERVER);
			ircdlog(LOG_KILL, "client %s attempted to kill server %s", get_client_name(cptr, FALSE),
				acptr->name);
			continue;
		}

		kcount++;
		if (!IsServer(sptr) && kcount > MAXKILLS) {
			send_me_notice(sptr, ":Too many targets, kill list truncated (maximum %d)", MAXKILLS);
			break;
		}

		if (MyClient(sptr)) {
			char myname[HOSTLEN + 1], *s;

			if (!BadPtr(path)) {
				ircsnprintf(mymsg, KILLLEN + 1, "(%s)", path);
				reason = mymsg;
			}
			else {
				reason = "(No reason specified)";
			}

			strncpy(myname, me.name, HOSTLEN + 1);
			if ((s = strchr(myname, '.')) != NULL) {
				*s = '\0';
			}

			ircsnprintf(mypath, KILLLEN + 1, "%s!%s!%s", myname, MaskedHost(sptr), sptr->name);
		}
		else {
			if (BadPtr(path) || (reason = strchr(path, ' ')) == NULL) {
				path = sptr->name;
				reason = "(No reason specified)";
			}
			else {
				*reason++ = '\0';
			}
			strncpyzt(mypath, path, KILLLEN + 1);
		}

		sendto_realops_lev(SKILL_LEV, "Received KILL message for %s!%s@%s. "
			"From: %s Path: %s %s", acptr->name, acptr->username,
			acptr->user ? MaskedHost(acptr) : "<unknown>", parv[0], mypath, reason);

		if (HasMode(sptr, UMODE_OPER)) {
			ircdlog(LOG_KILL, "KILL for %s!%s@%s. From: %s Path: %s %s", acptr->name,
				*acptr->username != '\0' ? acptr->username : "<unknown>",
				*acptr->host != '\0' ? MaskedHost(acptr) : "<unknown>", parv[0], mypath,
				reason);
		}

		if (!MyConnect(acptr) || !MyConnect(sptr) || !HasMode(sptr, UMODE_OPER)) {
			sendto_serv_msg_butone(cptr, sptr, acptr, &CMD_KILL, ":%s %s", mypath, reason);
			if (chasing && IsServer(cptr)) {
				sendto_one_client_real(cptr, acptr, &me, &CMD_KILL, ":%s %s", mypath, reason);
			}
			SetKilled(acptr);
		}
		if (MyConnect(acptr)) {
			sendto_one_client_prefixed(acptr, sptr, &CMD_KILL, ":%s %s", mypath, reason);
		}
		if (MyConnect(acptr) && MyConnect(sptr) && HasMode(sptr, UMODE_OPER)) {
			ircsprintf(buf, "Local kill by %s %s", sptr->name, reason);
		}
		else {
			ircsprintf(buf, "Killed (%s %s)", sptr->name, reason);
		}
		if (exit_client(cptr, acptr, sptr, buf) == FLUSH_BUFFER) {
			return FLUSH_BUFFER;
		}
	}
	return 0;
}
