/*
 * RageIRCd, an advanced Internet Relay Chat daemon (IRCd)
 * modules/m_snick.c: SNICK protocol module
 *
 * Copyright (C) 2002 the past and present IRCd developers, and others.
 * Refer to doc/authors/ for full credits and copyright details.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA.
 *
 * $Id: m_snick.c,v 1.32 2004/04/07 13:32:25 rageircd Exp $
 */

#include "config.h"
#include "struct.h"
#include "common.h"
#include "sys.h"
#include "numeric.h"
#include "msg.h"
#include "channel.h"
#include "h.h"
#include "memory.h"
#include "modules.h"
#include "xmode.h"
#include <time.h>
#include <sys/stat.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>

Module MOD_HEADER(m_snick) = {
	"modules/commands/m_snick.so",
	"SNICK protocol",
	5, "$Revision: 1.32 $"
};

int MOD_LOAD(m_snick)()
{
	register_command(&MOD_HEADER(m_snick), &CMD_SNICK, m_snick);
	return MOD_SUCCESS;
}

int MOD_UNLOAD(m_snick)()
{
	return MOD_SUCCESS;
}

/*
 * m_snick
 *	parv[0] = sender prefix
 *	parv[1] = nickname
 *	parv[2] = TS or !TS
 *	parv[3] = hop
 *	parv[4] = username
 *	parv[5] = hostname
 *	parv[6] = ip or !ip
 *	parv[7] = maskedhost
 *	parv[8] = servername or !suid
 *	parv[9] = servicestamp or !servicestamp
 *	parv[10] = user modes
 *	parv[11] = gcos
 */
int m_snick(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
	char *b64id = NULL, *p, ubuf[16];
	unsigned long newid, newip;
	aClient *acptr, *uplink = NULL;
	long newts, oldts;
	int sameuser = 0, mindex;

	if (!IsServer(sptr) || parc < 12) {
		return 0;
	}

	newts = (*parv[2] == '!') ? base64dec(parv[2] + 1) : strtoul(parv[2], NULL, 0);
	newip = (*parv[6] == '!') ? base64dec(parv[6] + 1) : strtoul(parv[6], NULL, 0);

	Debug((DEBUG_DEBUG, "SNICK[%s] got TS(%s) [%ld] and IP(%s) [%ld]", parv[1], parv[2],
		newts, parv[6], newip));

	/* Check for a nick collision. */
	do {
		if ((acptr = find_client(parv[1], NULL)) == NULL) {
			break;
		}

		if (IsUnknown(acptr)) {
			Debug((DEBUG_DEBUG, "SNICK collision (%s) with unknown", parv[1]));

			if (MyConnect(acptr)) {
				exit_client(NULL, acptr, &me, "Overridden");
				break;
			}

			if (acptr->user == NULL) {
				sendto_realops_lev(SKILL_LEV, "Nick collision on %s",
					parv[1]);
				sendto_serv_msg_butone(NULL, &me, NULL, &CMD_KILL,
					"%s :Nick collision", acptr->name);
				SetKilled(acptr);
				return exit_client(cptr, acptr, &me, "Overridden");
			}
		}

		Debug((DEBUG_DEBUG, "ARGH! Client collision: %s", get_client_name(acptr, FALSE)));

		ASSERT(acptr->user != NULL);

		oldts = acptr->tsinfo;

		/* Collision logic: we are either dealing with a race condition
		 * (two users signing onto different servers at the same time),
		 * or a server<->server synch with two identical nicknames.
		 *
		 * TS are equal or zero: kill both
		 * Same user: kill oldest
		 * Different user: kill youngest
		 */

		/* TS are equal or either TS is zero -- kill both */
		if (!newts || !oldts || (newts == oldts)) {
			ircstp->is_kill++;

			sendto_realops_lev(SKILL_LEV, "Nick collision on %s (all nicks)",
				acptr->name);
			sendto_serv_msg_butone(NULL, &me, NULL, &CMD_KILL, "%s :Nick collision",
				acptr->name);

			SetKilled(acptr);
			return exit_client(sptr, acptr, &me, "Nick collision");
		}

		sameuser = ((acptr->ip.s_addr == newip) && !mycmp(acptr->username, parv[4])) ? 1 : 0;

		if ((sameuser && (newts < oldts)) || (!sameuser && (newts > oldts))) {
			/* Either:
			 *	- same users, but they're older
			 *	- diff users, but they're younger
			 * We can safely ignore this SNICK now, no kills required!
			 * The other side will exit their client and send the
			 * appropriate kills backwards. Then, it will accept the
			 * SNICK for our client.
			 */
			Debug((DEBUG_DEBUG, "SNICK[%s] ignoring, our client won!", parv[1]));
			return 0;
		}

		/* OK -- we have to kill our client. So be it! Send the message
		 * backwards to the rest of the network.
		 */
		Debug((DEBUG_DEBUG, "Nick collision on %s", parv[1]));
		ircstp->is_kill++;

		sendto_realops_lev(SKILL_LEV, "Nick collision on %s", acptr->name);
		sendto_serv_msg_butone(sptr, &me, NULL, &CMD_KILL, "%s :Nick collision",
			acptr->name);

		/* Tell our client what's going on. Exit them, and then gracefully
		 * accept the remote servers SNICK.
		 */
		send_me_numeric(acptr, ERR_NICKCOLLISION, acptr->name);

		SetKilled(acptr);
		exit_client(sptr, acptr, &me, "Nick collision");

		break;
	} while (0);

	/* Accept the SNICK and validate the clients server. */
	if (*parv[8] == '!' && (*(parv[8] + 1) != '\0')) {
		b64id = parv[8] + 1;

		Debug((DEBUG_DEBUG, "SNICK[%s] got server suid [%s]", parv[1], b64id));

		if ((acptr = find_by_base64_id(b64id)) != NULL) {
			sendto_realops_lev(DEBUG_LEV, "IDENTITIY COLLISION! (%s[%s][%s] <> "
				"%s[%s][%s])", sptr->name, sptr->uplink->name, b64id,
				acptr->name, acptr->uplink->name, acptr->id.string);
			exit_client(cptr, acptr, &me, "Identity Collision");
		}
		uplink = find_serv_by_base64_id(b64id, &newid);
	}
	else {
		uplink = find_server(parv[8], NULL);
	}
	if (uplink == NULL) {
		sendto_realops_lev(DEBUG_LEV, "Unknown server in SNICK for %s", parv[1]);
		return 0;
	}

	Debug((DEBUG_DEBUG, "SNICK[%s] building client", parv[1]));

	acptr = sptr;
	sptr = make_client(cptr, uplink);

	if (IsULine(uplink)) {
		SetULine(sptr); /* If the uplink is U:lined, the client is also U:lined */
	}

	add_client_to_list(sptr);
	strcpy(sptr->name, parv[1]);
	sptr->hopcount = atoi(parv[3]);
	sptr->tsinfo = newts;
	sptr->ip.s_addr = newip;
	inetntop(&sptr->ip, sptr->hostip, HOSTIPLEN + 1);

	if (b64id != NULL) {
		sptr->id.id = newid;
		strncpy(sptr->id.string, b64id, 8);
		add_userid_to_serv(uplink, sptr);
		SetSUID(sptr);
	}

	add_to_client_hash_table(parv[1], sptr);

	strncpyzt(sptr->username, parv[4], USERLEN + 1);
	strncpyzt(sptr->host, parv[5], HOSTLEN + 1);
	strncpyzt(sptr->info, parv[11], REALLEN + 1);

	make_user(sptr);
	sptr->user->server = find_or_add(uplink->name);
	sptr->user->servicestamp = (*parv[9] == '!') ? base64dec(parv[9] + 1) : strtoul(parv[9], NULL, 0);

	SetClient(sptr);

	/* Parse the specified user modes */
	for (p = parv[10]; *p != '\0'; p++) {
		if (*p == '-' || (*p == '+')) {
			continue; /* Skip + and - flags */
		}

		/* Loop through the table to find the flag */
		/* We now use the index map, no loop required! */
		if ((mindex = usermodes->map[(unsigned char)*p]) == -1) {
			/* Paranoia: if this is true, we don't support the user mode. For
			 * the time being, this _can_ happen. However it shouldn't, once
			 * server<->server the USERMODES capability is added.
			 */
			sendto_realops_lev(DEBUG_LEV, "Warning! Skipping unknown user "
				"mode flag %c in SNICK for %s", *p, sptr->name);
		}
		else {
			/* Add it, if it's global */
			sptr->umode |= (usermodes->table[mindex].mode & SEND_UMODES);
		}
	}

	/* New clients are distributed through networks using SNICK now */
	if (!IsULine(sptr) && !SendingBurst(acptr)) {
		sendto_realops_lev(GCCONN_LEV, "Client connecting at %s: %s (%s@%s)", uplink->name,
			sptr->name, sptr->username, sptr->host);
	}

	/* Update internal counters */
	if (++Count.total > Count.max_tot) {
		Count.max_tot = Count.total;
	}
	if (HasMode(sptr, UMODE_OPER)) {
		Count.oper++;
	}
	if (HasMode(sptr, UMODE_INVISIBLE)) {
		Count.invisi++;
	}
	if (HasMode(sptr, UMODE_MASKED)) {
		if (*parv[7] == '*') {
			sendto_realops_lev(DEBUG_LEV, "SNICK for %s contains no maskedhost, "
				"when user is masked", sptr->name);
			DelMode(sptr, UMODE_MASKED);
		}
		else {
			strncpyzt(sptr->user->maskedhost, parv[7], HOSTLEN + 1);
		}
	}

	/* Send the user mode changes out to the servers */
	send_umode(NULL, sptr, 0, SEND_UMODES, ubuf);
	if (*ubuf == '\0') {
		*ubuf = '+';
		*(ubuf + 1) = '\0';
	}

	hash_check_watch(sptr, RPL_LOGON);

	send_nick_to_serv_butone(cptr, sptr, ubuf); /* Send backwards to the network */

	return 0;
}
