/*
 * RageIRCd, an advanced Internet Relay Chat daemon (IRCd)
 * modules/m_squit.c: SQUIT command module
 *
 * Copyright (C) 2002 the past and present IRCd developers, and others.
 * Refer to doc/authors/ for full credits and copyright details.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA.
 *
 * $Id: m_squit.c,v 1.23 2004/04/08 20:24:34 rageircd Exp $
 */

#include "config.h"
#include "struct.h"
#include "common.h"
#include "sys.h"
#include "numeric.h"
#include "msg.h"
#include "channel.h"
#include "h.h"
#include "memory.h"
#include "modules.h"
#include "xmode.h"
#include <time.h>
#include <sys/stat.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>

Module MOD_HEADER(m_squit) = {
	"modules/commands/m_squit.so",
	"/SQUIT command",
	5, "$Revision: 1.23 $"
};

int MOD_LOAD(m_squit)()
{
	register_command(&MOD_HEADER(m_squit), &CMD_SQUIT, m_squit);
	return MOD_SUCCESS;
}

int MOD_UNLOAD(m_squit)()
{
	return MOD_SUCCESS;
}

/*
 * m_squit
 *	parv[0] = sender prefix
 *	parv[1] = servername
 *	parv[2] = optional comment
 */
int m_squit(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
	char *server, *comment = (parc > 2) ? parv[2] : sptr->name;
	aClient *acptr = NULL;

	if (!IsPrivileged(sptr)) {
		send_me_numericNA(sptr, ERR_NOPRIVILEGES);
		return 0;
	}
	if (parc > 1) {
		server = parv[1];

		if (IsServer(cptr)) {
			acptr = find_server_target(server);
		}
		else {
			for (acptr = client; (acptr = next_client(acptr, server)); acptr = acptr->next) {
				if (IsServer(acptr) || IsMe(acptr)) {
					break;
				}
			}
		}
		if (acptr && IsMe(acptr)) {
			ASSERT(cptr->localClient != NULL);

			acptr = cptr;
			server = cptr->localClient->sockhost;
		}
	}
	else {
		if (IsClient(cptr)) {
			send_me_numeric(sptr, ERR_NEEDMOREPARAMS, "SQUIT");
			return 0;
		}

		ASSERT(cptr->localClient != NULL);

		acptr = cptr;
		server = cptr->localClient->sockhost;
	}
	if (acptr == NULL) {
		send_me_numeric(sptr, ERR_NOSUCHSERVER, server);
		return 0;
	}

	if (MyClient(sptr) && ((MyConnect(acptr) && !OPHasFlag(sptr, OFLAG_LROUTE)) ||
	  (!MyConnect(acptr) && !OPHasFlag(sptr, OFLAG_GROUTE)))) {
		send_me_numericNA(sptr, ERR_NOPRIVILEGES);
		return 0;
	}
	if (MyConnect(acptr)) {
		sendto_gnotice("Received SQUIT %s from %s (%s)", acptr->name,
			get_client_name(sptr, HIDE_IP), comment);
		sendto_serv_msg_butone(NULL, &me, NULL, &CMD_GNOTICE, ":Received SQUIT %s from %s (%s)",
			acptr->name, get_client_name(sptr, HIDE_IP), comment);
		ircdlog(LOG_SERVER, "SQUIT from %s: %s (%s)", parv[0], server, comment);

		if (cptr == sptr) {
			exit_client(&me, acptr, sptr, comment);
			return FLUSH_BUFFER;
		}
		return exit_client(&me, acptr, sptr, comment);
	}
	if (sptr->from == acptr->from) {
		sendto_realops_lev(DEBUG_LEV, "Exiting server %s due to upstream SQUIT by %s [%s]",
			acptr->name, sptr->name, comment);
		return exit_client(cptr, acptr, sptr, comment);
	}
	if (!CapUNCONNECT(acptr->from)) {
		sendto_realops_lev(DEBUG_LEV, "Exiting server %s due to non-unconnect server %s [%s]",
			acptr->name, acptr->from->name, comment);
		return exit_client(&me, acptr, sptr, comment);
	}

	sendto_realops_lev(DEBUG_LEV, "Passing along SQUIT for %s by %s [%s]", acptr->name, sptr->name,
		comment);
	sendto_one_client_real(acptr->from, acptr, sptr, &CMD_SQUIT, ":%s", comment);
	return 0;
}
