/*
 * RageIRCd, an advanced Internet Relay Chat daemon (IRCd)
 * modules/m_stats.c: STATS command module
 *
 * Copyright (C) 2002 the past and present IRCd developers, and others.
 * Refer to doc/authors/ for full credits and copyright details.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA.
 *
 * $Id: m_stats.c,v 1.84 2004/04/09 20:31:51 rageircd Exp $
 */

#include "config.h"
#include "struct.h"
#include "common.h"
#include "sys.h"
#include "numeric.h"
#include "msg.h"
#include "channel.h"
#include "h.h"
#include "memory.h"
#include "res.h"
#include "fd.h"
#include "hook.h"
#include "modules.h"
#include "xmode.h"
#include "conf2.h"
#include <time.h>
#include <sys/stat.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>

Module MOD_HEADER(m_stats) = {
	"modules/commands/m_stats.so",
	"/STATS command",
	5, "$Revision: 1.84 $"
};

int MOD_LOAD(m_stats)()
{
	register_command(&MOD_HEADER(m_stats), &CMD_STATS, m_stats);
	return MOD_SUCCESS;
}

int MOD_UNLOAD(m_stats)()
{
	return MOD_SUCCESS;
}

static void stats_adns(aClient *, char *);
static void stats_ban_file(aClient *, char *);
static void stats_link(aClient *, char *);
static void stats_except(aClient *, char *);
static void stats_event(aClient *, char *);
static void stats_fd(aClient *, char *);
static void stats_ban_gcos(aClient *, char *);
static void stats_hook(aClient *, char *);
static void stats_allow(aClient *, char *);
static void stats_iphash(aClient *, char *);
static void stats_ban(aClient *, char *);
static void stats_linkinfo(aClient *, char *);
static void stats_command(aClient *, char *);
static void stats_usercount(aClient *, char *);
static void stats_oper(aClient *, char *);
static void stats_listener(aClient *, char *);
static void stats_ban_nickchan(aClient *, char *);
static void stats_rusage(aClient *, char *);
static void stats_scache(aClient *, char *);
static void stats_throttle(aClient *, char *);
static void stats_server(aClient *, char *);
static void stats_uline(aClient *, char *);
static void stats_uptime(aClient *, char *);
static void stats_vhost(aClient *, char *);
static void stats_class(aClient *, char *);
static void stats_memory(aClient *, char *);
static void stats_info(aClient *, char *);

typedef struct _stats_request StatsRequest;
struct _stats_request {
	char higher;
	char lower;
	void (*func)(aClient *, char *);
	unsigned short options;
};

#define STATS_NONE		0x00
#define STATS_OPERONLY		0x01
#define STATS_ADMINONLY		0x02
#define STATS_NETADMINONLY	0x04
#define STATS_PACESIMPLE	0x08
#define STATS_PACEINTENSE	0x10

static StatsRequest stats_table[] = {
	{ 'A',	'a',	stats_adns,		STATS_ADMINONLY },
	{ 'B',	'b',	stats_ban_file,		STATS_PACEINTENSE },
	{ 'C',	'c',	stats_link,		STATS_PACESIMPLE },
	{ 'E',	0,	stats_except,		STATS_OPERONLY },
	{ 0,	'e',	stats_event,		STATS_ADMINONLY },
	{ 'F',	'f',	stats_fd,		STATS_ADMINONLY },
	{ 'G',	'g',	stats_ban_gcos,		STATS_OPERONLY },
	{ 'H',	'h',	stats_hook,		STATS_ADMINONLY },
	{ 'I',	0,	stats_allow,		STATS_PACESIMPLE },
	{ 0,	'i',	stats_iphash,		STATS_OPERONLY },
	{ 'K',	'k',	stats_ban,		STATS_PACEINTENSE },
	{ 'L',	'l',	stats_linkinfo,		STATS_PACEINTENSE },
	{ 'M',	'm',	stats_command,		STATS_PACEINTENSE },
	{ 'N',	'n',	stats_usercount,	STATS_PACESIMPLE },
	{ 'O',	'o',	stats_oper,		STATS_PACESIMPLE },
	{ 'P',	'p',	stats_listener,		STATS_OPERONLY },
	{ 'Q',	'q',	stats_ban_nickchan,	STATS_PACESIMPLE },
	{ 'R',	'r',	stats_rusage,		STATS_OPERONLY },
	{ 'S',	's',	stats_scache,		STATS_OPERONLY },
	{ 'T',	0,	stats_throttle,		STATS_OPERONLY },
	{ 0,	't',	stats_server,		STATS_OPERONLY },
	{ 'U',	0,	stats_uline,		STATS_PACESIMPLE },
	{ 0,	'u',	stats_uptime,		STATS_NONE },
	{ 'V',	'v',	stats_vhost,		STATS_OPERONLY },
	{ 'Y',	'y',	stats_class,		STATS_PACESIMPLE },
	{ 'Z',	'z',	stats_memory,		STATS_ADMINONLY },
	{ '?',	'?',	stats_info,		STATS_PACEINTENSE },

	{ 0,	0,	NULL,			0 }
};

static void stats_adns(aClient *sptr, char *unused)
{
	report_dns_nameservers(sptr);
}

static void stats_ban_file(aClient *sptr, char *unused)
{
	dlink_node *node;
	ConfigItem_ban *ban;

	DLINK_FOREACH_DATA(conf_ban_list.head, node, ban, ConfigItem_ban) {
		if (ban->flags & BAN_FILE) {
			send_me_numeric(sptr, RPL_STATSKLINE, 'B', ban->mask,
				BadPtr(ban->reason) ? "no reason specified" : ban->reason);
		}
	}
}

static void stats_link(aClient *sptr, char *unused)
{
	dlink_node *node;
	ConfigItem_link *linkp;

	if (GeneralConfig.hide_ulined_servers && !HasMode(sptr, UMODE_OPER)) {
		send_me_numericNA(sptr, ERR_NOPRIVILEGES);
		return;
	}
	DLINK_FOREACH_DATA(conf_link_list.head, node, linkp, ConfigItem_link) {
		send_me_numeric(sptr, RPL_STATSCLINE, "C",
			HasMode(sptr, UMODE_RSTAFF) ? linkp->host : "0.0.0.0",
			linkp->servername, linkp->port, linkp->class->name);
	}
}

static void stats_except(aClient *sptr, char *unused)
{
	dlink_node *node;
	ConfigItem_except *except;

	DLINK_FOREACH_DATA(conf_except_list.head, node, except, ConfigItem_except) {
		send_me_numeric(sptr, RPL_STATSKLINE, 'E', except->mask, "*");
	}
}

static void stats_event(aClient *sptr, char *unused)
{
	show_events(sptr);
}

static void stats_fd(aClient *sptr, char *unused)
{
#ifdef DEBUGMODE
	fd_report(sptr);
#endif
}

static void stats_ban_gcos(aClient *sptr, char *unused)
{
	dlink_node *node;
	ConfigItem_ban *ban;

	DLINK_FOREACH_DATA(conf_ban_list.head, node, ban, ConfigItem_ban) {
		if (ban->flags & BAN_GCOS) {
			send_me_numeric(sptr, RPL_STATSKLINE, (ban->item.temp) ? 'g' : 'G',
				ban->mask,
				BadPtr(ban->reason) ? "no reason specified" : ban->reason);
		}
	}
}

static void stats_hook(aClient *sptr, char *unused)
{
	hook_report(sptr);
}

static void stats_allow(aClient *sptr, char *unused)
{
	dlink_node *node;
	ConfigItem_allow *allow;

	DLINK_FOREACH_DATA(conf_allow_list.head, node, allow, ConfigItem_allow) {
		send_me_numeric(sptr, RPL_STATSILINE, allow->ipaddr, allow->hostname,
			allow->max_per_ip, allow->class->name);
	}
}

static void stats_iphash(aClient *sptr, char *unused)
{
	list_ip_hash(sptr);
}

static void stats_ban(aClient *sptr, char *unused)
{
	dlink_node *node;
	ConfigItem_ban *ban;
	char c;

	DLINK_FOREACH_DATA(conf_ban_list.head, node, ban, ConfigItem_ban) {
		if ((ban->flags & (BAN_USERHOST|BAN_SVS)) == (BAN_USERHOST|BAN_SVS)) {
			c = 'A';
		}
		else if (ban->flags & BAN_USERHOST) {
			c = (ban->item.temp) ? 'k' : 'K';
		}
		else if (ban->flags & BAN_IP) {
			c = (ban->item.temp) ? 'z' : 'Z';
		}
		else if (ban->flags & BAN_SERVER) {
			c = 'S';
		}
		else {
			continue;
		}
		send_me_numeric(sptr, RPL_STATSKLINE, c, ban->mask,
			BadPtr(ban->reason) ? "no reason specified" : ban->reason);
	}
}

static void stats_linkinfo(aClient *sptr, char *servermask)
{
	char *Sformat = ":Name SendQ SendM SendBytes RcveM RcveBytes OpenSince Idle TS";
	char *Lformat = ":%s %u %u %u %u %u %u %u %s";
	dlink_node *node;
	aClient *acptr;
	time_t sincetime;
	char *path;

	send_me_numeric_bufNA(sptr, Sformat, RPL_STATSLINKINFO);
	if (!BadPtr(servermask) && match(servermask, me.name) && !has_wilds(servermask)) {
		if (!MyConnect(sptr)) {
			return;
		}
		if ((acptr = find_person(servermask, NULL)) == NULL) {
			return;
		}

		ASSERT(acptr->localClient != NULL); /* FIXME: is this correct?? */

		sincetime = (acptr->since > timeofday) ? 0 : (timeofday - acptr->since);
		path = get_client_name(acptr, HasMode(sptr, UMODE_RSTAFF) ? SHOW_IP : HIDE_IP);

		send_me_numeric_buf(sptr, Lformat, RPL_STATSLINKINFO, path,
			(int)SBufLength(&acptr->localClient->sendQ), (int)acptr->localClient->sendM,
			(int)acptr->localClient->sendK, (int)acptr->localClient->receiveM,
			(int)acptr->localClient->receiveK, (timeofday - acptr->firsttime),
			sincetime, IsServer(acptr) ? CapTS(acptr) ? "TS" : "Non-TS" : "-");
		return;
	}

	DLINK_FOREACH_DATA(lserver_list.head, node, acptr, aClient) {
		if (GeneralConfig.hide_ulined_servers && IsULine(acptr) && !HasMode(sptr, UMODE_OPER)) {
			continue;
		}

		sincetime = (acptr->since > timeofday) ? 0 : (timeofday - acptr->since);
		path = get_client_name(acptr, HIDE_IP);

		send_me_numeric_buf(sptr, Lformat, RPL_STATSLINKINFO, path,
			(int)SBufLength(&acptr->localClient->sendQ), (int)acptr->localClient->sendM,
			(int)acptr->localClient->sendK, (int)acptr->localClient->receiveM,
			(int)acptr->localClient->receiveK, (timeofday - acptr->firsttime),
			sincetime, IsServer(acptr) ? CapTS(acptr) ? "TS" : "Non-TS" : "-");
	}
}

static void stats_command(aClient *sptr, char *unused)
{
	int i;
	Command *cmd;
	char *token;

	for (i = 0; i < 256; i++) {
		for (cmd = command_table[i]; cmd != NULL; cmd = cmd->next) {
			ASSERT(cmd->msg != NULL);

			if ((token = cmd->msg->tok_str) != NULL) {
				ASSERT(token_table[(unsigned char)*token] == cmd);
			}
			send_me_numeric(sptr, RPL_STATSCOMMANDS, cmd->msg->msg_str, cmd->count,
				cmd->bytes, (token != NULL) ? token : "N/A");
		}
	}
}

static void stats_usercount(aClient *sptr, char *unused)
{
	send_me_numeric(sptr, RPL_STATSCOUNT, "User connects today: ", Count.today);
	send_me_numeric(sptr, RPL_STATSCOUNT, "User connects past week: ", Count.weekly);
	send_me_numeric(sptr, RPL_STATSCOUNT, "User connects past month: ", Count.monthly);
	send_me_numeric(sptr, RPL_STATSCOUNT, "User connects past year: ", Count.yearly);
}

static void stats_oper(aClient *sptr, char *unused)
{
	dlink_node *node;
	ConfigItem_oper *oper;
	int i;

	DLINK_FOREACH_DATA(conf_oper_list.head, node, oper, ConfigItem_oper) {
		if (!HasMode(sptr, UMODE_OPER)) {
			send_me_numeric(sptr, RPL_STATSOLINE, "O", "*", oper->name, oper->class->name);
			continue;
		}

		for (i = 0; i < oper->from.host_count; i++) {
			send_me_numeric(sptr, RPL_STATSOLINE, "O", oper->from.hosts[i], oper->name,
				oper->class->name);
		}
	}
}

static void stats_listener(aClient *sptr, char *unused)
{
	char *Sformat = ":Name Clients TotalConnections SendM SendBytes RcvM RcvBytes Idle Options";
	char *Lformat = ":%s %u %u %u %u %u %u %u (%s)";
	dlink_node *node;
	Listener *l;
	char *flags;

	send_me_numeric_bufNA(sptr, Sformat, RPL_STATSLINKINFO);
	DLINK_FOREACH_DATA(listener_list.head, node, l, Listener) {
		flags = get_listener_flags(l);
		send_me_numeric_buf(sptr, Lformat, RPL_STATSLINKINFO, get_listener_name(l), l->clients,
			l->count, (int)l->sendM, (int)l->sendK, (int)l->receiveM, (int)l->receiveK,
			(timeofday - l->lasttime), (flags != NULL) ? flags : "none");
	}
}

static void stats_ban_nickchan(aClient *sptr, char *unused)
{
	dlink_node *node;
	ConfigItem_ban *ban;

	DLINK_FOREACH_DATA(conf_ban_list.head, node, ban, ConfigItem_ban) {
		if (ban->flags & (BAN_NICKNAME|BAN_CHANNEL)) {
			send_me_numeric(sptr, RPL_STATSKLINE, 'Q', ban->mask,
				BadPtr(ban->reason) ? "no reason specified" : ban->reason);
		}
	}
}

static void stats_rusage(aClient *sptr, char *unused)
{
#ifdef DEBUGMODE
	send_usage(sptr);
#endif
}

static void stats_scache(aClient *sptr, char *unused)
{
	list_scache(sptr);
}

static void stats_throttle(aClient *sptr, char *unused)
{
#ifdef USE_THROTTLE
	throttle_stats(sptr);
#endif
}

static void stats_server(aClient *sptr, char *unused)
{
	serv_stats(sptr);
}

static void stats_uline(aClient *sptr, char *unused)
{
	dlink_node *node;
	char *uline;

	if (GeneralConfig.hide_ulined_servers && !HasMode(sptr, UMODE_OPER)) {
		send_me_numericNA(sptr, ERR_NOPRIVILEGES);
		return;
	}
	DLINK_FOREACH_DATA(conf_uline_list.head, node, uline, char) {
		send_me_numeric(sptr, RPL_STATSULINE, "U", uline);
	}
}

static void stats_uptime(aClient *sptr, char *unused)
{
	time_t now = timeofday - me.since;
	send_me_numeric(sptr, RPL_STATSUPTIME, (now / 86400), (now / 3600) % 24, (now / 60) % 60, now % 60);
	send_me_numeric(sptr, RPL_STATSCONN, Internal.max_con_count, Internal.max_cli_count, Count.cli_restart);
}

static void stats_vhost(aClient *sptr, char *unused)
{
	dlink_node *node;
	ConfigItem_vhost *vhost = NULL;

	DLINK_FOREACH_DATA(conf_vhost_list.head, node, vhost, ConfigItem_vhost) {
		send_me_numeric_buf(sptr, ":V %s %s", RPL_TEXT, vhost->mask, vhost->login);
	}
}

static void stats_class(aClient *sptr, char *unused)
{
	dlink_node *node;
	ConfigItem_class *class = NULL;

	DLINK_FOREACH_DATA(conf_class_list.head, node, class, ConfigItem_class) {
		send_me_numeric(sptr, RPL_STATSYLINE, "Y", class->name, class->ping_time,
			GeneralConfig.auto_connect_freq, class->max_clients, class->sendq_length,
			class->clients);
	}
}

static void stats_memory(aClient *sptr, char *unused)
{
	count_memory(sptr);
}

static void stats_info(aClient *sptr, char *unused)
{
	serv_info(sptr);
}

/*
 * m_stats
 *	parv[0] = sender prefix
 *	parv[1] = stat
 *	parv[2] = server name
 */
int m_stats(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
	static time_t last_used = 0L;
	StatsRequest *sr;
	char flag = (parc > 1) ? *parv[1] : '\0';
	char *para = (parc > 2) ? parv[2] : NULL;
	int need_oper = 0;

	if (use_or_deliver(cptr, sptr, &CMD_STATS, "%s :%s", 2, parc, parv) != HUNTED_ISME) {
		return 0;
	}
	if (parc < 2) {
		send_me_numeric(sptr, ERR_NEEDMOREPARAMS, "STATS");
		return 0;
	}

	for (sr = stats_table; sr->func != NULL && (flag != '\0'); sr++) {
		if ((sr->higher != flag) && (sr->lower != flag)) {
			continue;
		}
		if (!HasMode(sptr, UMODE_OPER)) {
			time_t pace_wait = 0;

			if (sr->options & STATS_PACESIMPLE) {
				pace_wait = FloodConfig.pace_wait_simple;
			}
			else if (sr->options & STATS_PACEINTENSE) {
				pace_wait = FloodConfig.pace_wait_intense;
			}

			if (pace_wait && ((last_used + pace_wait) > timeofday)) {
				send_me_numericNA(sptr, RPL_LOAD2HI);
				return 0;
			}
			else {
				last_used = timeofday;
			}

			if (GeneralConfig.spy_notices && IsPerson(sptr)) {
				sendto_realops_lev(SPY_LEV, "STATS '%c' requested by %s (%s@%s) [%s]",
					flag, sptr->name, sptr->username, MaskedHost(sptr),
					sptr->user->server);
			}
		}
		if (!HasMode(sptr, UMODE_NETADMIN) && (sr->options & STATS_NETADMINONLY)) {
			need_oper = 1;
		}
		else if (!HasMode(sptr, UMODE_ADMIN) && (sr->options & STATS_ADMINONLY)) {
			need_oper = 1;
		}
		else if (!HasMode(sptr, UMODE_OPER) && (sr->options & STATS_OPERONLY)) {
			need_oper = 1;
		}
		if (need_oper) {
			send_me_numericNA(sptr, ERR_NOPRIVILEGES);
			return 0;
		}

		(*sr->func)(sptr, para);
		break;
	}

	send_me_numeric(sptr, RPL_ENDOFSTATS, flag);
	return 0;
}
