/*
 * RageIRCd, an advanced Internet Relay Chat daemon (IRCd)
 * modules/m_svinfo.c: SVINFO protocol module
 *
 * Copyright (C) 2002 the past and present IRCd developers, and others.
 * Refer to doc/authors/ for full credits and copyright details.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA.
 *
 * $Id: m_svinfo.c,v 1.36 2004/03/22 00:18:10 rageircd Exp $
 */

#include "config.h"
#include "struct.h"
#include "common.h"
#include "sys.h"
#include "numeric.h"
#include "msg.h"
#include "channel.h"
#include "h.h"
#include "memory.h"
#include "patchlevel.h"
#include "zlink.h"
#include "modules.h"
#include <time.h>
#include <sys/stat.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>

Module MOD_HEADER(m_svinfo) = {
	"modules/commands/m_svinfo.so",
	"SVINFO protocol",
	5, "$Revision: 1.36 $"
};

int MOD_LOAD(m_svinfo)()
{
	register_command(&MOD_HEADER(m_svinfo), &CMD_SVINFO, m_svinfo);
	return MOD_SUCCESS;
}

int MOD_UNLOAD(m_svinfo)()
{
	return MOD_SUCCESS;
}

/*
 * m_svinfo
 *	parv[0] = sender prefix
 *
 *	if (parc == 2)
 *		parv[1] = ZIP (compression initialisation)
 *
 *	if (parc > 2)
 *		parv[1] = TS_CURRENT
 *		parv[2] = TS_MIN
 *		parv[3] = standalone or connected to non-TS (unused)
 *		parv[4] = UTC time
 *		parv[5] = ircd codename
 *		parv[6] = masking keys 
 */
int m_svinfo(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
	char *inpath = get_client_name(cptr, HIDE_IP);
	time_t delta, tmptime = timeofday, theirtime;

	if (!IsServer(sptr) || !MyConnect(sptr)) {
		return 0;
	}
	if (parc == 2 && !mycmp(parv[1], "ZIP")) {
		SetInputZIP(sptr);
		sptr->serv->zip_in = zip_create_input_session();

		sendto_gnotice("Input from %s is now compressed", inpath);
		if (ServerInfo->hub) {
			sendto_serv_msg_butone(sptr, &me, NULL, &CMD_GNOTICE, ":Input from %s is "
				"now compressed", inpath);
		}
		return ZIP_NEXT_BUFFER;
	}
	if (parc < 7) {
		sendto_gnotice("Link %s dropped, incorrect number of SVINFO parameters", inpath);
		return exit_client(sptr, sptr, sptr, "Incorrect number of SVINFO parameters");
	}
	if (TS_CURRENT < atoi(parv[2]) || atoi(parv[1]) < TS_MIN) {
		sendto_gnotice("Link %s dropped, wrong TS protocol version (%s, %s)",
			inpath, parv[1], parv[2]);
		return exit_client(sptr, sptr, sptr, "Wrong TS protocol version");
	}

	theirtime = atol(parv[4]);
	delta = abs(theirtime - tmptime);

	if (delta > GeneralConfig.ts_delta_max) {
		sendto_gnotice("Link %s dropped, excessive TS delta (our TS=%d, their TS=%d, delta=%d)",
			inpath, tmptime, theirtime, delta);
		return exit_client(sptr, sptr, sptr, "Excessive TS delta");
	}
	if (delta > GeneralConfig.ts_delta_warn) {
		sendto_gnotice("Link %s has a notable TS delta (our TS=%d, their TS=%d, delta=%d)",
			inpath, tmptime, theirtime, delta);
	}

	if (irccmp(CODENAME, parv[5])) {
		sendto_gnotice("Link %s dropped, incompatible server version (ours=%s, theirs=%s)",
			inpath, CODENAME, parv[5]);
		return exit_client(sptr, sptr, sptr, "Incompatible server version (please upgrade)");
	}

	if ((*parv[6] != '0' && (*(parv[6] + 1) != '\0')) && irccmp(Internal.masking_keys, parv[6])) {
		sendto_gnotice("WARNING! Link %s has different masking keys", inpath);
	}
	return 0;
}
