/*
 * RageIRCd, an advanced Internet Relay Chat daemon (IRCd)
 * modules/m_zline.c: ZLINE command module
 *
 * Copyright (C) 2002 the past and present IRCd developers, and others.
 * Refer to doc/authors/ for full credits and copyright details.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA.
 *
 * $Id: m_zline.c,v 1.29 2004/05/13 17:43:52 rageircd Exp $
 */

#include "config.h"
#include "struct.h"
#include "common.h"
#include "sys.h"
#include "numeric.h"
#include "msg.h"
#include "channel.h"
#include "h.h"
#include "memory.h"
#include "modules.h"
#include <time.h>
#include <sys/stat.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>

Module MOD_HEADER(m_zline) = {
	"modules/commands/m_zline.so",
	"/ZLINE command",
	6, "$Revision: 1.29 $"
};

int MOD_LOAD(m_zline)()
{
	register_command(&MOD_HEADER(m_zline), &CMD_ZLINE, m_zline);
	return MOD_SUCCESS;
}

int MOD_UNLOAD(m_zline)()
{
	return MOD_SUCCESS;
}

/*
 * m_zline
 *	parv[0] = sender prefix
 *	parv[1] = ip address
 *	parv[2] = reason
 */
int m_zline(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
	dlink_node *node;
	ConfigItem_ban *ban;
	aClient *acptr;
	char *host, *reason = "No reason", *p;
	int dots = 0;

	if (!OPHasFlag(sptr, OFLAG_ZAP)) {
		send_me_numericNA(sptr, ERR_NOPRIVILEGES);
		return 0;
	}
	if (parc < 2 || BadPtr(parv[1])) {
		send_me_numeric(sptr, ERR_NEEDMOREPARAMS, "ZLINE");
		return 0;
	}
	if (strchr(parv[1], '@') != NULL) {
		send_me_noticeNA(sptr, ":Please specify IP address of client to Z:line");
		return 0;
	}

	host = parv[1];
	for (p = host; *p != '\0'; p++) {
		if (*p == '.') {
			dots++;
		}
		else if (*p == '*') {
			if (dots != 3) {
				send_me_noticeNA(sptr, ":Z:line format error! Please use "
					"xxx.xxx.xxx.xxx or xxx.xxx.xxx.*");
				return 0;
			}
			else {
				*(p + 1) = '\0';
				break;
			}
		}
		else if (!isdigit(*p)) {
			if ((acptr = find_chasing(sptr, parv[1], NULL, MSG_ZLINE)) == NULL) {
				return 0;
			}
			if (acptr->user == NULL || !IsClient(acptr)) {
				return 0;
			}
			if (!MyConnect(acptr)) {
				send_me_noticeNA(sptr, ":You cannot Z:line remote clients");
				return 0;
			}
			host = acptr->hostip;
			break;
		}
	}
	if (parc > 2) {
		reason = parv[2];
		if (strlen(reason) > TOPICLEN) {
			reason[TOPICLEN] = '\0';
		}
	}

	DLINK_FOREACH_DATA(conf_ban_list.head, node, ban, ConfigItem_ban) {
		if (!(ban->flags & BAN_IP)) {
			continue;
		}
		if (!match(ban->mask, host)) {
			send_me_notice(sptr, ":%s is already on the temporary Z:line list as %s",
				host, ban->mask);
			return 0;
		}
	}

	ban = (ConfigItem_ban *)MyMallocX(sizeof(ConfigItem_ban));
	ban->flags |= BAN_IP;
	DupString(ban->mask, host);
	if (!BadPtr(ban->reason)) {
		DupString(ban->reason, reason);
	}
	ban->item.temp = 1;
	ban->kill_on_rehash = 1;
	dlink_add(&conf_ban_list, ban);

	ircdlog(LOG_BAN, "zline for %s added by %s", ban->mask, get_client_name(sptr, SHOW_IP));
	sendto_realops("%s has added a temporary Z:line on *@%s (%s).", sptr->name, ban->mask,
		BadPtr(ban->reason) ? "no reason specified" : ban->reason);
	Internal.do_ban_check = 1;
	return 0;
}
