/*
 * RageIRCd, an advanced Internet Relay Chat daemon (IRCd)
 * src/auth.c: password authentication engine
 *
 * Copyright (C) 2002 the past and present IRCd developers, and others.
 * Refer to doc/authors/ for full credits and copyright details.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA.
 *
 * $Id: auth.c,v 1.30 2004/05/10 12:39:56 rageircd Exp $
 */

#include "struct.h"
#include "conf2.h"
#include "common.h"
#include "sys.h"
#include "numeric.h"
#include "msg.h"
#include "channel.h"
#include "patchlevel.h"
#include "ssl.h"
#include "config.h"
#include "memory.h"
#include "send.h"
#include "h.h"
#include <time.h>
#include <sys/stat.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>

/* FIXME -- this lot could do with some updates and cleanups... */

static const char enc_chars[] = "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789./";

static Auth AuthTypes[] = {
#ifdef USE_OPENSSL
	{ "ripemd/160", AUTH_RIPEMD_160 },
	{ "sha1/160", AUTH_SHA1_160 },
	{ "md5/128", AUTH_MD5_128 },
#endif
	{ "des/56", AUTH_DES_56 },
	{ "md5/56", AUTH_MD5_56 },
	{ "plaintext", AUTH_PLAINTEXT },
	{ NULL, 0 }
};

static short auth_str_to_type(char *);

extern int base64_encode(const char *, size_t, char *, size_t);
extern char *crypt(const char *, const char *);

static short auth_str_to_type(char *str)
{
	Auth *auth;

	for (auth = AuthTypes; auth->string != NULL; auth++) {
		if (!mycmp(auth->string, str)) {
			return auth->type;
		}
	}

	return -1;
}

Auth *auth_parse(ConfigEntry *entry, char *dir)
{
	short type = -1;
	char *string = NULL;
	ConfigEntry *e;
	Auth *auth;

	BLOCK_ENTRIES(entry, dir)

	for (EACH_ENTRY(e, entry)) {
		VARIABLE(e, dir)
		PARAMETER(e, dir)

		if (!mycmp(e->varname, "string")) {
			string = e->vardata;
		}
		else if (!mycmp(e->varname, "type")) {
			if ((type = auth_str_to_type(e->vardata)) == -1) {
				report(0, "%s:%d: %s::type: unknown type %s", e->file->filename,
					e->varlinenum, dir, e->vardata);
				continue;
			}
		}
		else {
			UNKNOWN_VAR(e, dir);
		}
	}

	if (string == NULL) {
		report(0, "%s: %s::string: missing parameter, ignoring section",
			entry->file->filename, dir);
		return NULL;
	}
	if (type == -1) {
		report(0, "%s: %s::type: missing parameter, ignoring section",
			entry->file->filename, dir);
		return NULL;
	}

	auth = (Auth *)MyMalloc(sizeof(Auth));
	DupString(auth->string, string);
	auth->type = type;

	return auth;
}

int auth_check(Auth *auth, char *passwd)
{
#ifdef USE_OPENSSL
	static char buf[512];
#endif
	char salt[12], *encr;
	int i = 0;

	if (auth == NULL || passwd == NULL) {
		return 0;
	}

	switch (auth->type) {
#ifdef USE_OPENSSL
		case AUTH_RIPEMD_160:
			if ((i = base64_encode(RIPEMD160(passwd, strlen(passwd), NULL),
			  RIPEMD160_DIGEST_LENGTH, buf, 512)) > 0) {
				if (!irccmp(buf, auth->string)) {
					return 1;
				}
			}
			break;
		case AUTH_SHA1_160:
			if ((i = base64_encode(SHA1(passwd, strlen(passwd), NULL),
			  SHA_DIGEST_LENGTH, buf, 512)) > 0) {
				if (!irccmp(buf, auth->string)) {
					return 1;
				}
			}
			break;
		case AUTH_MD5_128:
			if ((i = base64_encode(MD5(passwd, strlen(passwd), NULL),
			  MD5_DIGEST_LENGTH, buf, 512)) > 0) {
				if (!irccmp(buf, auth->string)) {
					return 1;
				}
			}
			break;
#endif
		case AUTH_MD5_56:
			salt[0] = '$';
			salt[1] = '1';
			salt[2] = '$';
			for (i = 0; i < 8; i++) {
				if (auth->string[i + 3] == '$') {
					break;
				}
				salt[i + 3] = auth->string[i + 3];
			}
			salt[i + 3] = '\0';
			encr = crypt(passwd, salt);

			if (!irccmp(encr, auth->string)) {
				return 1;
			}
			break;
		case AUTH_DES_56:
			if (strlen(passwd) >= 2) {
				for (i = 0; i < 2; i++) {
					salt[i] = auth->string[i];
				}
				salt[i] = '\0';
				encr = crypt(passwd, salt);
				if (!irccmp(encr, auth->string)) {
					return 1;
				}
			}
			break;
		case AUTH_PLAINTEXT:
			if (!irccmp(passwd, auth->string)) {
				return 1;
			}
			break;
	}
	return 0;
}

void auth_destroy_struct(Auth *auth)
{
	ASSERT(auth != NULL);

	if (auth->string != NULL) {
		MyFree(auth->string);
	}
	MyFree(auth);
	return;
}

void auth_handle_mkpasswd(aClient *sptr, char *type_str, char *passwd)
{
	short type = -1;
#ifdef USE_OPENSSL
	static char buf[512];
#endif
	char salt[12];
	int i = 0;

	if ((type = auth_str_to_type(type_str)) == -1 || (type == AUTH_PLAINTEXT)) {
		send_me_notice(sptr, ":%s is an unknown encryption method.", type_str);
		return;
	}

	srandom(time(NULL));
	switch (type) {
#ifdef USE_OPENSSL
		case AUTH_RIPEMD_160:
			if ((i = base64_encode(RIPEMD160(passwd, strlen(passwd), NULL),
			  RIPEMD160_DIGEST_LENGTH, buf, 512)) > 0) {
				send_me_notice(sptr, ":Encrypted string is: %s (ripemd/160)", buf);
				return;
			}
			break;
		case AUTH_SHA1_160:
			if ((i = base64_encode(SHA(passwd, strlen(passwd), NULL),
			  SHA_DIGEST_LENGTH, buf, 512)) > 0) {
				send_me_notice(sptr, ":Encrypted string is: %s (sha1/160)", buf);
				return;
			}
			break;
		case AUTH_MD5_128:
			if ((i = base64_encode(MD5(passwd, strlen(passwd), NULL),
			  MD5_DIGEST_LENGTH, buf, 512)) > 0) {
				send_me_notice(sptr, ":Encrypted string is: %s (md5/128)", buf);
				return;
			}
			break;
#endif
		case AUTH_DES_56:
			for (i = 0; i < 2; i++) {
				salt[i] = enc_chars[random() % 64];
			}
			salt[i] = '\0';
			send_me_notice(sptr, ":Encrypted string is: %s (des/56)", crypt(passwd, salt));
			break;
		case AUTH_MD5_56:
			salt[0] = '$';
			salt[1] = '1';
			salt[2] = '$';
			for (i = 0; i < 8; i++) {
				salt[i + 3] = enc_chars[random() % 64];
			}
			salt[i + 3] = '\0';
			send_me_notice(sptr, ":Encrypted string is: %s (md5/56)", crypt(passwd, salt));
			break;
		default:
			ircdlog(LOG_ERROR, "auth_handle_mkpasswd() encountered unknown encryption %d", type);
	}
	return;
}
