/*
 * RageIRCd, an advanced Internet Relay Chat daemon (IRCd)
 * src/connauth.c: dns/identd connection authentication
 *
 * Copyright (C) 2002 the past and present IRCd developers, and others.
 * Refer to doc/authors/ for full credits and copyright details.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA.
 *
 * $Id: connauth.c,v 1.50 2004/04/09 01:45:52 rageircd Exp $
 */

#include "memory.h"
#include "struct.h"
#include "common.h"
#include "config.h"
#include "sys.h"
#include "res.h"
#include "numeric.h"
#include "patchlevel.h"
#include "sock.h"
#include "setup.h"
#include "ssl.h"
#include "h.h"
#include "fd.h"
#include "dlink.h"
#include "hook.h"
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/file.h>
#include <sys/ioctl.h>
#include <fcntl.h>

BlockHeap *connauth_heap;
BlockHeap *dnsquery_heap;

dlink_list connauth_list;

typedef struct _header header_t;
struct _header {
	const char *string;
	int length;
};
enum {
	CONNAUTH_DO_DNS,
	CONNAUTH_FIN_DNS,
	CONNAUTH_FAIL_DNS,
	CONNAUTH_DO_ID,
	CONNAUTH_FIN_ID,
	CONNAUTH_FAIL_ID,
	CONNAUTH_HOST_TOOLONG,
	CONNAUTH_NOMATCH
};

static header_t headers[] = {
	{ "NOTICE AUTH :*** Looking up your hostname...\r\n", 46 },
	{ "NOTICE AUTH :*** Found your hostname\r\n", 38 },
	{ "NOTICE AUTH :*** Couldn't lookup your hostname\r\n", 48 },
	{ "NOTICE AUTH :*** Checking identd...\r\n", 37 },
	{ "NOTICE AUTH :*** Received identd response\r\n", 43 },
	{ "NOTICE AUTH :*** No identd response\r\n", 37 },
	{ "NOTICE AUTH :*** Your hostname is too long; using ip address instead\r\n", 70 },
	{ "NOTICE AUTH :*** Your forward and reverse DNS do not match; using ip address instead\r\n", 86 }
};

static void send_header(aClient *cptr, int num)
{
	ASSERT(cptr != NULL);
	ASSERT(cptr->localClient != NULL);

	if (GeneralConfig.show_headers
#ifdef USE_OPENSSL
	  && !(cptr->localClient->listener->flags & LISTEN_SECURE)
#endif
	  ) {
		ircsend(cptr, headers[num].string, (size_t)headers[num].length);
	}
}

static void release_client(aClient *cptr)
{
	add_client_to_list(cptr);
	Debug((DEBUG_DEBUG, "ConnectAuth: releasing client fd %d", cptr->localClient->fd));
	engine_read_packet(cptr->localClient->fd, cptr);
}

static void ident_error(ConnectAuth *auth)
{
	ircstp->is_abad++;

	ASSERT(auth != NULL);

	fd_close(auth->fd);
	auth->fd = -1;
	auth->flag.ident_connect = 0;
	auth->flag.ident_pending = 0;

	send_header(auth->cptr, CONNAUTH_FAIL_ID);

	if (!auth->flag.dns_pending) {
		aClient *cptr = auth->cptr;

		ASSERT(cptr != NULL);
		dlink_del(&connauth_list, auth, NULL);
		free_connauth(auth);
		release_client(cptr);
	}
}

static char *ident_parse(char *id)
{
	int remp = 0, locp = 0;
	char *colon1, *colon2, *colon3, *comma, *remstring = id;

	if ((colon1 = strchr(remstring, ':')) == NULL) {
		return NULL;
	}
	*colon1 = '\0';
	colon1++;

	if ((colon2 = strchr(colon1, ':')) == NULL) {
		return NULL;
	}
	*colon2 = '\0';
	colon2++;

	if ((comma = strchr(remstring, ',')) == NULL) {
		return NULL;
	}
	*comma = '\0';
	comma++;

	if (!(remp = atoi(remstring))) {
		return NULL;
	}
	if (!(locp = atoi(comma))) {
		return NULL;
	}
	if (!strstr(colon1, "USERID")) {
		return NULL;
	}
	if ((colon3 = strchr(colon2, ':')) == NULL) {
		return NULL;
	}
	*colon3++ = '\0';
	return colon3;
}

static void ident_read_reply(int fd, void *data)
{
	ConnectAuth *auth = (ConnectAuth *)data;
	char buf[129], *s = NULL, *t = NULL;
	int len, usercnt;

	ASSERT(auth != NULL);

	if ((len = recv(auth->fd, buf, 128, 0)) > 0) {
		buf[len] = '\0';
		if ((s = ident_parse(buf)) != NULL) {
			t = auth->cptr->username;
			for (usercnt = USERLEN; *s && *s != '@' && usercnt > 0; s++) {
				if (!isspace(*s) && *s != ':') {
					*t++ = *s;
					usercnt--;
				}
			}
			*t = '\0';
		}
	}
	if (len < 0 && engine_ignore_errno(errno)) {
		engine_set_call(fd, FDL_CLIENT, FD_WANT_READ, ident_read_reply, auth, 0);
		return;
	}

	fd_close(auth->fd);
	auth->fd = -1;
	auth->flag.ident_connect = 0;
	auth->flag.ident_pending = 0;

	if (s == NULL) {
		ircstp->is_abad++;
		strcpy(auth->cptr->username, "unknown");
		send_header(auth->cptr, CONNAUTH_FAIL_ID);
	}
	else {
		ircstp->is_asuc++;
		SetGotID(auth->cptr);
		send_header(auth->cptr, CONNAUTH_FIN_ID);
	}

	if (!auth->flag.dns_pending) {
		aClient *cptr = auth->cptr;

		dlink_del(&connauth_list, auth, NULL);
		free_connauth(auth);
		release_client(cptr);
	}
}

static void ident_connect_callback(int fd, int error, void *data)
{
	ConnectAuth *auth = (ConnectAuth *)data;
	struct sockaddr_in us, them;
	char authbuf[32];
	int ulen, tlen, alen;

	ASSERT(auth != NULL);

	if (error != ENGINE_OK) {
		ident_error(auth);
		return;
	}

	tlen = ulen = sizeof(us);
	if (getsockname(auth->cptr->localClient->fd, (struct sockaddr *)&us, &ulen) ||
	  getpeername(auth->cptr->localClient->fd, (struct sockaddr *)&them, &tlen)) {
		ircdlog(LOG_ERROR, "ConnectAuth get{sock,peer}name error for %s",
			get_client_name(auth->cptr, SHOW_IP));
		ident_error(auth);
		return;
	}

	alen = ircsprintf(authbuf, "%u , %u\r\n", (unsigned int)ntohs(them.sin_port),
		(unsigned int)ntohs(us.sin_port));
	if (send(auth->fd, authbuf, alen, 0) == -1) {
		ident_error(auth);
		return;
	}

	auth->flag.ident_connect = 0;
	auth->flag.ident_pending = 1;

	ident_read_reply(auth->fd, auth);
}

static int ident_init(ConnectAuth *auth)
{
	struct sockaddr_in addr;
	int sinlen;
	int fd = -1;

	ASSERT(auth != NULL);

	if ((fd = engine_open(SOCK_STREAM, 0, "identd lookup")) == -1) {
		char *tmp = get_client_name(auth->cptr, SHOW_IP);
		Debug((DEBUG_ERROR, "Unable to create ConnectAuth id socket for %s", tmp));
		ircdlog(LOG_ERROR, "Unable to create ConnectAuth id socket for %s", tmp);
		ircstp->is_abad++;
		return 0;
	}
	if (fd > (HARD_FDLIMIT - 10)) {
		sendto_realops_lev(DEBUG_LEV, "Cannot allocate fd for ConnectAuth on %s",
			get_client_name(auth->cptr, SHOW_IP));
		fd_close(fd);
		return 0;
	}

	send_header(auth->cptr, CONNAUTH_DO_ID);

	sinlen = sizeof(struct sockaddr_in);
	memset(&addr, '\0', sinlen);
	getsockname(auth->cptr->localClient->fd, (struct sockaddr *)&addr, &sinlen);
	addr.sin_port = 0;

	auth->fd = fd;
	auth->flag.ident_connect = 1;

	engine_connect_tcp(fd, auth->cptr->localClient->sockhost, 113, (struct sockaddr *)&addr,
		sinlen, ident_connect_callback, auth, GeneralConfig.connauth_timeout);

	return 1;
}

static void connauth_dns_callback(void *vauth, adns_answer *answer)
{
	ConnectAuth *auth = (ConnectAuth *)vauth;

	ASSERT(auth != NULL);
	auth->flag.dns_pending = 0;
	if (answer != NULL && (answer->status == adns_s_ok)) {
		if (strlen(*answer->rrs.str) <= HOSTLEN) {
			strncpyzt(auth->cptr->host, *answer->rrs.str, HOSTLEN + 1);
			send_header(auth->cptr, CONNAUTH_FIN_DNS);
		}
		else {
			strncpyzt(auth->cptr->host, auth->cptr->localClient->sockhost, HOSTLEN + 1);
			send_header(auth->cptr, CONNAUTH_HOST_TOOLONG);
		}
	}
	else {
		strncpyzt(auth->cptr->host, auth->cptr->localClient->sockhost, HOSTLEN + 1);
		send_header(auth->cptr, CONNAUTH_FAIL_DNS);
	}

	if (answer != NULL) {
		MyFree(answer);
	}

	free_dnsquery(auth->cptr->localClient->dns_query);
	auth->cptr->localClient->dns_query = NULL;

	if (!auth->flag.ident_connect && !auth->flag.ident_pending) {
		aClient *cptr = auth->cptr;

		dlink_del(&connauth_list, auth, NULL);
		free_connauth(auth);
		release_client(cptr);
	}
}

void connauth_init_client(aClient *cptr)
{
	ConnectAuth *auth = NULL;
	HookData hdata;

	ASSERT(cptr != NULL);
	ASSERT(cptr->localClient != NULL);

	hdata.cptr = cptr;
	hook_run(h_pre_connauth, &hdata);

	if (GeneralConfig.resolve_hostnames || GeneralConfig.check_identd) {
		auth = make_connauth();
		auth->cptr = cptr;
		auth->timeout = timeofday + GeneralConfig.connauth_timeout;

		if (GeneralConfig.resolve_hostnames) {
			cptr->localClient->dns_query = make_dnsquery();
			cptr->localClient->dns_query->ptr = auth;
			cptr->localClient->dns_query->callback = connauth_dns_callback;

			send_header(cptr, CONNAUTH_DO_DNS);

			if (!dns_getaddr(&cptr->ip, cptr->localClient->dns_query)) {
				auth->flag.dns_pending = 1;
			}
		}
		if (GeneralConfig.check_identd) {
			ident_init(auth);
		}

		dlink_add(&connauth_list, auth);
		return;
	}

	release_client(cptr);
}

static void connauth_timeout_queries()
{
	dlink_node *node, *next = NULL;
	ConnectAuth *auth;
	aClient *cptr;

	DLINK_FOREACH_SAFE_DATA(connauth_list.head, node, next, auth, ConnectAuth) {
		if (auth->timeout > timeofday) {
			continue;
		}

		if (auth->fd >= 0) {
			fd_close(auth->fd);
			auth->fd = -1;
		}

		cptr = auth->cptr;
		ASSERT(cptr != NULL);

		if (auth->flag.ident_connect || auth->flag.ident_pending) {
			send_header(cptr, CONNAUTH_FAIL_ID);
		}
		if (auth->flag.dns_pending) {
			dns_delete_queries(cptr->localClient->dns_query);
			free_dnsquery(cptr->localClient->dns_query);
			cptr->localClient->dns_query = NULL;
			send_header(cptr, CONNAUTH_FAIL_DNS);
		}

		Debug((DEBUG_NOTICE, "AUTH/DNS timeout for %s", get_client_name(cptr, SHOW_IP)));

		cptr->since = timeofday;
		dlink_del(&connauth_list, NULL, node);
		free_connauth(auth);
		release_client(cptr);
	}
}

void connauth_init()
{
	connauth_heap = BlockHeapCreate((size_t)sizeof(ConnectAuth), CONNAUTH_PREALLOCATE);
	dnsquery_heap = BlockHeapCreate((size_t)sizeof(DNSQuery), CONNAUTH_PREALLOCATE);
	add_event("connauth_timeout_queries", connauth_timeout_queries, NULL, 1, 1);
}

void ident_delete_queries(aClient *cptr)
{
	dlink_node *node, *next = NULL;
	ConnectAuth *auth;

	ASSERT(cptr != NULL);

	DLINK_FOREACH_SAFE_DATA(connauth_list.head, node, next, auth, ConnectAuth) {
		if (auth->cptr != cptr) {
			continue;
		}
		if (auth->fd >= 0) {
			fd_close(auth->fd);
			auth->fd = -1;
		}
		dlink_del(&connauth_list, NULL, node);
		free_connauth(auth);
	}
}
