/*
 * RageIRCd, an advanced Internet Relay Chat daemon (IRCd)
 * src/dlist.c: double-linked list management
 *
 * Copyright (C) 2002 the past and present IRCd developers, and others.
 * Refer to doc/authors/ for full credits and copyright details.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA.
 *
 * $Id: dlink.c,v 1.20 2004/03/22 23:17:14 rageircd Exp $
 */

#include "struct.h"
#include "common.h"
#include "sys.h"
#include "h.h"
#include "numeric.h"
#include "setup.h"
#include "memory.h"
#include "hook.h"
#include "dlink.h"

BlockHeap *dlinknode_heap;

void dlink_init()
{
	dlinknode_heap = BlockHeapCreate((size_t)sizeof(dlink_node), DLINKNODES_PREALLOCATE);
}

static inline void add_node_to_list_head(dlink_list *list, dlink_node *node)
{
	ASSERT(list != NULL);
	ASSERT(node != NULL);

	node->next = list->head;

	if (list->head != NULL) {
		list->head->prev = node;
	}
	else {
		list->tail = node;
	}

	list->head = node;
	list->length++;
}

static inline void add_node_to_list_tail(dlink_list *list, dlink_node *node)
{
	ASSERT(list != NULL);
	ASSERT(node != NULL);

	node->next = NULL;
	node->prev = list->tail;

	if (list->tail != NULL) {
		list->tail->next = node;
	}
	else {
		list->head = node;
	}

	list->tail = node;
	list->length++;
}

static inline void del_node_from_list(dlink_list *list, dlink_node *node)
{
	ASSERT(list != NULL);
	ASSERT(node != NULL);

	if (node->next != NULL) {
		node->next->prev = node->prev;
	}
	else {
		list->tail = node->prev;
	}
	if (node->prev != NULL) {
		node->prev->next = node->next;
	}
	else {
		list->head = node->next;
	}

	node->next = node->prev = NULL;
	list->length--;
}

dlink_node *dlink_find(dlink_list *list, void *data)
{
	dlink_node *node;

	ASSERT(list != NULL);
	ASSERT(data != NULL);

	DLINK_FOREACH(list->head, node) {
		ASSERT(node != NULL);
		ASSERT(node->data != NULL);

		if (node->data == data) {
			return node;
		}
	}

	return NULL;
}

dlink_node *dlink_add_node(dlink_list *list, dlink_node *node, void *data)
{
	ASSERT(list != NULL);
	ASSERT(node != NULL);

	node->data = data;
	add_node_to_list_head(list, node);

	return node;
}

dlink_node *dlink_add(dlink_list *list, void *data)
{
	dlink_node *node = make_dlink_node();
	return dlink_add_node(list, node, data);
}

dlink_node *dlink_add_tail(dlink_list *list, void *data)
{
	dlink_node *node = NULL;

	ASSERT(list != NULL);

	node = make_dlink_node();
	node->data = data;
	add_node_to_list_tail(list, node);

	return node;
}

void dlink_del(dlink_list *list, void *data, dlink_node *node_p)
{
	dlink_node *node = node_p;

	ASSERT(list != NULL);

	if (node == NULL) {
		node = dlink_find(list, data);
	}
	if (node != NULL) {
		del_node_from_list(list, node);
		free_dlink_node(node);
	}
}

dlink_node *dlink_del_nofree(dlink_list *list, void *data, dlink_node *node_p)
{
	dlink_node *node = node_p;

	ASSERT(list != NULL);

	if (node == NULL) {
		node = dlink_find(list, data);
	}
	if (node != NULL) {
		del_node_from_list(list, node);
		return node;
	}

	ircdlog(LOG_ERROR, "dlink_del_nofree(%x,%x,%x) failed", list, data, node_p);
	return NULL;
}

void dlink_array_init(dlink_array *array)
{
	int i;

	ASSERT(array != NULL);

	for (i = 0; i < MAXCONNECTIONS; i++) {
		array->nodes[i].next = NULL;
		array->nodes[i].prev = NULL;
		array->nodes[i].data = NULL;
	}

	array->node_count = 0;
}

void dlink_array_add(dlink_array *array, void *data)
{
	ASSERT(array != NULL);
	ASSERT(data != NULL);

	array->nodes[array->node_count].data = data;
	array->node_count++;
}
