/*
 * RageIRCd, an advanced Internet Relay Chat daemon (IRCd)
 * src/engine_select.c: select socket engine
 *
 * Copyright (C) 2002 the past and present IRCd developers, and others.
 * Refer to doc/authors/ for full credits and copyright details.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA.
 *
 * $Id: engine_select.c,v 1.15 2004/04/10 15:22:48 rageircd Exp $
 */

#include "setup.h"
#include "struct.h"
#include "common.h"
#include "sys.h"
#include "h.h"
#include "fd.h"
#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>

#ifdef USE_SELECT

static fd_set select_read_fds;
static fd_set select_write_fds;

static void select_update_fds(int fd, short event, void (*h)())
{
	fd_entry *fde = fd_get_entry(fd);

	if (event & FD_WANT_READ) {
		if (h != NULL) {
			FD_SET(fd, &select_read_fds);
		}
		else {
			FD_CLR(fd, &select_read_fds);
		}
	}
	if (event & FD_WANT_WRITE) {
		if (h != NULL) {
			FD_SET(fd, &select_write_fds);
		}
		else {
			FD_CLR(fd, &select_write_fds);
		}
	}
	if (h != NULL) {
		fde->flags |= event;
	}
	else {
		fde->flags &= ~event;
	}
}

void engine_init()
{
	FD_ZERO(&select_read_fds);
	FD_ZERO(&select_write_fds);
}

void engine_set_call(int fd, fd_list list, unsigned int flags, void (*h)(), void *d, time_t timeout)
{
	fd_entry *fde = fd_get_entry(fd);
	FD_USED(fd)

	fde->list = list;
	fde->flags |= flags;

	if (flags & FD_WANT_READ) {
		fde->h_read = h;
		fde->d_read = d;
		select_update_fds(fd, FD_WANT_READ, h);
	}
	if (flags & FD_WANT_WRITE) {
		fde->h_write = h;
		fde->d_write = d;
		select_update_fds(fd, FD_WANT_WRITE, h);
	}
	if (timeout) {
		engine_set_timeout(fd, timeout, NULL, NULL);
	}
}

void engine_do_netio(time_t delay)
{
	int cnt, i;
	fd_set read_set, write_set;
	struct timeval tv;
	fd_entry *fde;
	void (*h)();

	memcpy(&read_set, &select_read_fds, sizeof(fd_set));
	memcpy(&write_set, &select_write_fds, sizeof(fd_set));

	for (;;) {
		tv.tv_sec = 0;
		tv.tv_usec = delay;

		if ((cnt = select(fds.highest + 1, &read_set, &write_set, NULL, &tv)) >= 0) {
			break;
		}
		if (!engine_ignore_errno(errno)) {
			cnt = 0;
			break;
		}
	}

	set_time();

	if (!cnt) {
		return;
	}

	for (i = 0; i < fds.highest + 1; i++) {
		fde = fd_get_entry(i);

		if (FD_ISSET(i, &read_set)) {
			select_update_fds(i, FD_WANT_READ, NULL);
			if ((h = fde->h_read) != NULL) {
				fde->h_read  = NULL;
				h(i, fde->d_read);
			}
		}
		if (FD_ISSET(i, &write_set)) {
			select_update_fds(i, FD_WANT_WRITE, NULL);
			if ((h = fde->h_write) != NULL) {
				fde->h_write = NULL;
				h(i, fde->d_write);
			}
		}
	}
}

#endif
