/*
 * RageIRCd, an advanced Internet Relay Chat daemon (IRCd)
 * src/iphash.c: ip hash system
 *
 * Copyright (C) 2002 the past and present IRCd developers, and others.
 * Refer to doc/authors/ for full credits and copyright details.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA.
 *
 * $Id: iphash.c,v 1.15 2004/03/27 02:51:12 rageircd Exp $
 */

#include "struct.h"
#include "common.h"
#include "sys.h"
#include "numeric.h"
#include "h.h"
#include "memory.h"
#include <fcntl.h>

IPEntry *ip_hash_table[IP_HASH_SIZE];
BlockHeap *ipentry_heap;

void init_ip_hash()
{
	ipentry_heap = BlockHeapCreate(sizeof(IPEntry), CLIENTS_PREALLOCATE);
	memset(ip_hash_table, '\0', IP_HASH_SIZE);
}

static int hash_ip(struct in_addr *addr)
{
	unsigned long ip = ntohl(addr->s_addr);
	return (((ip >> 12) + ip) & (IP_HASH_SIZE - 1));
}

IPEntry *find_or_add_ip(struct in_addr *addr)
{
	int hash = hash_ip(addr);
	IPEntry *ipe;

	for (ipe = ip_hash_table[hash]; ipe != NULL; ipe = ipe->next) {
		if (ipe->ip.s_addr == addr->s_addr) {
			return ipe;
		}
	}

	ipe = BlockHeapAlloc(ipentry_heap);
	memcpy(&ipe->ip, addr, sizeof(struct in_addr));
	ipe->count = 0;

	if (ip_hash_table[hash] != NULL) {
		ipe->next = ip_hash_table[hash];
	}
	else {
		ipe->next = NULL;
	}
	ip_hash_table[hash] = ipe;

	return ipe;
}

void remove_one_ip(struct in_addr *addr)
{
	int hash = hash_ip(addr);
	IPEntry *ipe, *last = NULL;

	for (ipe = ip_hash_table[hash]; ipe != NULL; ipe = ipe->next) {
		if (ipe->ip.s_addr != addr->s_addr) {
			last = ipe;
			continue;
		}

		ASSERT(ipe->count > 0);

		if (--ipe->count) {
			return;
		}
		if (last != NULL) {
			last->next = ipe->next;
		}
		else {
			ip_hash_table[hash] = ipe->next;
		}
		BlockHeapFree(ipentry_heap, ipe);
		break;
	}
}

void count_ip_hash_memory(int *ip_entry_cnt, unsigned long *ip_entry_mem)
{
	int hash, cnt = 0;
	IPEntry *ipe;

	for (hash = 0; hash < IP_HASH_SIZE; hash++) {
		for (ipe = ip_hash_table[hash]; ipe != NULL; ipe = ipe->next) {
			cnt++;
		}
	}

	*ip_entry_cnt = cnt;
	*ip_entry_mem = (cnt * sizeof(IPEntry));
}

void list_ip_hash(aClient *cptr)
{
	int hash;
	IPEntry *ipe;

	for (hash = 0; hash < IP_HASH_SIZE; hash++) {
		for (ipe = ip_hash_table[hash]; ipe != NULL; ipe = ipe->next) {
			send_me_notice(cptr, ":%s (%d)", inetntoa((char *)&ipe->ip), ipe->count);
		}
	}
}
