/*
 * RageIRCd, an advanced Internet Relay Chat daemon (IRCd)
 * src/ircd_signal.c: internal signal handling
 *
 * Copyright (C) 2002 the past and present IRCd developers, and others.
 * Refer to doc/authors/ for full credits and copyright details.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA.
 *
 * $Id: ircd_signal.c,v 1.18 2004/04/10 23:40:01 rageircd Exp $
 */

#include "struct.h"
#include "common.h"
#include "sys.h"
#include "numeric.h"
#include "msg.h"
#include "h.h"
#include "setup.h"
#include "hook.h"
#include <signal.h>

#if defined(PROFILING) && (!defined(__GLIBC__) || (__GLIBC__ < 2))
#undef PROFILING
#endif

#ifdef PROFILING
#include SYSGMONH
#define monstartup __monstartup

extern void _start, etext;
static int profiling_state = 1;
static int profiling_newmsg = 0;
static char profiling_msg[512];
#endif

static void dummy_handler()
{
	/* Nothing */
}

static void sigterm_handler()
{
	flush_queued_buffers(me.localClient->fd);
	ircdlog(LOG_DEFAULT, "Server killed by SIGTERM");
	exit(EXIT_SUCCESS);
}

static void sighup_handler()
{
	sendto_ops("Received signal SIGHUP, reloading ircd.conf.");
	ircdlog(LOG_DEFAULT, "Received SIGHUP, reloading ircd.conf.");
	Internal.do_rehash = 1;
}

static void sigint_handler()
{
	static int been_here = 0;

	if (Internal.run_in_terminal) {
		ircdlog(LOG_DEFAULT, "Server exiting on SIGINT");
		exit(0);
	}

	if (!been_here) {
		ircdlog(LOG_DEFAULT, "Server restarting on SIGINT");
		been_here = 1;
		server_reboot();
	}
}

#ifdef PROFILING
static void sigusr1_handler()
{
	char buf[32];

	ircsprintf(buf, "gmon.%ld", timeofday);

	setenv("GMON_OUT_PREFIX", buf, 1);
	_mcleanup();
	monstartup((u_long)&_start, (u_long)&etext);
	setenv("GMON_OUT_PREFIX", "gmon.auto", 1);

	ircsprintf(profiling_msg, "Reset profile, saved past profile data to %s", buf);
	profiling_newmsg = 1;
}

static void sigusr2_handler()
{
	char buf[32];

	if (profiling_state) {
		ircsprintf(buf, "gmon.%ld", timeofday);
		setenv("GMON_OUT_PREFIX", buf, 1);
		_mcleanup();
		ircsprintf(profiling_msg, "Turned profiling OFF, saved profile data to %s", buf);
		profiling_state = 0;
	}
	else {
		monstartup((u_long)&_start, (u_long)&etext);
		setenv("GMON_OUT_PREFIX", "gmon.auto", 1);
		ircsprintf(profiling_msg, "Turned profiling ON");
		profiling_state = 1;
	}
	profiling_newmsg = 1;
}
#endif

void server_die()
{
	sigterm_handler();
}

void init_signal()
{
	struct sigaction act;

	act.sa_handler = SIG_IGN;
	act.sa_flags = 0;

	sigemptyset(&act.sa_mask);
	sigaddset(&act.sa_mask, SIGPIPE);
	sigaddset(&act.sa_mask, SIGALRM);
#ifdef SIGWINCH
	sigaddset(&act.sa_mask, SIGWINCH);
	sigaction(SIGWINCH, &act, NULL);
#endif
	sigaction(SIGPIPE, &act, NULL);
	act.sa_handler = dummy_handler;
	sigaction(SIGALRM, &act, NULL);

	act.sa_handler = sighup_handler;
	sigemptyset(&act.sa_mask);
	sigaddset(&act.sa_mask, SIGHUP);
	sigaction(SIGHUP, &act, NULL);

	act.sa_handler = sigint_handler;
	sigaddset(&act.sa_mask, SIGINT);
	sigaction(SIGINT, &act, NULL);

	act.sa_handler = sigterm_handler;
	sigaddset(&act.sa_mask, SIGTERM);
	sigaction(SIGTERM, &act, NULL);
}

#ifdef PROFILING

static HookEvent *he_check_profiling;

static int check_profiling(HookData *hdata)
{
	if (profiling_newmsg) {
		sendto_realops("PROFILING: %s", profiling_msg);
		profiling_newmsg = 0;
	}
	return 0;
}

void init_profiling_signal()
{
	setenv("GMON_OUT_PREFIX", "gmon.out", 1);
	signal(SIGUSR1, sigusr1_handler);
	signal(SIGUSR2, sigusr2_handler);
}

void init_profiling_event()
{
	he_check_profiling = hook_add_event(h_post_netio, check_profiling);
}

#endif
