/*
 * RageIRCd, an advanced Internet Relay Chat daemon (IRCd)
 * src/ircsprintf.c: ircsprintf functions
 *
 * Copyright (C) 2002 the past and present IRCd developers, and others.
 * Refer to doc/authors/ for full credits and copyright details.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA.
 *
 * $Id: ircsprintf.c,v 1.14 2004/04/01 23:47:56 rageircd Exp $
 */

#include "ircsprintf.h"
#include "memory.h"
#include "common.h"
#include "struct.h"
#include "h.h"

char num[12] = { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 };
char itoa_tab[10] = { '0', '1', '2', '3', '4', '5', '6', '7', '8', '9' };
char xtoa_tab[16] = { '0', '1', '2', '3', '4', '5', '6', '7', '8', '9', 'a', 'b', 'c', 'd', 'e', 'f' };

static inline int irc_sprintf(char *str, const char *pattern, va_list vl)
{
	const char *p = pattern;
	char *buf = str, *s, c;
	va_list ap;
	unsigned long i, u;
	int len = 0;

	VA_COPY(ap, vl);

	for (p = pattern; *p != '\0'; p++) {
		if (*p != '%') {
			buf[len++] = *p;
			continue;
		}

		p++;
		u = 0;

		switch (*p) {
			case 's':
				if ((s = va_arg(ap, char *)) == NULL) {
					break;
				}
				while (*s != '\0') {
					buf[len++] = *s++;
				}
				break;
			case 'c':
				buf[len++] = (char)va_arg(ap, int);
				break;
			case 'u':
				p--;
			case 'l':
				u = (*(p + 1) == 'u') ? 1 : 0;
				if (*(p + 1) == 'u' || *(p + 1) == 'd') {
					p++;
				}
			case 'd':
			case 'i':
				i = va_arg(ap, unsigned long);
				s = &num[11];

				if (!u && (i & 0x80000000)) {
					buf[len++] = '-';
					i = 0x80000000 - (i & ~0x80000000);
				}

				do {
					*--s = itoa_tab[i % 10];
					i /= 10;
				} while (i != 0);

				while (*s != '\0') {
					buf[len++] = *s++;
				}
				break;
			case 'B':
				buf[len++] = '!';
			case 'b':
				i = va_arg(ap, long);
				for (s = base64enc(i); *s != '\0'; s++) {
					buf[len++] = *s;
				}
				break;
			case 'X':
			case 'x':
			case 'n':
				i = va_arg(ap, long);
				s = &num[11];

				do {
					c = xtoa_tab[i % 16];
					if (*p == 'X' && (c >= 'a' || c <= 'f')) {
						c = touppertab[(unsigned char)c];
					}

					*--s = c;
					i /= 16;
				} while (i != 0);

				while (*s != '\0') {
					buf[len++] = *s++;
				}
				break;
			default:
				return vsprintf(str, pattern, vl);
				break;
		}
	}
	buf[len] = '\0';
	return len;
}

static inline int irc_snprintf(char *str, size_t size, const char *pattern, va_list vl)
{
	const char *p = pattern;
	char *buf = str, *s, c;
	va_list ap;
	unsigned long i, u;
	int len = 0;

	VA_COPY(ap, vl);

	for (p = pattern; *p != '\0' && (len < size); p++) {
		if (*p != '%') {
			buf[len++] = *p;
			continue;
		}

		p++;
		u = 0;

		switch (*p) {
			case 's':
				if ((s = va_arg(ap, char *)) == NULL) {
					break;
				}
				while (*s != '\0' && (len < size)) {
					buf[len++] = *s++;
				}
				break;
			case 'c':
				buf[len++] = (char)va_arg(ap, int);
				break;
			case 'u':
				p--;
			case 'l':
				u = (*(p + 1) == 'u') ? 1 : 0;
				if (*(p + 1) == 'u' || *(p + 1) == 'd') {
					p++;
				}
			case 'd':
			case 'i':
				i = va_arg(ap, unsigned long);
				s = &num[11];

				if (!u && (i & 0x80000000)) {
					buf[len++] = '-';
					i = 0x80000000 - (i & ~0x80000000);
				}

				do {
					*--s = itoa_tab[i % 10];
					i /= 10;
				} while (i != 0);

				while (*s != '\0' && (len < size)) {
					buf[len++] = *s++;
				}
				break;
			case 'B':
				buf[len++] = '!';
			case 'b':
				i = va_arg(ap, long);
				for (s = base64enc(i); *s != '\0'; s++) {
					buf[len++] = *s;
				}
				break;
			case 'X':
			case 'x':
			case 'n':
				i = va_arg(ap, long);
				s = &num[11];

				do {
					c = xtoa_tab[i % 16];
					if (*p == 'X' && (c >= 'a' || c <= 'f')) {
						c = touppertab[(unsigned char)c];
					}

					*--s = c;
					i /= 16;
				} while (i != 0);

				while (*s != '\0' && (len < size)) {
					buf[len++] = *s++;
				}
				break;
			default:
				return vsprintf(str, pattern, vl);
				break;
		}
	}
	buf[len] = '\0';
	return len;
}

int ircsprintf(char *str, const char *pattern, ...)
{
	int len;
	va_list vl;

	va_start(vl, pattern);
	len = irc_sprintf(str, pattern, vl);
	va_end(vl);

	return len;
}

int ircsnprintf(char *str, size_t size, const char *pattern, ...)
{
	int len;
	va_list vl;

	va_start(vl, pattern);
	len = irc_snprintf(str, size, pattern, vl);
	va_end(vl);

	return len;
}

int ircvsprintf(char *str, const char *pattern, va_list vl)
{
	return irc_sprintf(str, pattern, vl);
}

int ircvsnprintf(char *str, size_t size, const char *pattern, va_list vl)
{
	return irc_snprintf(str, size, pattern, vl);
}

void make_numeric_string(char *string, int value)
{
	char *s = string + 3;
	*s-- = '\0';
	*s-- = itoa_tab[value % 10];
	value /= 10;
	*s-- = itoa_tab[value % 10];
	value /= 10;
	*s = itoa_tab[value % 10];
}
