/*
 * RageIRCd, an advanced Internet Relay Chat daemon (IRCd)
 * src/s_conf2.c: server config parsing
 *
 * Copyright (C) 2002 the past and present IRCd developers, and others.
 * Refer to doc/authors/ for full credits and copyright details.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA.
 *
 * $Id: s_conf2.c,v 1.233 2004/05/13 22:26:56 rageircd Exp $
 */

#include "struct.h"
#include "common.h"
#include "h.h"
#include "sys.h"
#include "numeric.h"
#include "msg.h"
#include "channel.h"
#include "patchlevel.h"
#include "ssl.h"
#include "conf2.h"
#include "config.h"
#include "memory.h"
#include "res.h"
#include "hook.h"
#include "dlink.h"
#include "fd.h"
#include "xmode.h"
#include "modules.h"
#include <time.h>
#include <signal.h>
#include <sys/stat.h>
#include <stdlib.h>
#include <string.h>
#include <utmp.h>
#include <fcntl.h>

#undef CDEBUG

ConfigItem_network NetworkConfig, tmpNetworkConfig;
ConfigItem_masking MaskingConfig, tmpMaskingConfig;
ConfigItem_flood FloodConfig, tmpFloodConfig;
ConfigItem_general GeneralConfig, tmpGeneralConfig;

ConfigItem_servinfo *ServerInfo = NULL;
ConfigItem_admin *AdminInfo = NULL;
ConfigItem_class *DefaultClass = NULL;

dlink_list conf_class_list;
dlink_list conf_allow_list;
dlink_list conf_oper_list;
dlink_list conf_uline_list;
dlink_list conf_link_list;
dlink_list conf_listen_list;
dlink_list conf_ban_list;
dlink_list conf_except_list;
dlink_list conf_vhost_list;
dlink_list conf_include_list;

ConfigDirective conf_directives[] = {
	{ "servinfo",	parse_servinfo,	test_servinfo,	free_servinfo },
	{ "admin",	parse_admin,	NULL,		NULL },
	{ "class",	parse_class,	test_class,	free_class },
	{ "allow",	parse_allow,	test_allow,	free_allow },
	{ "oper",	parse_oper,	test_oper,	free_oper },
	{ "uline",	parse_uline,	NULL,		NULL },
	{ "link",	parse_link,	test_link,	free_link },
	{ "listen",	parse_listen,	NULL,		NULL },
	{ "ban",	parse_ban,	test_ban,	free_ban },
	{ "except",	parse_except,	NULL,		NULL },
	{ "vhost",	parse_vhost,	test_vhost,	free_vhost },
	{ "include",	parse_include,	NULL,		NULL },
#ifndef STATIC_MODULES
	{ "modules",	parse_modules,	NULL,		NULL },
#endif
	{ "network",	parse_network,	test_network,	free_network },
	{ "masking",	parse_masking,	test_masking,	free_masking },
	{ "flood",	parse_flood,	NULL,		NULL },
	{ "general",	parse_general,	test_general,	free_general },
	{ NULL,		NULL,		NULL,		NULL },
};

ConfigOption servinfo_logs[] = {
	{ "default",		LOG_DEFAULT },
	{ "error",		LOG_ERROR },
	{ "kill",		LOG_KILL },
	{ "ban",		LOG_BAN },
	{ "client",		LOG_CLIENT },
	{ "server",		LOG_SERVER },
	{ "oper",		LOG_OPER },
	{ "override",		LOG_OVERRIDE },
	{ NULL, 0 }
};

ConfigOption oper_can_perform[] = {
	{ "die",		OFLAG_DIE },
	{ "restart",		OFLAG_RESTART },
	{ "rehash",		OFLAG_REHASH },
	{ "kline",		OFLAG_KLINE },
	{ "unkline",		OFLAG_UNKLINE },
	{ "zap",		OFLAG_ZAP },
	{ "local_routing",	OFLAG_LROUTE },
	{ "global_routing",	OFLAG_GROUTE },
	{ "local_kills",	OFLAG_LKILL },
	{ "global_kills",	OFLAG_GKILL },
	{ "local_notices",	OFLAG_LNOTICE },
	{ "global_notices",	OFLAG_GNOTICE },
	{ NULL, 0 }
};

ConfigOption oper_can_see[] = {
	{ "routing",		OFLAG_RSTAFF },
	{ "floods",		OFLAG_FNOTICE },
	{ "local_cliconn",	OFLAG_LCLICONN },
	{ "global_cliconn",	OFLAG_GCLICONN },
	{ NULL, 0 }
};

ConfigOption ban_types[] = {
	{ "userhost",		BAN_USERHOST },
	{ "ip",			BAN_IP },
	{ "nickname",		BAN_NICKNAME },
	{ "gcos",		BAN_GCOS },
	{ "server",		BAN_SERVER },
	{ "file",		BAN_FILE },
	{ "channel",		BAN_CHANNEL },
	{ NULL, 0 }
};

static void conf_add_from(ConfigFrom *from, char *str)
{
	char *s, *userhost;

	if (from->host_count >= MAXCONFHOSTS) {
		return;
	}

	for (s = str; *s != '\0'; s++) {
		if (*s == '@') {
			break;
		}
	}
	if (*s == '\0') {
		userhost = (char *)MyMalloc(strlen(str) + 3);
		ircsprintf(userhost, "*@%s", str);
	}
	else {
		DupString(userhost, str);
	}

	if (conf_check_from(from, userhost)) {
		MyFree(userhost);
	}
	else {
		from->hosts[from->host_count++] = userhost;
	}
}

int conf_check_from(ConfigFrom *from, char *str)
{
	int i;

	for (i = 0; i < from->host_count; i++) {
		if (!match(from->hosts[i], str)) {
			return 1;
		}
	}

	return 0;
}

static void conf_clear_from(ConfigFrom *from)
{
	while (from->host_count) {
		MyFree(from->hosts[from->host_count - 1]);
		from->host_count--;
	}
}

static void add_unknown(ConfigEntry *entry, void *item, short type)
{
	ConfigUnknown *unknown = (ConfigUnknown *)MyMalloc(sizeof(ConfigUnknown));

	unknown->item = item;
	unknown->entry = entry;
	unknown->type = type;

	dlink_add_tail(&conf_unknown_list, unknown);
}

static unsigned int parse_token_list(char *token_list, ConfigOption *option_table)
{
	char *s, *p;
	unsigned int flags = 0;
	ConfigOption *opt;

	for (s = strtoken(&p, token_list, ","); s != NULL; s = strtoken(&p, NULL, ",")) {
		while (isspace(*s)) {
			s++;
		}
		if (*s == '\0') {
			continue;
		}
		for (opt = option_table; opt->token != NULL; opt++) {
			if (!mycmp(opt->token, s)) {
				flags |= opt->flag;
				break;
			}
		}
	}

	return flags;
}

void *parse_servinfo(ConfigEntry *entry)
{
	ConfigEntry *e;

	BLOCK_ENTRIES(entry, "servinfo")

	if (ServerInfo == NULL) {
		ServerInfo = (ConfigItem_servinfo *)MyMalloc(sizeof(ConfigItem_servinfo));
	}

	for (EACH_ENTRY(e, entry)) {
		VARIABLE(e, "servinfo")
		PARAMETER(e, "servinfo")

		if (!mycmp(e->varname, "name")) {
			cDupString(ServerInfo->name, e->vardata);
		}
		else if (!mycmp(e->varname, "description")) {
			cDupString(ServerInfo->desc, e->vardata);
		}
		else if (!mycmp(e->varname, "default_bind_ip")) {
			cDupString(ServerInfo->default_bind_ip, e->vardata);
		}
		else if (!mycmp(e->varname, "identity")) {
			ServerInfo->identity = atoi(e->vardata);
		}
		else if (!mycmp(e->varname, "hub")) {
			ServerInfo->hub = get_option(e->vardata, 1);
		}
		else if (!mycmp(e->varname, "max_clients")) {
			ServerInfo->max_clients = atoi(e->vardata);
		}
		else if (!mycmp(e->varname, "kline_address")) {
			cDupString(ServerInfo->kline_address, e->vardata);
		}
		else if (!mycmp(e->varname, "logs")) {
			ServerInfo->logs |= parse_token_list(e->vardata, servinfo_logs);
		}
#ifdef USE_OPENSSL
		else if (!mycmp(e->varname, "ssl_certificate")) {
			cDupString(ServerInfo->ssl_certificate, e->vardata);
		}
		else if (!mycmp(e->varname, "ssl_private_key")) {
			cDupString(ServerInfo->ssl_private_key, e->vardata);
		}
#endif
		else if (!mycmp(e->varname, "resolver_conf_file")) {
			cDupString(ServerInfo->resolver_conf_file, e->vardata);
		}
		else {
			add_unknown(e, ServerInfo, CONF_SERVINFO);
		}
	}

	return (void *)ServerInfo;
}

int test_servinfo(ConfigFile *file, void *item)
{
	ConfigItem_servinfo *servinfo = (ConfigItem_servinfo *)item;
	int retval = CONF_SUCCESS;

	if (BadPtr(servinfo->name)) {
		report(1, "%s: servinfo::name: missing parameter", file->filename);
		retval = CONF_FAILURE;
	}
	else if (!valid_server_name(servinfo->name)) {
		report(1, "%s: servinfo::name: invalid server name", file->filename);
		retval = CONF_FAILURE;
	}
	if (BadPtr(servinfo->desc)) {
		cDupString(servinfo->desc, "IRCers United!");
	}
	if (!BadPtr(servinfo->default_bind_ip) && !valid_ip_addr(servinfo->default_bind_ip, 0)) {
		cMyFree(servinfo->default_bind_ip);
		report(0, "%s: servinfo::default_bind_ip: invalid ip address, using INADDR_ANY",
			file->filename);
	}
	if (servinfo->identity < 1 || servinfo->identity > 4096) {
		if (servinfo->identity < 0) {
			report(0, "%s: servinfo::identity: invalid identity, using 4095", file->filename);
		}
		servinfo->identity = 0xFFF;
	}
	if (servinfo->max_clients < 10) {
		servinfo->max_clients = DEFAULT_MAX_CLIENTS;
	}
	if (BadPtr(servinfo->kline_address)) {
		cDupString(servinfo->kline_address, DEFAULT_KLINE_ADDRESS);
	}
	else if (!valid_email_addr(servinfo->kline_address, 1)) {
		cDupString(servinfo->kline_address, DEFAULT_KLINE_ADDRESS);
		report(0, "%s: servinfo::kline_address: invalid e-mail address/url, using %s",
			file->filename, servinfo->kline_address);
	}
#ifdef USE_OPENSSL
	if (BadPtr(servinfo->ssl_certificate)) {
		report(1, "%s: ERROR: servinfo::ssl_certificate: missing parameter", file->filename);
		retval = CONF_FAILURE;
	}
	if (BadPtr(servinfo->ssl_private_key)) {
		report(1, "%s: ERROR: servinfo::ssl_private_key: missing parameter", file->filename);
		retval = CONF_FAILURE;
	}
#endif

	return retval;
}

void free_servinfo(void *item)
{
	ConfigItem_servinfo *servinfo = (ConfigItem_servinfo *)item;

	cMyFree(servinfo->name);
	cMyFree(servinfo->desc);
	cMyFree(servinfo->default_bind_ip);
	cMyFree(servinfo->kline_address);
#ifdef USE_OPENSSL
	cMyFree(servinfo->ssl_certificate);
	cMyFree(servinfo->ssl_private_key);
#endif
	cMyFree(servinfo->resolver_conf_file);
	cMyFree(servinfo);
}

void *parse_admin(ConfigEntry *entry)
{
	ConfigEntry *e;

	BLOCK_ENTRIES(entry, "admin")

	if (AdminInfo == NULL) {
		AdminInfo = (ConfigItem_admin *)MyMalloc(sizeof(ConfigItem_admin));
	}

	for (EACH_ENTRY(e, entry)) {
		VARIABLE(e, "admin")
		PARAMETER(e, "admin")

		if (!mycmp(e->varname, "name")) {
			cDupString(AdminInfo->name, e->vardata);
		}
		else if (!mycmp(e->varname, "description")) {
			cDupString(AdminInfo->desc, e->vardata);
		}
		else if (!mycmp(e->varname, "email")) {
			cDupString(AdminInfo->email, e->vardata);
		}
		else {
			add_unknown(e, AdminInfo, CONF_ADMIN);
		}
	}

	return (void *)AdminInfo;
}

void *parse_class(ConfigEntry *entry)
{
	ConfigEntry *e;
	ConfigItem_class *class;

	BLOCK_NAME(entry, "class", "class name")
	BLOCK_ENTRIES(entry, "class")

	if ((class = find_class(entry->vardata)) == NULL) {
		class = (ConfigItem_class *)MyMalloc(sizeof(ConfigItem_class));
		cDupString(class->name, entry->vardata);
		dlink_add(&conf_class_list, class);
	}
	else {
		class->item.temp = 0;
		class->flags = 0;
	}

	for (EACH_ENTRY(e, entry)) {
		VARIABLE(e, "class")
		PARAMETER(e, "class")

		if (!mycmp(e->varname, "ping_time")) {
			class->ping_time = atime(e->vardata);
		}
		else if (!mycmp(e->varname, "max_clients")) {
			class->max_clients = atoi(e->vardata);
		}
		else if (!mycmp(e->varname, "sendq_length")) {
			class->sendq_length = get_size(e->vardata);
		}
		else if (!mycmp(e->varname, "ignore_identd")) {
			class->flags |= get_option(e->vardata, CLASS_IGNOREIDENTD);
		}
		else if (!mycmp(e->varname, "require_identd")) {
			class->flags |= get_option(e->vardata, CLASS_REQUIREIDENTD);
		}
		else if (!mycmp(e->varname, "no_tilde")) {
			class->flags |= get_option(e->vardata, CLASS_NOTILDE);
		}
		else {
			add_unknown(e, class, CONF_CLASS);
		}
	}

	return (void *)class;
}

int test_class(ConfigFile *file, void *item)
{
	ConfigItem_class *class = (ConfigItem_class *)item;

	if (BadPtr(class->name)) {
		report(0, "%s: class: missing name, ignoring block", file->filename);
		return CONF_FAILURE;
	}
	if (class->ping_time < 30) {
		class->ping_time = DEFAULT_PINGTIME;
	}
	if (class->max_clients < 1) {
		class->max_clients = DEFAULT_MAX_CLIENTS;
	}
	if (class->sendq_length < 1) {
		class->sendq_length = MAXSENDQLENGTH;
	}

	return CONF_SUCCESS;
}

void free_class(void *item)
{
	ConfigItem_class *class = (ConfigItem_class *)item;
	cMyFree(class->name);
	cMyFree(class);
}

void *parse_allow(ConfigEntry *entry)
{
	ConfigEntry *e;
	ConfigItem_allow *allow;

	BLOCK_ENTRIES(entry, "allow")

	allow = (ConfigItem_allow *)MyMalloc(sizeof(ConfigItem_allow));
	for (EACH_ENTRY(e, entry)) {
		VARIABLE(e, "allow")
		if (!mycmp(e->varname, "auth")) {
			allow->auth = auth_parse(e, "allow::auth");
			continue;
		}

		PARAMETER(e, "allow")
		if (!mycmp(e->varname, "hostname")) {
			cDupString(allow->hostname, e->vardata);
			allow->hostname_at = (in_str(allow->hostname, '@')) ? 1 : 0;
		}
		else if (!mycmp(e->varname, "ip")) {
			cDupString(allow->ipaddr, e->vardata);
			allow->ipaddr_at = (in_str(allow->ipaddr, '@')) ? 1 : 0;
		}
		else if (!mycmp(e->varname, "port")) {
			allow->port = get_port(e->vardata);
		}
		else if (!mycmp(e->varname, "max_per_ip")) {
			allow->max_per_ip = atoi(e->vardata);
		}
		else if (!mycmp(e->varname, "class")) {
			if ((allow->class = find_class(e->vardata)) == NULL) {
				allow->class = DefaultClass;
			}
			allow->class->item.refcount++;
		}
		else if (!mycmp(e->varname, "redir_serv")) {
			cDupString(allow->redir_serv, e->vardata);
		}
		else if (!mycmp(e->varname, "redir_port")) {
			allow->redir_port = get_port(e->vardata);
		}
		else {
			add_unknown(e, allow, CONF_ALLOW);
		}
	}

	dlink_add(&conf_allow_list, allow);
	return (void *)allow;
}

int test_allow(ConfigFile *file, void *item)
{
	ConfigItem_allow *allow = (ConfigItem_allow *)item;

	if (BadPtr(allow->hostname)) {
		cDupString(allow->hostname, "*");
	}
	if (BadPtr(allow->ipaddr)) {
		cDupString(allow->ipaddr, "*");
	}
	else if (!valid_ip_addr(allow->ipaddr, 1)) {
		cDupString(allow->ipaddr, "*");
		report(0, "%s: allow::ip: invalid ip address, using *", file->filename);
	}
	if (allow->class == NULL) {
		report(0, "%s: allow::class: missing parameter, using default", file->filename);
		allow->class = DefaultClass;
		allow->class->item.refcount++;
	}
	if (BadPtr(allow->redir_serv)) {
		allow->redir_port = 0;
	}
	else if (allow->redir_port == -1) {
		cMyFree(allow->redir_serv);
	}

	return CONF_SUCCESS;
}

void free_allow(void *item)
{
	ConfigItem_allow *allow = (ConfigItem_allow *)item;

	allow->class->item.refcount--;

	if (allow->auth != NULL) {
		auth_destroy_struct(allow->auth);
	}

	cMyFree(allow->hostname);
	cMyFree(allow->ipaddr);
	cMyFree(allow->redir_serv);
	cMyFree(allow);
}

void *parse_oper(ConfigEntry *entry)
{
	ConfigEntry *e;
	ConfigItem_oper *oper;

	BLOCK_NAME(entry, "oper", "oper name")
	BLOCK_ENTRIES(entry, "oper")

	if ((oper = find_oper(entry->vardata)) == NULL) {
		oper = (ConfigItem_oper *)MyMalloc(sizeof(ConfigItem_oper));
		cDupString(oper->name, entry->vardata);
		dlink_add(&conf_oper_list, oper);
	}
	else {
		oper->flags = 0;
		oper->item.temp = 0;
	}

	for (EACH_ENTRY(e, entry)) {
		VARIABLE(e, "oper")
		if (!mycmp(e->varname, "auth")) {
			if (oper->auth != NULL) {
				auth_destroy_struct(oper->auth);
			}
			oper->auth = auth_parse(e, "oper::auth");
			continue;
		}

		PARAMETER(e, "oper")
		if (!mycmp(e->varname, "from")) {
			conf_add_from(&oper->from, e->vardata);
		}
		else if (!mycmp(e->varname, "class")) {
			if ((oper->class = find_class(e->vardata)) == NULL) {
				oper->class = DefaultClass;
			}
			oper->class->item.refcount++;
		}
		else if (!mycmp(e->varname, "join_on_oper")) {
			cDupString(oper->join_on_oper, e->vardata);
		}
		else if (!mycmp(e->varname, "show_oper_motd")) {
			oper->flags |= get_option(e->vardata, OFLAG_OPERMOTD);
		}
		else if (!mycmp(e->varname, "no_recvq_throttle")) {
			oper->flags |= get_option(e->vardata, OFLAG_NORECVQTHROTTLE);
		}
		else if (!mycmp(e->varname, "locops")) {
			oper->flags |= get_option(e->vardata, OFLAG_LOCOPS);
		}
		else if (!mycmp(e->varname, "globops")) {
			oper->flags |= get_option(e->vardata, OFLAG_GLOBOPS);
		}
		else if (!mycmp(e->varname, "wallops")) {
			oper->flags |= get_option(e->vardata, OFLAG_WALLOPS);
		}
		else if (!mycmp(e->varname, "netadmin")) {
			oper->flags |= get_option(e->vardata, OFLAG_NETADMIN);
		}
		else if (!mycmp(e->varname, "admin")) {
			oper->flags |= get_option(e->vardata, OFLAG_ADMIN);
		}
		else if (!mycmp(e->varname, "sadmin")) {
			oper->flags |= get_option(e->vardata, OFLAG_SADMIN);
		}
		else if (!mycmp(e->varname, "can_perform")) {
			oper->flags |= parse_token_list(e->vardata, oper_can_perform);
		}
		else if (!mycmp(e->varname, "can_see")) {
			oper->flags |= parse_token_list(e->vardata, oper_can_see);
		}
		else {
			add_unknown(e, oper, CONF_OPER);
		}
	}

	return (void *)oper;
}

int test_oper(ConfigFile *file, void *item)
{
	ConfigItem_oper *oper = (ConfigItem_oper *)item;

	if (BadPtr(oper->name)) {
		report(0, "%s: oper: missing oper name, ignoring block", file->filename);
		return CONF_FAILURE;
	}
	if (oper->auth == NULL) {
		report(0, "%s: oper::auth: missing section, ignoring block", file->filename);
		return CONF_FAILURE;
	}
	if (!oper->from.host_count) {
		report(0, "%s: oper::from: missing parameter, ignoring block", file->filename);
		return CONF_FAILURE;
	}
	if (oper->class == NULL) {
		report(0, "%s: oper::class: missing parameter, using default", file->filename);
		oper->class = DefaultClass;
		oper->class->item.refcount++;
	}

	return CONF_SUCCESS;
}

void free_oper(void *item)
{
	ConfigItem_oper *oper = (ConfigItem_oper *)item;

	oper->class->item.refcount--;

	conf_clear_from(&oper->from);
	if (oper->auth != NULL) {
		auth_destroy_struct(oper->auth);
	}

	cMyFree(oper->name);
	cMyFree(oper->join_on_oper);
	cMyFree(oper);
}

void *parse_uline(ConfigEntry *entry)
{
	char *server;

	BLOCK_NAME(entry, "uline", "server name")

	if ((server = find_uline(entry->vardata)) == NULL) {
		cDupString(server, entry->vardata);
		dlink_add(&conf_uline_list, server);
	}

	return NULL;
}

void *parse_link(ConfigEntry *entry)
{
	ConfigEntry *e;
	ConfigItem_link *linkp;

	BLOCK_NAME(entry, "link", "server name")
	BLOCK_ENTRIES(entry, "link")

	if ((linkp = find_link(entry->vardata, "*")) == NULL) {
		linkp = (ConfigItem_link *)MyMalloc(sizeof(ConfigItem_link));
		cDupString(linkp->servername, entry->vardata);
		dlink_add(&conf_link_list, linkp);
	}
	else {
		linkp->flags = 0;
		linkp->item.temp = 0;
	}

	for (EACH_ENTRY(e, entry)) {
		VARIABLE(e, "link")
		if (!mycmp(e->varname, "auth")) {
			if (linkp->auth != NULL) {
				auth_destroy_struct(linkp->auth);
			}
			linkp->auth = auth_parse(e, "link::auth");
			continue;
		}

		PARAMETER(e, "link")
		if (!mycmp(e->varname, "host")) {
			cDupString(linkp->host, e->vardata);
		}
		else if (!mycmp(e->varname, "bind_ip")) {
			cDupString(linkp->bind_ip, e->vardata);
		}
		else if (!mycmp(e->varname, "port")) {
			linkp->port = get_port(e->vardata);
		}
		else if (!mycmp(e->varname, "class")) {
			if ((linkp->class = find_class(e->vardata)) == NULL) {
				linkp->class = DefaultClass;
			}
			linkp->class->item.refcount++;
		}
		else if (!mycmp(e->varname, "auto_connect")) {
			linkp->flags |= get_option(e->vardata, LINK_AUTOCONNECT);
		}
		else if (!mycmp(e->varname, "compressed")) {
			linkp->flags |= get_option(e->vardata, LINK_COMPRESS);
		}
#ifdef USE_OPENSSL
		else if (!mycmp(e->varname, "encrypted")) {
			linkp->flags |= get_option(e->vardata, LINK_SECURE);
		}
#endif
		else {
			add_unknown(e, linkp, CONF_LINK);
		}
	}

	return (void *)linkp;
}

int test_link(ConfigFile *file, void *item)
{
	ConfigItem_link *linkp = (ConfigItem_link *)item;

	if (BadPtr(linkp->servername)) {
		report(0, "%s: link: missing server name, ignoring block", file->filename);
		return CONF_FAILURE;
	}
	if (linkp->auth == NULL) {
		report(0, "%s: link::auth: missing section, ignoring block", file->filename);
		return CONF_FAILURE;
	}
	if (BadPtr(linkp->host)) {
		cDupString(linkp->host, "*");
	}
	if (!BadPtr(linkp->bind_ip) && !valid_ip_addr(linkp->bind_ip, 0)) {
		cMyFree(linkp->bind_ip);
		report(0, "%s: link::bind_ip: invalid ip address, using default_bind_ip",
			file->filename);
	}
	if (linkp->port <= 0) {
		report(0, "%s: link::port: %s, using %d", file->filename,
			(linkp->port == -1) ? "invalid port" : "missing parameter",
			DEFAULT_PORTNUM);
		linkp->port = DEFAULT_PORTNUM;
	}
	if (linkp->class == NULL) {
		report(0, "%s: link::class: missing parameter, using default", file->filename);
		linkp->class = DefaultClass;
		linkp->class->item.refcount++;
	}

	return CONF_SUCCESS;
}

void free_link(void *item)
{
	ConfigItem_link *linkp = (ConfigItem_link *)item;

	linkp->class->item.refcount--;

	if (linkp->auth != NULL) {
		auth_destroy_struct(linkp->auth);
	}

	cMyFree(linkp->servername);
	cMyFree(linkp->host);
	cMyFree(linkp->bind_ip);
	cMyFree(linkp);
}

void *parse_listen(ConfigEntry *entry)
{
	ConfigEntry *e, *ee;
	ConfigItem_listen *listenp;
	char addr[HOSTIPLEN + 1], portbuf[16];
	int first, last, i;

	BLOCK_NAME(entry, "listen", "ip address")
	BLOCK_ENTRIES(entry, "listen")

	if (*entry->vardata != '*') {
		strncpyzt(addr, entry->vardata, HOSTIPLEN + 1);
	}
	else {
		*addr = '*';
		*(addr + 1) = '\0';
	}

	for (EACH_ENTRY(e, entry)) {
		VARIABLE(e, "listen")
		PARAMETER(e, "listen")

		first = last = -1;
		if (!mycmp(e->varname, "port")) {
			if (!(first = get_port(e->vardata))) {
				report(0, "%s:%d: listen: invalid port, ignoring entry", e->file->filename,
					e->varlinenum);
				continue;
			}
			last = first;
		}
		else if (!mycmp(e->varname, "range")) {
			get_port_range(e->vardata, &first, &last);
			if (!first || !last) {
				report(0, "%s:%d: listen: invalid range, ignoring entry", e->file->filename,
					e->varlinenum);
				continue;
			}
		}
		else {
			UNKNOWN_VAR(e, "listen");
			continue;
		}

		if (first == -1 || (last == -1)) {
			continue;
		}

		for (i = first; i <= last; i++) {
			if ((listenp = find_listen(addr, i)) == NULL) {
				listenp = (ConfigItem_listen *)MyMalloc(sizeof(ConfigItem_listen));
				cDupString(listenp->ipaddr, addr);
				listenp->port = i;
				dlink_add(&conf_listen_list, listenp);
			}
			else {
				listenp->flags = 0;
				listenp->item.temp = 0;
			}

			if (e->entries == NULL) {
				continue;
			}

			ircsprintf(portbuf, "listen::%d", i);
			for (EACH_ENTRY(ee, e)) {
				VARIABLE(ee, portbuf)
				PARAMETER(ee, portbuf)

				if (!mycmp(ee->varname, "client_only")) {
					listenp->flags |= get_option(ee->vardata, LISTEN_CLIENTONLY);
				}
				else if (!mycmp(ee->varname, "server_only")) {
					listenp->flags |= get_option(ee->vardata, LISTEN_SERVERONLY);
				}
#ifdef USE_OPENSSL
				else if (!mycmp(ee->varname, "secure")) {
					listenp->flags |= get_option(ee->vardata, LISTEN_SECURE);
				}
#endif
				else if (!mycmp(ee->varname, "confroom_java")) {
					listenp->flags |= get_option(ee->vardata, LISTEN_CRJAVA);
				}
				else {
					UNKNOWN_VAR(ee, portbuf);
				}
			}
		}
	}

	return NULL;
}

void *parse_ban(ConfigEntry *entry)
{
	ConfigEntry *e;
	ConfigItem_ban *ban;
	ConfigOption *opt;

	BLOCK_NAME(entry, "ban", "ban type")
	BLOCK_ENTRIES(entry, "ban")

	if ((opt = get_bantype(entry->vardata)) == NULL) {
		report(0, "%s:%d: ban: unknown parameter %s, ignoring block", entry->file->filename,
			entry->varlinenum, entry->vardata);
		return NULL;
	}

	ban = (ConfigItem_ban *)MyMalloc(sizeof(ConfigItem_ban));
	ban->flags |= opt->flag;
	ban->in_file = 1;
	ban->kill_on_rehash = 1;

	for (EACH_ENTRY(e, entry)) {
		VARIABLE(e, "ban")
		PARAMETER(e, "ban")

		if (!mycmp(e->varname, "mask")) {
			cDupString(ban->mask, e->vardata);
		}
		else if (!mycmp(e->varname, "reason")) {
			cDupString(ban->reason, e->vardata);
		}
		else {
			add_unknown(e, ban, CONF_BAN);
		}
	}

	dlink_add(&conf_ban_list, ban);
	return (void *)ban;
}

int test_ban(ConfigFile *file, void *item)
{
	ConfigItem_ban *ban = (ConfigItem_ban *)item;

	if (BadPtr(ban->mask)) {
		report(0, "%s: ban::mask: missing parameter, ignoring block", file->filename);
		return CONF_FAILURE;
	}

	return CONF_SUCCESS;
}

void free_ban(void *item)
{
	ConfigItem_ban *ban = (ConfigItem_ban *)item;
	cMyFree(ban->mask);
	cMyFree(ban->reason);
	cMyFree(ban);
}

void *parse_except(ConfigEntry *entry)
{
	ConfigEntry *e;
	ConfigItem_except *except;
	ConfigOption *opt;

	BLOCK_ENTRIES(entry, "except")

	for (EACH_ENTRY(e, entry)) {
		VARIABLE(e, "except")
		PARAMETER(e, "except")

		if ((opt = get_bantype(e->varname)) == NULL) {
			UNKNOWN_VAR(e, "except");
			continue;
		}
		if ((except = find_except(opt->flag, e->vardata)) != NULL) {
			continue;
		}

		except = (ConfigItem_except *)MyMalloc(sizeof(ConfigItem_except));
		except->flags |= opt->flag;
		cDupString(except->mask, e->vardata);
		dlink_add(&conf_except_list, except);
	}

	return NULL;
}

void *parse_vhost(ConfigEntry *entry)
{
	ConfigEntry *e;
	ConfigItem_vhost *vhost;

	BLOCK_NAME(entry, "vhost", "vhost mask")
	BLOCK_ENTRIES(entry, "vhost")

	vhost = (ConfigItem_vhost *)MyMalloc(sizeof(ConfigItem_vhost));
	cDupString(vhost->mask, entry->vardata);

	for (EACH_ENTRY(e, entry)) {
		VARIABLE(e, "vhost")
		if (!mycmp(e->varname, "auth")) {
			vhost->auth = auth_parse(e, "vhost::auth");
			continue;
		}

		PARAMETER(e, "vhost")
		if (!mycmp(e->varname, "login")) {
			cDupString(vhost->login, e->vardata);
		}
		else if (!mycmp(e->varname, "from")) {
			conf_add_from(&vhost->from, e->vardata);
		}
		else {
			UNKNOWN_VAR(e, "vhost");
		}
	}

	dlink_add(&conf_vhost_list, vhost);
	return (void *)vhost;
}

int test_vhost(ConfigFile *file, void *item)
{
	ConfigItem_vhost *vhost = (ConfigItem_vhost *)item;

	if (BadPtr(vhost->mask)) {
		report(0, "%s: vhost: missing vhost mask, ignoring block", file->filename);
		return CONF_FAILURE;
	}
	if (BadPtr(vhost->login)) {
		report(0, "%s: vhost::login: missing parameter, ignoring block", file->filename);
		return CONF_FAILURE;
	}
	if (vhost->auth == NULL) {
		report(0, "%s: vhost::auth: missing section, ignoring block", file->filename);
		return CONF_FAILURE;
	}
	if (!vhost->from.host_count) {
		report(0, "%s: vhost::from: missing parameter, ignoring block", file->filename);
		return CONF_FAILURE;
	}

	return CONF_SUCCESS;
}

void free_vhost(void *item)
{
	ConfigItem_vhost *vhost = (ConfigItem_vhost *)item;

	conf_clear_from(&vhost->from);
	if (vhost->auth != NULL) {
		auth_destroy_struct(vhost->auth);
	}

	cMyFree(vhost->mask);
	cMyFree(vhost->login);
	cMyFree(vhost);
}

void *parse_include(ConfigEntry *entry)
{
	ConfigItem_include *include;
	dlink_node *node;

	BLOCK_NAME(entry, "include", "file name")

	DLINK_FOREACH_DATA(conf_include_list.head, node, include, ConfigItem_include) {
		if (!mycmp(include->filename, entry->vardata)) {
			report(0, "%s:%d: include: config file %s already %s", entry->file->filename,
				entry->varlinenum, (include->loaded) ? "loaded" : "marked for inclusion");
			return NULL;
		}
	}

	include = (ConfigItem_include *)MyMalloc(sizeof(ConfigItem_include));
	cDupString(include->filename, entry->vardata);

	dlink_add(&conf_include_list, include);
	return (void *)include;
}

#ifndef STATIC_MODULES
void *parse_modules(ConfigEntry *entry)
{
	ConfigEntry *e;

	BLOCK_ENTRIES(entry, "modules")
	for (EACH_ENTRY(e, entry)) {
		VARIABLE(e, "modules")
		PARAMETER(e, "modules")

		if (!mycmp(e->varname, "module")) {
			do_module("load", e->vardata, NULL);
		}
		else if (!mycmp(e->varname, "path")) {
			load_modules_from_path(e->vardata);
		}
		else {
			UNKNOWN_VAR(e, "modules");
		}
	}

	return NULL;
};
#endif

void *parse_network(ConfigEntry *entry)
{
	ConfigEntry *e;
	ConfigItem_network *network = &tmpNetworkConfig;
	char *s;

	BLOCK_ENTRIES(entry, "network")
	for (EACH_ENTRY(e, entry)) {
		VARIABLE(e, "network")
		PARAMETER(e, "network")

		if (!mycmp(e->varname, "name")) {
			cDupString(network->name, e->vardata);
			for (s = e->vardata; *s != '\0'; s++) {
				if (*s == ' ') {
					*s = '-';
				}
			}
			cDupString(network->name_005, e->vardata);
		}
		else if (!mycmp(e->varname, "kline_address")) {
			cDupString(network->kline_address, e->vardata);
		}
		else if (!mycmp(e->varname, "services_server")) {
			cDupString(network->services_server, e->vardata);
		}
		else if (!mycmp(e->varname, "stats_server")) {
			cDupString(network->stats_server, e->vardata);
		}
		else if (!mycmp(e->varname, "max_link_depth")) {
			network->max_link_depth = atoi(e->vardata);
		}
		else {
			add_unknown(e, &NetworkConfig, CONF_NETWORK);
		}
	}

	return (void *)network;
}

int test_network(ConfigFile *file, void *item)
{
	ConfigItem_network *network = (ConfigItem_network *)item;
	static int retval = CONF_SUCCESS;

	for (;;) {
		if (BadPtr(network->name)) {
			if (Internal.rehashing) {
				report(0, "%s: network::name: missing parameter", file->filename);
				retval = CONF_FALLBACK;
				break;
			}

			cDupString(network->name, DEFAULT_NETWORK_NAME);
			cDupString(network->name_005, DEFAULT_NETWORK_NAME_005);
			report(0, "%s: network::name: missing parameter, using %s", file->filename,
				network->name);
		}
		if (BadPtr(network->kline_address) || !valid_email_addr(network->kline_address, 1)) {
			cDupString(network->kline_address, DEFAULT_KLINE_ADDRESS);
		}
		if (network->max_link_depth < 1) {
			network->max_link_depth = DEFAULT_LINK_DEPTH;
		}
		break;
	}

	return retval;
}

void free_network(void *item)
{
	ConfigItem_network *network = (ConfigItem_network *)item;
	cMyFree(network->name);
	cMyFree(network->name_005);
	cMyFree(network->kline_address);
	cMyFree(network->services_server);
	cMyFree(network->stats_server);
}

void *parse_masking(ConfigEntry *entry)
{
	ConfigEntry *e;
	ConfigItem_masking *masking = &tmpMaskingConfig;

	BLOCK_ENTRIES(entry, "masking")
	for (EACH_ENTRY(e, entry)) {
		VARIABLE(e, "masking")
		PARAMETER(e, "masking")

		if (!mycmp(e->varname, "netadmin_mask")) {
			cDupString(masking->netadmin_mask, e->vardata);
		}
		else if (!mycmp(e->varname, "admin_mask")) {
			cDupString(masking->admin_mask, e->vardata);
		}
		else if (!mycmp(e->varname, "oper_mask")) {
			cDupString(masking->oper_mask, e->vardata);
		}
		else if (!mycmp(e->varname, "user_mask_prefix")) {
			cDupString(masking->user_mask_prefix, e->vardata);
		}
		else if (!mycmp(e->varname, "mask_key1")) {
			masking->mask_key1 = atol(e->vardata);
		}
		else if (!mycmp(e->varname, "mask_key2")) {
			masking->mask_key2 = atol(e->vardata);
		}
		else if (!mycmp(e->varname, "mask_key3")) {
			masking->mask_key3 = atol(e->vardata);
		}
		else if (!mycmp(e->varname, "auto_mask_opers")) {
			masking->auto_mask_opers = get_option(e->vardata, 1);
		}
		else {
			add_unknown(e, masking, CONF_MASKING);
		}
	}

	return (void *)masking;
}

int test_masking(ConfigFile *file, void *item)
{
	ConfigItem_masking *masking = (ConfigItem_masking *)item;
	static int retval = CONF_SUCCESS;

	for (;;) {
		if (BadPtr(masking->user_mask_prefix)) {
			cDupString(masking->user_mask_prefix, DEFAULT_USER_MASK_PREFIX);
		}
		if (masking->mask_key1 < 10000) {
			if (Internal.rehashing) {
				report(0, "%s: masking::mask_key1: invalid key", file->filename);
				retval = CONF_FALLBACK;
			}
			else {
				report(1, "%s: ERROR: masking::mask_key1: invalid key", file->filename);
				retval = CONF_FAILURE;
			}
		}
		if (masking->mask_key2 < 10000) {
			if (Internal.rehashing) {
				report(0, "%s: masking::mask_key2: invalid key", file->filename);
				retval = CONF_FALLBACK;
			}
			else {
				report(1, "%s: ERROR: masking::mask_key2: invalid key", file->filename);
				retval = CONF_FAILURE;
			}
		}
		if (masking->mask_key3 < 10000) {
			if (Internal.rehashing) {
				report(0, "%s: masking::mask_key3: invalid key", file->filename);
				retval = CONF_FALLBACK;
			}
			else {
				report(1, "%s: ERROR: masking::mask_key3: invalid key", file->filename);
				retval = CONF_FAILURE;
			}
		}
		break;
	}

	return retval;
}

void free_masking(void *item)
{
	ConfigItem_masking *masking = (ConfigItem_masking *)item;
	cMyFree(masking->netadmin_mask);
	cMyFree(masking->admin_mask);
	cMyFree(masking->oper_mask);
	cMyFree(masking->user_mask_prefix);
}

void *parse_flood(ConfigEntry *entry)
{
	ConfigEntry *e;
	ConfigItem_flood *flood = &tmpFloodConfig;

	BLOCK_ENTRIES(entry, "flood")
	for (EACH_ENTRY(e, entry)) {
		VARIABLE(e, "flood")
		PARAMETER(e, "flood")

		if (!mycmp(e->varname, "msg_flood_count")) {
			flood->msg_flood_count = atoi(e->vardata);
		}
		else if (!mycmp(e->varname, "knock_delay")) {
			flood->knock_delay = atime(e->vardata);
		}
		else if (!mycmp(e->varname, "accept_notice_time")) {
			flood->accept_notice_time = atime(e->vardata);
		}
		else if (!mycmp(e->varname, "antispam_quit_msg_time")) {
			flood->antispam_quit_msg_time = atime(e->vardata);
		}
		else if (!mycmp(e->varname, "pace_wait_simple")) {
			flood->pace_wait_simple = atime(e->vardata);
		}
		else if (!mycmp(e->varname, "pace_wait_intense")) {
			flood->pace_wait_intense = atime(e->vardata);
		}
#ifdef USE_THROTTLE
		else if (!mycmp(e->varname, "max_connect_count")) {
			flood->max_connect_count = atoi(e->vardata);
		}
		else if (!mycmp(e->varname, "min_connect_time")) {
			flood->min_connect_time = atime(e->vardata);
		}
		else if (!mycmp(e->varname, "throttle_rejected_clients")) {
			flood->throttle_rejected_clients = get_option(e->vardata, 1);
		}
#endif
		else if (!mycmp(e->varname, "no_oper_flood")) {
			flood->no_oper_flood = get_option(e->vardata, 1);
		}
		else if (!mycmp(e->varname, "anti_nick_flood")) {
			flood->anti_nick_flood = get_option(e->vardata, 1);
		}
		else if (!mycmp(e->varname, "anti_away_flood")) {
			flood->anti_away_flood = get_option(e->vardata, 1);
		}
		else {
			add_unknown(e, flood, CONF_FLOOD);
		}
	}

	return (void *)flood;
}

void *parse_general(ConfigEntry *entry)
{
	ConfigEntry *e;
	ConfigItem_general *general = &tmpGeneralConfig;
	char *s;
	int mindex;
	xMode *t;

	BLOCK_ENTRIES(entry, "general")
	for (EACH_ENTRY(e, entry)) {
		VARIABLE(e, "general")
		PARAMETER(e, "general")

		if (!mycmp(e->varname, "max_chans_per_user")) {
			general->max_chans_per_user = atoi(e->vardata);
		}
		else if (!mycmp(e->varname, "connauth_timeout")) {
			general->connauth_timeout = atime(e->vardata);
		}
		else if (!mycmp(e->varname, "statsfile_save_freq")) {
			general->statsfile_save_freq = atime(e->vardata);
		}
		else if (!mycmp(e->varname, "modes_on_connect")) {
			for (s = e->vardata; *s != '\0'; s++) {
				if ((mindex = usermodes->map[(unsigned char)*s]) == -1) {
					continue;
				}

				t = &usermodes->table[mindex];

				if (NEEDOPER_UMODES & t->mode || (t->mode == UMODE_REGNICK)
				  || (t->mode == UMODE_DEAF)) {
					continue;
				}
				general->modes_on_connect |= t->mode;
			}
		}
		else if (!mycmp(e->varname, "join_on_connect")) {
			cDupString(general->join_on_connect, e->vardata);
		}
		else if (!mycmp(e->varname, "compression_level")) {
			general->compression_level = atoi(e->vardata);
		}
		else if (!mycmp(e->varname, "auto_connect_freq")) {
			general->auto_connect_freq = atime(e->vardata);
		}
		else if (!mycmp(e->varname, "ts_delta_warn")) {
			general->ts_delta_warn = atime(e->vardata);
		}
		else if (!mycmp(e->varname, "ts_delta_max")) {
			general->ts_delta_max = atime(e->vardata);
		}
		else if (!mycmp(e->varname, "short_motd")) {
			general->short_motd = get_option(e->vardata, 1);
		}
		else if (!mycmp(e->varname, "hide_ulined_servers")) {
			general->hide_ulined_servers = get_option(e->vardata, 1);
		}
		else if (!mycmp(e->varname, "check_identd")) {
			general->check_identd = get_option(e->vardata, 1);
		}
		else if (!mycmp(e->varname, "resolve_hostnames")) {
			general->resolve_hostnames = get_option(e->vardata, 1);
		}
		else if (!mycmp(e->varname, "show_headers")) {
			general->show_headers = get_option(e->vardata, 1);
		}
		else if (!mycmp(e->varname, "failed_oper_notice")) {
			general->failed_oper_notice = get_option(e->vardata, 1);
		}
		else if (!mycmp(e->varname, "custom_channels")) {
			general->custom_channels = get_option(e->vardata, 1);
		}
		else if (!mycmp(e->varname, "enable_knock")) {
			general->enable_knock = get_option(e->vardata, 1);
		}
		else if (!mycmp(e->varname, "custom_quit_msgs")) {
			general->custom_quit_msgs = get_option(e->vardata, 1);
		}
		else if (!mycmp(e->varname, "show_cliconn_quit_msgs")) {
			general->show_cliconn_quit_msgs = get_option(e->vardata, 1);
		}
		else if (!mycmp(e->varname, "show_invisible_lusers")) {
			general->show_invisible_lusers = get_option(e->vardata, 1);
		}
		else if (!mycmp(e->varname, "write_perm_klines_to_file")) {
			general->write_perm_klines_to_file = get_option(e->vardata, 1);
		}
		else if (!mycmp(e->varname, "flatten_links")) {
			general->flatten_links = get_option(e->vardata, 1);
		}
		else if (!mycmp(e->varname, "enable_map")) {
			general->enable_map = get_option(e->vardata, 1);
		}
		else if (!mycmp(e->varname, "spy_notices")) {
			general->spy_notices = get_option(e->vardata, 1);
		}
		else if (!mycmp(e->varname, "no_oper_accept")) {
			general->no_oper_accept = get_option(e->varname, 1);
		}
		else {
			add_unknown(e, general, CONF_GENERAL);
		}
	}

	return (void *)general;
}

int test_general(ConfigFile *file, void *item)
{
	ConfigItem_general *general = (ConfigItem_general *)item;
	static int retval = CONF_SUCCESS;

	if (!Internal.valid_conf && (retval != CONF_SUCCESS)) {
		return retval;
	}
	for (;;) {
		if (general->max_chans_per_user < 1) {
			general->max_chans_per_user = DEFAULT_CHANS_PER_USER;
		}
		if (general->connauth_timeout) {
			general->connauth_timeout = DEFAULT_CONNAUTH_TIMEOUT;
		}
		if (general->compression_level < 0 || general->compression_level > 9) {
			general->compression_level = DEFAULT_COMPRESSION_LEVEL;
		}
		if (general->ts_delta_max <= 0) {
			general->ts_delta_max = DEFAULT_TS_DELTA_MAX;
		}
		if (general->ts_delta_warn < 0) {
			general->ts_delta_warn = DEFAULT_TS_DELTA_WARN;
		}
		break;
	}

	return retval;
}

void free_general(void *item)
{
	ConfigItem_general *general = (ConfigItem_general *)item;
	cMyFree(general->join_on_connect);
}

ConfigItem_class *find_class(char *name)
{
	dlink_node *node;
	ConfigItem_class *class;

	DLINK_FOREACH_DATA(conf_class_list.head, node, class, ConfigItem_class) {
		if (!mycmp(class->name, name)) {
			return class;
		}
	}

	return NULL;
}

ConfigItem_oper *find_oper(char *name)
{
	dlink_node *node;
	ConfigItem_oper *oper;

	DLINK_FOREACH_DATA(conf_oper_list.head, node, oper, ConfigItem_oper) {
		if (!mycmp(oper->name, name)) {
			return oper;
		}
	}

	return NULL;
}

char *find_uline(char *servername)
{
	dlink_node *node;
	char *uline;

	DLINK_FOREACH_DATA(conf_uline_list.head, node, uline, char) {
		if (!match(uline, servername)) {
			return uline;
		}
	}

	return NULL;
}

ConfigItem_link *find_link(char *servername, char *host)
{
	dlink_node *node;
	ConfigItem_link *linkp;

	DLINK_FOREACH_DATA(conf_link_list.head, node, linkp, ConfigItem_link) {
		if (match(linkp->servername, servername) && match(servername, linkp->servername)) {
			continue;
		}
		if (!match(linkp->host, host) || !match(host, linkp->host)) {
			return linkp;
		}
	}

	return NULL;
}

ConfigItem_listen *find_listen(char *ipaddr, int port)
{
	dlink_node *node;
	ConfigItem_listen *listenp;

	DLINK_FOREACH_DATA(conf_listen_list.head, node, listenp, ConfigItem_listen) {
		if (listenp->port != port) {
			continue;
		}
		if (*listenp->ipaddr == '*') {
			return listenp;
		}
		else if (!match(listenp->ipaddr, ipaddr)) {
			return listenp;
		}
	}

	return NULL;
}

ConfigItem_ban *find_ban(unsigned int flags, char *mask)
{
	dlink_node *node;
	ConfigItem_ban *ban;

	DLINK_FOREACH_DATA(conf_ban_list.head, node, ban, ConfigItem_ban) {
		if ((ban->flags & flags) == flags && !match(ban->mask, mask)) {
			if (find_except(flags, mask) == NULL) {
				return ban;
			}
		}
	}

	return NULL;
}

ConfigItem_except *find_except(unsigned int flags, char *mask)
{
	dlink_node *node;
	ConfigItem_except *except;

	DLINK_FOREACH_DATA(conf_except_list.head, node, except, ConfigItem_except) {
		if ((except->flags & flags) && !match(except->mask, mask)) {
			return except;
		}
	}

	return NULL;
}

int conf_do_kline(ConfigItem_ban *ban, aClient *cptr, aClient *sptr)
{
	char *path = get_client_name(sptr, SHOW_IP);
	int netwide = (ban->flags & BAN_SVS) ? 1 : 0;
	char *type = (netwide) ? "AutoKill" : "K:Line";

	ircstp->is_ref++;
	ircdlog(LOG_BAN, "%s active for %s", type, path);
	sendto_realops_lev(REJ_LEV, "%s active for %s", type, path);

	send_me_notice(sptr, ":*** You are not welcome on this %s: %s",
		(netwide) ? "network" : "server", BadPtr(ban->reason) ? "no reason specified" : ban->reason);
	send_me_notice(sptr, ":*** For more information contact %s, including everything shown here.",
		(netwide) ? NetworkConfig.kline_address : ServerInfo->kline_address);
	if (ban->expire > 0) {
		send_me_notice(sptr, ":*** This %s will expire in %s", type,
			time_to_str((ban->set_at - timeofday) + ban->expire));
	}

#ifdef USE_THROTTLE
	if (FloodConfig.throttle_rejected_clients) {
		throttle_force(sptr->host);
	}
#endif

	send_me_numeric(sptr, ERR_YOUREBANNEDCREEP, (netwide) ? "AutoKilled" : "K:Lined");
	return exit_client(cptr, sptr, &me, (netwide) ? "AutoKilled" : "K:Lined");
}

int conf_do_gline(ConfigItem_ban *ban, aClient *cptr, aClient *sptr)
{
	char *path = get_client_name(sptr, SHOW_IP);
	int netwide = (ban->flags & BAN_SVS) ? 1 : 0;
	char *type = (netwide) ? "SG:Line" : "G:Line";

	ircstp->is_ref++;
	ircdlog(LOG_BAN, "%s active for %s", type, path);
	sendto_realops_lev(REJ_LEV, "%s active for %s (%s)", type, path,
		BadPtr(ban->reason) ? "no reason" : ban->reason);

	send_me_notice(sptr, ":*** Your gcos (real name) is not permitted on this %s: %s",
		(netwide) ? "network" : "server",
		BadPtr(ban->reason) ? "no reason specified" : ban->reason);
	send_me_notice(sptr, ":*** For more information contact %s, including everything shown here.",
		(netwide) ? NetworkConfig.kline_address : ServerInfo->kline_address);

#ifdef USE_THROTTLE
	if (FloodConfig.throttle_rejected_clients) {
		throttle_force(sptr->host);
	}
#endif

	send_me_numeric(sptr, ERR_YOUREBANNEDCREEP, (netwide) ? "SG:Lined" : "G:Lined");
	return exit_client(cptr, sptr, &me, (netwide) ? "SG:Lined" : "G:Lined");
}

int conf_do_zap(ConfigItem_ban *ban, char *ipaddr, int fd)
{
	char zbuf[512];
	int zlen;

	ircstp->is_ref++;
	ircdlog(LOG_BAN, "Zap active for *@%s", ipaddr);
	sendto_realops_lev(REJ_LEV, "Zap active for *@%s (%s)", ipaddr,
		BadPtr(ban->reason) ? "no reason" : ban->reason);

#ifdef USE_THROTTLE
	if (FloodConfig.throttle_rejected_clients) {
		throttle_force(ipaddr);
	}
#endif

	zlen = ircsnprintf(zbuf, 512, "ERROR :Your host has been zapped: %s\r\n",
		BadPtr(ban->reason) ? "no reason specified" : ban->reason);
	write(fd, zbuf, zlen);
	return 0;
}

void conf_rehash()
{
	dlink_node *node, *next = NULL;
	ConfigItem_class *class;
	ConfigItem_allow *allow;
	ConfigItem_oper *oper;
	ConfigItem_link *linkp;
	ConfigItem_listen *listenp;
	ConfigItem_ban *ban;
	ConfigItem_except *except;
	ConfigItem_vhost *vhost;
	char *cp;

	Internal.rehashing = 1;

	if (ServerInfo != NULL) {
		free_servinfo(ServerInfo);
		ServerInfo = NULL;
	}
	if (AdminInfo != NULL) {
		cMyFree(AdminInfo->name);
		cMyFree(AdminInfo->desc);
		cMyFree(AdminInfo->email);
		cMyFree(AdminInfo);
	}

	/* WARNING! Items with class references _must_ be cleared before conf_class_list
	 * in order to allow any class references to be removed.
	 */

	DLINK_FOREACH_SAFE_DATA(conf_allow_list.head, node, next, allow, ConfigItem_allow) {
		if (allow->item.refcount) {
			allow->item.temp = 1;
		}
		else {
			dlink_del(&conf_allow_list, NULL, node);
			free_allow(allow);
		}
	}
	DLINK_FOREACH_SAFE_DATA(conf_oper_list.head, node, next, oper, ConfigItem_oper) {
		if (oper->item.refcount) {
			oper->item.temp = 1;
		}
		else {
			dlink_del(&conf_oper_list, NULL, node);
			free_oper(oper);
		}
	}
	DLINK_FOREACH_SAFE_DATA(conf_link_list.head, node, next, linkp, ConfigItem_link) {
		if (linkp->item.refcount || linkp->servers) {
			linkp->item.temp = 1;
		}
		else {
			dlink_del(&conf_link_list, NULL, node);
			free_link(linkp);
		}
	}

	DLINK_FOREACH_SAFE_DATA(conf_class_list.head, node, next, class, ConfigItem_class) {
		if (class->item.perm) {
			continue;
		}
		if (class->item.refcount || class->clients) {
			class->item.temp = 1;
		}
		else {
			dlink_del(&conf_class_list, NULL, node);
			free_class(class);
		}
	}

	DLINK_FOREACH_SAFE_DATA(conf_listen_list.head, node, next, listenp, ConfigItem_listen) {
		if ((listenp->l != NULL) && listenp->l->clients) {
			listenp->item.temp = 1;
		}
		else {
			if (listenp->l != NULL) {
				close_one_listener(listenp->l, NULL);
			}
			dlink_del(&conf_listen_list, NULL, node);
			cMyFree(listenp->ipaddr);
			cMyFree(listenp);
		}
	}

	DLINK_FOREACH_SAFE_DATA(conf_uline_list.head, node, next, cp, char) {
		dlink_del(&conf_uline_list, NULL, node);
		cMyFree(cp);
	}
	DLINK_FOREACH_SAFE_DATA(conf_ban_list.head, node, next, ban, ConfigItem_ban) {
		if (!(ban->flags & BAN_SVS) && ban->kill_on_rehash) {
			dlink_del(&conf_ban_list, NULL, node);
			free_ban(ban);
		}
	}
	DLINK_FOREACH_SAFE_DATA(conf_except_list.head, node, next, except, ConfigItem_except) {
		dlink_del(&conf_except_list, NULL, node);
		cMyFree(except->mask);
		cMyFree(except);
	}
	DLINK_FOREACH_SAFE_DATA(conf_vhost_list.head, node, next, vhost, ConfigItem_vhost) {
		dlink_del(&conf_vhost_list, NULL, node);
		free_vhost(vhost);
	}

	hook_run(h_conf_rehash, NULL);

#ifndef STATIC_MODULES
	unload_all_modules();
#endif

	init_conf();
	check_core_commands();

	close_listeners();
	restart_resolver();
	open_listeners();

	build_isupport();
#ifdef USE_THROTTLE
	throttle_rehash();
#endif
	rehash_message_files();

	Internal.rehashing = 0;
	Internal.do_ban_check = 1;

	sendto_ops("Rehash complete.");
}

int attach_allow(aClient *cptr, ConfigItem_allow *allow)
{
	IPEntry *ipe;

	ASSERT(cptr != NULL);
	ASSERT(cptr->localClient != NULL);
	ASSERT(cptr->user != NULL);
	ASSERT(cptr->localUser != NULL);
	ASSERT(allow != NULL);

	ipe = find_or_add_ip(&cptr->ip);
	ipe->count++;
	SetHashedIP(cptr);

	if (allow->max_per_ip > 0 && (ipe->count > allow->max_per_ip)) {
		return CLIENTAUTH_TOOMANYIPS;
	}
	if ((allow->class->clients + 1) > allow->class->max_clients) {
		return CLIENTAUTH_ILINEFULL;
	}
	if (allow->auth != NULL && !auth_check(allow->auth, cptr->localClient->passwd)) {
		return CLIENTAUTH_INVALIDPW;
	}
	memset(cptr->localClient->passwd, '\0', PASSWDLEN + 1);

#ifdef CDEBUG
	Debug((DEBUG_DEBUG, "Attached allow %x to client %s(%x)", allow, cptr->name, cptr));
#endif

	cptr->localUser->allow = allow;
	allow->item.refcount++;

	return CLIENTAUTH_MATCHED;
}

void detach_allow(aClient *cptr)
{
	ConfigItem_allow *allow;

	ASSERT(cptr != NULL);
	ASSERT(cptr->localUser != NULL);

	if (cptr->localUser->allow == NULL) {
		return;
	}

	allow = cptr->localUser->allow;
	cptr->localUser->allow = NULL;
	allow->item.refcount--;

#ifdef CDEBUG
	Debug((DEBUG_DEBUG, "Detached allow %x from client %s(%x)", allow, cptr->name, cptr));
#endif

	if (!allow->item.refcount && allow->item.temp) {
#ifdef CDEBUG
		Debug((DEBUG_DEBUG, "Freeing allow conf"));
#endif
		dlink_del(&conf_allow_list, allow, NULL);
		free_allow(allow);
	}
}

void attach_class(aClient *cptr)
{
	ConfigItem_class *class = NULL;

	ASSERT(cptr != NULL);
	ASSERT(cptr->localClient != NULL);

#ifdef CDEBUG
	Debug((DEBUG_DEBUG, "Attempting to attach class to client %s(%x)", cptr->name, cptr));
#endif

	if (cptr->localClient->class != NULL) {
#ifdef CDEBUG
		Debug((DEBUG_DEBUG, "-> first removing client from old class %s(%x)",
			cptr->localClient->class->name, cptr->localClient->class));
#endif
		detach_class(cptr);
	}

	if (IsServer(cptr)) {
		if (cptr->serv->conf->class != NULL) {
			class = cptr->serv->conf->class;
		}
	}
	else if (cptr->localUser != NULL) {
		if (cptr->localUser->oper != NULL) {
			class = cptr->localUser->oper->class;
		}
		else if (cptr->localUser->allow != NULL) {
			class = cptr->localUser->allow->class;
		}
	}

	if (class == NULL) {
		class = DefaultClass;
	}

	cptr->localClient->class = class;
	class->item.refcount++;
	class->clients++;

#ifdef CDEBUG
	Debug((DEBUG_DEBUG, "Attached class %s(%x) to client %s(%x)", class->name, class, cptr->name, cptr));
#endif
}

void detach_class(aClient *cptr)
{
	ConfigItem_class *class;

	ASSERT(cptr != NULL);
	ASSERT(cptr->localClient != NULL);

	if (cptr->localClient->class == NULL) {
		return;
	}

	class = cptr->localClient->class;
	cptr->localClient->class = NULL;
	class->item.refcount--;
	class->clients--;

#ifdef CDEBUG
	Debug((DEBUG_DEBUG, "Detached class %s(%x) from client %s(%x)", class->name, class,
		cptr->name, cptr));
#endif

	if (!class->clients && !class->item.refcount && class->item.temp && !class->item.perm) {
#ifdef CDEBUG
		Debug((DEBUG_DEBUG, "Freeing class conf"));
#endif
		dlink_del(&conf_class_list, class, NULL);
		free_class(class);
	}
}

void attach_oper(aClient *cptr, ConfigItem_oper *oper)
{
	ASSERT(cptr != NULL);
	ASSERT(cptr->localUser != NULL);
	ASSERT(oper != NULL);

	cptr->localUser->oper = oper;
	oper->item.refcount++;

#ifdef CDEBUG
	Debug((DEBUG_DEBUG, "Attached oper %s(%x) to client %s(%x)", oper->name, oper, cptr->name, cptr));
#endif
}

void detach_oper(aClient *cptr)
{
	ConfigItem_oper *oper;

	ASSERT(cptr != NULL);
	ASSERT(cptr->localUser != NULL);

	if (cptr->localUser->oper == NULL) {
		return;
	}

	oper = cptr->localUser->oper;
	cptr->localUser->oper = NULL;
	oper->item.refcount--;

#ifdef CDEBUG
	Debug((DEBUG_DEBUG, "Detached oper %s(%x) from client %s(%x)", oper->name, oper, cptr->name, cptr));
#endif

	if (!oper->item.refcount && oper->item.temp) {
#ifdef CDEBUG
		Debug((DEBUG_DEBUG, "Freeing oper conf"));
#endif
		dlink_del(&conf_oper_list, oper, NULL);
		free_oper(oper);
	}
}

void attach_link(aClient *cptr, ConfigItem_link *conf)
{
	ASSERT(cptr != NULL);
	ASSERT(cptr->serv != NULL);
	ASSERT(conf != NULL);

	cptr->serv->conf = conf;
	conf->item.refcount++;
	conf->servers++;

#ifdef CDEBUG
	Debug((DEBUG_DEBUG, "Attached link %s(%x) to server %s(%x)", conf->servername, conf, cptr->name, cptr));
#endif
}

void detach_link(aClient *cptr)
{
	ConfigItem_link *conf;

	ASSERT(cptr != NULL);
	ASSERT(cptr->serv != NULL);

	if (cptr->serv->conf == NULL) {
		return;
	}

	conf = cptr->serv->conf;
	cptr->serv->conf = NULL;
	conf->item.refcount--;
	conf->servers--;

#ifdef CDEBUG
	Debug((DEBUG_DEBUG, "Detached link %s(%x) from server %s(%x)", conf->servername, conf, cptr->name, cptr));
#endif

	if (!conf->item.refcount && !conf->servers && conf->item.temp) {
#ifdef CDEBUG
		Debug((DEBUG_DEBUG, "Freeing link conf"));
#endif
		dlink_del(&conf_link_list, conf, NULL);
		free_link(conf);
	}
}

void attach_listen(Listener *l, ConfigItem_listen *conf)
{
	ASSERT(l != NULL);
	ASSERT(conf != NULL);

	l->conf = conf;
	conf->item.refcount++;

	conf->l = l;
	l->item.refcount++;

#ifdef CDEBUG
	Debug((DEBUG_DEBUG, "Attached listen %x to listener %s/%d(%x)", conf, l->ipaddr, l->port, l));
#endif
}

void detach_listen(Listener *l)
{
	ConfigItem_listen *conf;

	ASSERT(l != NULL);
	ASSERT(l->conf != NULL);

	conf = l->conf;
	l->conf = NULL;
	conf->item.refcount--;

	conf->l = NULL;
	l->item.refcount--;

#ifdef CDEBUG
	Debug((DEBUG_DEBUG, "Detached listen %x from listener %s/%d(%x)", conf, l->ipaddr, l->port, l));
#endif

	if (!conf->item.refcount && conf->item.temp) {
#ifdef CDEBUG
		Debug((DEBUG_DEBUG, "Freeing listen conf"));
#endif
		dlink_del(&conf_listen_list, conf, NULL);
		cMyFree(conf->ipaddr);
		cMyFree(conf);
	}
}

void detach_confs(aClient *cptr)
{
	ASSERT(cptr != NULL);

#ifdef CDEBUG
	Debug((DEBUG_DEBUG, "detach_confs() for client %s (%x)", cptr->name, cptr));
#endif

	if (IsServer(cptr)) {
		detach_link(cptr);
	}
	else if (cptr->user != NULL) {
		detach_allow(cptr);
		detach_oper(cptr);
	}

	detach_class(cptr);
}
