/*
 * RageIRCd, an advanced Internet Relay Chat daemon (IRCd)
 * src/s_misc.c: misc. functions
 *
 * Copyright (C) 2002 the past and present IRCd developers, and others.
 * Refer to doc/authors/ for full credits and copyright details.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA.
 *
 * $Id: s_misc.c,v 1.114 2004/05/13 22:11:47 rageircd Exp $
 */

#include "struct.h"
#include "common.h"
#include "sys.h"
#include "numeric.h"
#include "zlink.h"
#include "memory.h"
#include "h.h"
#include "fd.h"
#include "hook.h"
#include "dlink.h"
#include "res.h"
#include "xmode.h"
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/time.h>
#include <time.h>
#include <sys/param.h>

#ifdef ALWAYS_SEND_DURING_SPLIT
int currently_processing_netsplit = 0;
#endif

static char *months[] = {
	"January",
	"February",
	"March",
	"April",
	"May",
	"June",
	"July",
	"August",
	"September",
	"October",
	"November",
	"December"
};
static char *weekdays[] = {
	"Sunday",
	"Monday",
	"Tuesday",
	"Wednesday",
	"Thursday",
	"Friday",
	"Saturday"
};

static void exit_one_client(aClient *, aClient *, aClient *, char *);

struct stats ircst, *ircstp = &ircst;

char *date(time_t c)
{
	static char buf[80], plus;
	struct tm *lt, *gm, gmbuf;
	int minswest;

	if (!c) {
		time(&c);
	}

	gm = gmtime(&c);
	memcpy((char *)&gmbuf, (char *)gm, sizeof(gmbuf));
	gm = &gmbuf;
	lt = localtime(&c);

	if (lt->tm_yday == gm->tm_yday) {
		minswest = (gm->tm_hour - lt->tm_hour) * 60 + (gm->tm_min - lt->tm_min);
	}
	else if (lt->tm_yday > gm->tm_yday) {
		minswest = (gm->tm_hour - (lt->tm_hour + 24)) * 60;
	}
	else {
		minswest = ((gm->tm_hour + 24) - lt->tm_hour) * 60;
	}

	plus = (minswest > 0) ? '-' : '+';
	if (minswest < 0) {
		minswest = -minswest;
	}
	ircsprintf(buf, "%s %s %d %04d -- %02d:%02d %c%02d:%02d", weekdays[lt->tm_wday], months[lt->tm_mon],
		lt->tm_mday, lt->tm_year + 1900, lt->tm_hour, lt->tm_min, plus, minswest / 60, minswest % 60);
	return buf;
}

char *smalldate(time_t c)
{
	static char buf[MAX_DATE_STRING];
	struct tm *lt, *gm, gmbuf;

	if (!c) {
		time(&c);
	}
	gm = gmtime(&c);
	memcpy((char *)&gmbuf, (char *)gm, sizeof(gmbuf));
	gm = &gmbuf;
	lt = localtime(&c);
	ircsprintf(buf, "%04d/%02d/%02d %02d.%02d", lt->tm_year + 1900, lt->tm_mon + 1, lt->tm_mday,
		lt->tm_hour, lt->tm_min);
	return buf;
}

char *myctime(time_t value)
{
	static char buf[28];
	char *p;

	strcpy(buf, ctime(&value));
	if ((p = strchr(buf, '\n')) != NULL) {
		*p = '\0';
	}
	return buf;
}

char *check_string(char *str)
{
	char *p;

	if (BadPtr(str)) {
		return "*";
	}
	for (p = str; *p != '\0'; p++) {
		if (isspace(*p)) {
			*p = '\0';
			break;
		}
	}
	return BadPtr(str) ? "*" : str;
}

char *make_nick_user_host(char *nick, char *user, char *host)
{
	static char namebuf[NICKLEN + USERLEN + HOSTLEN + 6];
	int n;
	char *t = namebuf, *s;

	for (s = check_string(nick), n = NICKLEN; *s != '\0' && n-- > 0;) {
		*t++ = *s++;
	}

	*t++ = '!';
	for (s = check_string(user), n = USERLEN; *s != '\0' && n-- > 0;) {
		*t++ = *s++;
	}

	*t++ = '@';
	for (s = check_string(host), n = HOSTLEN; *s != '\0' && n-- > 0;) {
		*t++ = *s++;
	}

	*t = '\0';
	return namebuf;
}

char *make_user_host(char *user, char *host)
{
	static char namebuf[USERLEN + HOSTLEN + 4];
	char *t = namebuf, *s;
	int n;

	for (s = check_string(user), n = USERLEN; *s != '\0' && n-- > 0;) {
		*t++ = *s++;
	}

	*t++ = '@';
	for (s = check_string(host), n = HOSTLEN; *s != '\0' && n-- > 0;) {
		*t++ = *s++;
	}

	*t = '\0';
	return namebuf;
}

char *get_client_name(aClient *sptr, int show_ip)
{
	static char nbuf[HOSTLEN * 2 + USERLEN + 5];

	ASSERT(sptr != NULL);

	if (!MyConnect(sptr)) {
		return sptr->name;
	}

	ASSERT(sptr->localClient != NULL);

	switch (show_ip) {
		case SHOW_IP:
			ircsprintf(nbuf, "%s[%s@%s]", sptr->name, (!GotID(sptr)) ? "" : sptr->username,
				(*sptr->hostip != '\0') ? sptr->hostip : inetntoa((const char *)&sptr->ip));
			break;
		case HIDE_IP:
			ircsprintf(nbuf, "%s[%s@0.0.0.0]", sptr->name, (!GotID(sptr)) ? "" : sptr->username);
			break;
		default:
			if (!mycmp(sptr->name, sptr->localClient->sockhost)) {
				return sptr->name;
			}
			ircsprintf(nbuf, "%s[%s@%s]", sptr->name, sptr->username, sptr->localClient->sockhost);
			break;
	}

	return nbuf;
}

void target_left(aClient *cptr, char *target_name, char *cmd, char *msgbuf)
{
	if (IsMe(cptr)) {
		return;
	}
	if (MyClient(cptr) || *target_name != '!') {
		send_me_numeric(cptr, ERR_NOSUCHNICK, target_name);
	}
	else if (msgbuf != NULL) {
		char truncated_msgbuf[24], *p;
		int i = 0;

		for (p = msgbuf; (*p != '\0') && (i < 20); p++) {
			truncated_msgbuf[i++] = *p;
		}
		if (i == 20) {
			truncated_msgbuf[i++] = '.';
			truncated_msgbuf[i++] = '.';
			truncated_msgbuf[i++] = '.';
		}
		truncated_msgbuf[i] = '\0';

		send_me_numeric_buf(cptr, "* :Target lost in transit. Could not deliver %s: %s",
			ERR_NOSUCHNICK, cmd, truncated_msgbuf);
	}
	else {
		send_me_numeric_buf(cptr, "* :Target lost in transit. Could not deliver %s",
			ERR_NOSUCHNICK, cmd);
	}
}

static void exit_one_client_in_split(aClient *sptr, aClient *dead, char *reason)
{
	chanMember *cm;
	SLink *lp;
	HookData hdata;

	sendto_common_channels(sptr, ":%s QUIT :%s", sptr->name, reason);

	while ((cm = sptr->user->channel) != NULL) {
		remove_user_from_channel(sptr, cm->chptr);
	}
	while ((lp = sptr->user->silence) != NULL) {
		del_silence(sptr, lp->value.cp);
	}

	hdata.cptr = dead;
	hdata.sptr = sptr;
	hdata.c = reason;
	hook_run(h_exit_user_remote, &hdata);

	if (HasSUID(sptr)) {
		del_userid_from_serv(sptr->uplink, sptr);
	}

	del_from_client_hash_table(sptr->name, sptr);
	hash_check_watch(sptr, RPL_LOGOFF);
	remove_client_from_list(sptr);
}

static void exit_one_server(aClient *cptr, aClient *dead, aClient *from, aClient *lcptr, char *spinfo, char *comment)
{
	aClient *acptr, *next = NULL;
	dlink_node *node;
	HookData hdata;

	Debug((DEBUG_NOTICE, "server noquit: %s", cptr->name));

	for (acptr = client; acptr != NULL; acptr = next)  {
		next = acptr->next;
		if (acptr->uplink != cptr || !IsPerson(acptr)) {
			continue;
		}
		exit_one_client_in_split(acptr, dead, spinfo);
	}
	for (acptr = client; acptr != NULL; acptr = next) {
		next = acptr->next;
		if (acptr->uplink != cptr || !IsServer(acptr)) {
			continue;
		}
		exit_one_server(acptr, dead, from, lcptr, spinfo, comment);
		next = client;
	}

	Debug((DEBUG_NOTICE, "done exiting server: %s", cptr->name));

	DLINK_FOREACH_DATA(lserver_list.head, node, acptr, aClient) {
		if (acptr == cptr || IsMe(acptr) || acptr == dead || acptr == lcptr || cptr != dead) {
			continue;
		}
		sendto_one_client_real(acptr, cptr, (cptr->from == acptr) ? from : NULL,
			&CMD_SQUIT, ":%s", comment);
	}

	hdata.sptr = cptr;
	hdata.c = comment;
	hook_run(h_exit_server, &hdata);

	del_from_client_hash_table(cptr->name, cptr); 
	hash_check_watch(cptr, RPL_LOGOFF);

	if (HasSUID(cptr)) {
		del_base64_serv(cptr);
	}

	remove_client_from_list(cptr);
}

static void exit_server(aClient *lcptr, aClient *cptr, aClient *from, char *comment)
{
	char splitname[HOSTLEN + HOSTLEN + 2];

	Debug((DEBUG_NOTICE, "exit_server(%s, %s, %s)", cptr->name, from->name, comment));

	ircsprintf(splitname, "%s %s", cptr->uplink->name, cptr->name);
	exit_one_server(cptr, cptr, from, lcptr, splitname, comment);
}

int exit_client(aClient *cptr, aClient *sptr, aClient *from, char *comment)
{
	if (MyConnect(sptr)) {
		ASSERT(sptr->localClient != NULL);

		if (IsClosing(sptr)) {
			return 0;
		}
		SetClosing(sptr);

		dns_delete_queries(sptr->localClient->dns_query);
		ident_delete_queries(sptr);

		if (HashedIP(sptr)) {
			remove_one_ip(&sptr->ip);
		}
		if ((sptr->serv != NULL) && DoingDKEY(sptr)) {
			sendto_realops("Lost server %s during negotiation: %s", sptr->name, comment);
		}

		if (IsServer(sptr)) {
			Count.myserver--;
			if (IsULine(sptr)) {
				Count.myulined--;
			}

			dlink_del_nofree(&lserver_list, NULL, &sptr->localClient->self);
		}

		if (IsPerson(sptr)) {
			SLink *lp, *next = NULL;
			HookData hdata;

			hash_del_watch_list(sptr);

			if (sptr->localUser->lopt != NULL) {
				LOpts *lopt = sptr->localUser->lopt;

				dlink_del(&listingcli_list, sptr, NULL);
				for (lp = lopt->yeslist; lp != NULL; lp = next) {
					next = lp->next;
					MyFree(lp->value.cp);
					free_slink(lp);
				}
				for (lp = lopt->nolist; lp != NULL; lp = next) {
					next = lp->next;
					MyFree(lp->value.cp);
					free_slink(lp);
				}
				MyFree(sptr->localUser->lopt);
			}

			ircdlog(LOG_CLIENT, "disconnecting client %s (%s)", get_client_name(sptr, FALSE),
				comment);
			sendto_realops_lev(CCONN_LEV, "Client exiting: %s (%s@%s) [%s]", sptr->name, sptr->username,
				sptr->host, NormalExit(sptr) ?
				!GeneralConfig.show_cliconn_quit_msgs ? "Client Quit" : comment : comment);

			hdata.sptr = sptr;
			hdata.c = comment;
			hook_run(h_exit_user_local, &hdata);
		}
		else if (IsUnknown(sptr)) {
			HookData hdata;

			Count.unknown--;

			hdata.sptr = sptr;
			hdata.c = comment;
			hook_run(h_exit_unknown, &hdata);
		}
		if (sptr->localClient->fd >= 0) {
			if (cptr != NULL && (sptr != cptr)) {
				sendto_one_client_nopostfix(sptr, NULL, &CMD_ERROR, ":Closing Link: %s %s (%s)",
					IsPerson(sptr) ? sptr->localClient->sockhost : "0.0.0.0",
					sptr->name, comment);
			}
			else {
				sendto_one_client_nopostfix(sptr, NULL, &CMD_ERROR, ":Closing Link: %s (%s)",
					IsPerson(sptr) ? sptr->localClient->sockhost : "0.0.0.0", comment);
			}
		}

		close_connection(sptr);
		SetDeadSocket(sptr);

		if (IsServer(sptr)) {
			sendto_ops("%s was connected for %s. %lu/%lu sendK/recvK.", sptr->name,
				time_to_str(timeofday - sptr->firsttime), sptr->localClient->sendK,
				sptr->localClient->receiveK);
			ircdlog(LOG_SERVER, "%s was connected for %s. %lu/%lu sendK/recvK.",
				sptr->name, time_to_str(timeofday - sptr->firsttime),
				sptr->localClient->sendK, sptr->localClient->receiveK);
		}

		if (!IsRegistered(sptr)) {
			dlink_del_nofree(&lunknown_list, NULL, &sptr->localClient->self);
		}
		else if (IsClient(sptr)) {
			if (IsClient(sptr)) {
				Count.local--;
				dlink_del_nofree(&lclient_list, NULL, &sptr->localClient->self);
			}
			if (HasMode(sptr, UMODE_OPER)) {
				dlink_del(&oper_list, sptr, NULL);
			}
			if (ParsingRecvQ(sptr)) {
				dlink_del(&recvqcli_list, sptr, NULL);
			}
		}
	}
	else if (IsPerson(sptr) && !IsULine(sptr)) {
		sendto_realops_lev(GCCONN_LEV, "Client exiting at %s: %s (%s@%s) [%s]", sptr->user->server,
			sptr->name, sptr->username, sptr->host,
			NormalExit(sptr) ? !GeneralConfig.show_cliconn_quit_msgs ? "Client Quit" : comment : comment);
	}

#ifdef DEBUGMODE
	{
		dlink_node *lu_node = dlink_find(&lunknown_list, sptr);
		dlink_node *lc_node = dlink_find(&lclient_list, sptr);
		dlink_node *ls_node = dlink_find(&lserver_list, sptr);
		dlink_node *o_node = dlink_find(&oper_list, sptr);
		dlink_node *lst_node = dlink_find(&listingcli_list, sptr);
		dlink_node *rc_node = dlink_find(&recvqcli_list, sptr);

		if (lu_node || lc_node || ls_node || o_node || lst_node || rc_node) {
			sendto_realops_lev(DEBUG_LEV, "DLINK ERROR! %s still present in lists! lu "
				"%x lc %x ls %x o %x lst %x rc %x", get_client_name(sptr, FALSE),
				lu_node, lc_node, ls_node, o_node, lst_node, rc_node);
			ircdlog(LOG_ERROR, "DLINK ERROR! %s still present in lists! lu %x lc %x "
				"ls %x o %x lst %x rc %x", get_client_name(sptr, FALSE),
				lu_node, lc_node, ls_node, o_node, lst_node, rc_node);
			Debug((DEBUG_ERROR, "DLINK ERROR! %s still present in lists! lu %x lc %x "
				"ls %x o %x lst %x rc %x", get_client_name(sptr, FALSE),
				lu_node, lc_node, ls_node, o_node, lst_node, rc_node));
		}
		if (lu_node) dlink_del(&lunknown_list, NULL, lu_node);
		if (lc_node) dlink_del(&lclient_list, NULL, lc_node);
		if (ls_node) dlink_del(&lserver_list, NULL, ls_node);
		if (o_node) dlink_del(&oper_list, NULL, o_node);
		if (lst_node) dlink_del(&listingcli_list, NULL, lst_node);
		if (rc_node) dlink_del(&recvqcli_list, NULL, rc_node);
	}
#endif

	exit_one_client(cptr, sptr, from, comment);
	return cptr == sptr ? FLUSH_BUFFER : 0;
}

static void exit_one_client(aClient *cptr, aClient *sptr, aClient *from, char *comment)
{
	chanMember *cm;
	SLink *lp;
	HookData hdata;

	if (IsMe(sptr)) {
		sendto_ops("ERROR: tried to exit me! : %s", comment);
		return;
	}
	else if (IsServer(sptr)) {
#ifdef ALWAYS_SEND_DURING_SPLIT
		currently_processing_netsplit = 1;
#endif
		exit_server(cptr, sptr, from, comment);
#ifdef ALWAYS_SEND_DURING_SPLIT
		currently_processing_netsplit = 0;
#endif
		return;
	}
	else if (!(IsPerson(sptr))) {
		/* Nothing! */
	}
	else if (*sptr->name) {
		if (!IsKilled(sptr)) {
			sendto_serv_msg_butone(cptr, sptr, NULL, &CMD_QUIT, ":%s", comment);
		}
		if (sptr->user != NULL) {
			send_part_to_common_channels(sptr, comment);
			send_quit_to_common_channels(sptr, comment);

			while ((cm = sptr->user->channel) != NULL) {
				remove_user_from_channel(sptr, cm->chptr);
			}
			while ((lp = sptr->user->silence) != NULL) {
				del_silence(sptr, lp->value.cp);
			}
			if (MyConnect(sptr)) {
				free_local_userid(sptr);
				while ((lp = sptr->localUser->invited) != NULL) {
					del_invite(sptr, lp->value.chptr);
				}
			}
			else {
				hdata.sptr = sptr;
				hdata.c = comment;
				hook_run(h_exit_user_remote, &hdata);
			}
		}
	}
	if (HasSUID(sptr)) {
		if (IsServer(sptr)) {
			del_base64_serv(sptr);
		}
		else {
			del_userid_from_serv(sptr->uplink, sptr);
		}
	}
	if (IsHashed(sptr) && del_from_client_hash_table(sptr->name, sptr) != 1) {
		sendto_realops_lev(DEBUG_LEV, "HASHING ERROR! %#x !in tab %s[%s] from %#x next %#x "
			"hnext %#x prev %#x fd %d status %d user %#x", sptr, sptr->name,
			sptr->from ? sptr->from->localClient->sockhost : "??host", sptr->from, sptr->next,
			sptr->hnext, sptr->prev, sptr->localClient->fd, sptr->status, sptr->user);
		ircdlog(LOG_ERROR, "HASHING ERROR! %#x !in tab %s[%s] from %#x next %#x "
			"hnext %#x prev %#x fd %d status %d user %#x", sptr, sptr->name,
			sptr->from ? sptr->from->localClient->sockhost : "??host", sptr->from, sptr->next,
			sptr->hnext, sptr->prev, sptr->localClient->fd, sptr->status, sptr->user);
		Debug((DEBUG_ERROR, "HASHING ERROR! %#x !in tab %s[%s] from %#x next %#x "
			"hnext %#x prev %#x fd %d status %d user %#x", sptr, sptr->name,
			sptr->from ? sptr->from->localClient->sockhost : "??host", sptr->from, sptr->next,
			sptr->hnext, sptr->prev, sptr->localClient->fd, sptr->status, sptr->user));
	}
	if (IsRegistered(sptr)) {
		hash_check_watch(sptr, RPL_LOGOFF);
	}
	remove_client_from_list(sptr);
}

void init_stats()
{
	memset((char *)&ircst, '\0', sizeof(ircst));
}

void serv_stats(aClient *cptr)
{
	aClient *acptr;
	dlink_node *node;
	struct stats tmp, *sp = &tmp;

	memcpy(sp, ircstp, sizeof(*sp));

	sp->is_sv = dlink_length(&lserver_list);
	DLINK_FOREACH_DATA(lserver_list.head, node, acptr, aClient) {
		sp->is_sbs += acptr->localClient->sendB;
		sp->is_sbr += acptr->localClient->receiveB;
		sp->is_sks += acptr->localClient->sendK;
		sp->is_skr += acptr->localClient->receiveK;
		sp->is_sti += timeofday - acptr->firsttime;

		if (sp->is_sbs > 1023) {
			sp->is_sks += (sp->is_sbs >> 10);
			sp->is_sbs &= 0x3ff;
		}
		if (sp->is_sbr > 1023) {
			sp->is_skr += (sp->is_sbr >> 10);
			sp->is_sbr &= 0x3ff;
		}
	}

	sp->is_cl = dlink_length(&lclient_list);
	DLINK_FOREACH_DATA(lclient_list.head, node, acptr, aClient) {
		sp->is_cbs += acptr->localClient->sendB;
		sp->is_cbr += acptr->localClient->receiveB;
		sp->is_cks += acptr->localClient->sendK;
		sp->is_ckr += acptr->localClient->receiveK;
		sp->is_cti += timeofday - acptr->firsttime;

		if (sp->is_cbs > 1023) {
			sp->is_cks += (sp->is_cbs >> 10);
			sp->is_cbs &= 0x3ff;
		}
		if (sp->is_cbr > 1023) {
			sp->is_ckr += (sp->is_cbr >> 10);
			sp->is_cbr &= 0x3ff;
		}
	}

	sp->is_ni = dlink_length(&lunknown_list);

	send_me_debug(cptr, ":accepts %u refused %u", sp->is_ac, sp->is_ref);
	send_me_debug(cptr, ":unknown commands %u prefixes %u", sp->is_unco, sp->is_unpf);
	send_me_debug(cptr, ":nick collisions %u unknown closes %u", sp->is_kill, sp->is_ni);
	send_me_debug(cptr, ":wrong direction %u empty %u", sp->is_wrdi, sp->is_empt);
	send_me_debug(cptr, ":numerics seen %u mode fakes %u", sp->is_num, sp->is_fake);
	send_me_debug(cptr, ":auth successes %u fails %u", sp->is_asuc, sp->is_abad);
	send_me_debug(cptr, ":udp packets %u", sp->is_udp);

	send_me_debugNA(cptr, ":Client Server");

	send_me_debug(cptr, ":connected %u %u", sp->is_cl, sp->is_sv);
	send_me_debug(cptr, ":bytes sent %u.%uK %u.%uK", sp->is_cks, sp->is_cbs, sp->is_sks, sp->is_sbs);
	send_me_debug(cptr, ":bytes recv %u.%uK %u.%uK", sp->is_ckr, sp->is_cbr, sp->is_skr, sp->is_sbr);
	send_me_debug(cptr, ":time connected %u %u", sp->is_cti, sp->is_sti);
#ifdef FLUD
	send_me_debug(cptr, ":CTCP floods blocked %u", sp->is_flud);
#endif
}

#define	_1MEG	(1024.0)
#define	_1GIG	(1024.0 * 1024.0)
#define	_1TER	(1024.0 * 1024.0 * 1024.0)
#define	_GMKs(x)	((x > _1TER) ? "Terabytes" : ((x > _1GIG) ? "Gigabytes" : \
			((x > _1MEG) ? "Megabytes" : "Kilobytes")))
#define	_GMKv(x)	((x > _1TER) ? (float)(x / _1TER) : ((x > _1GIG) ? \
			(float)(x / _1GIG) : ((x > _1MEG) ? (float)(x / _1MEG) : (float)x)))

void serv_info(aClient *cptr)
{
	static char Lformat[] = "%s %u %u %u %u %u :%u %u %s";
	long sendK = 0, receiveK = 0, uptime;
	aClient *acptr;
	int i = 0;
	dlink_node *node;

	DLINK_FOREACH_DATA(lserver_list.head, node, acptr, aClient) {
		if (GeneralConfig.hide_ulined_servers && IsULine(acptr) && !HasMode(cptr, UMODE_OPER)) {
			continue;
		}

		sendK += acptr->localClient->sendK;
		receiveK += acptr->localClient->receiveK;

		send_me_numeric_buf(cptr, Lformat, RPL_STATSLINKINFO,
			get_client_name(acptr, HIDE_IP), (int)SBufLength(&acptr->localClient->sendQ),
			(int)acptr->localClient->sendM, (int)acptr->localClient->sendK,
			(int)acptr->localClient->receiveM, (int)acptr->localClient->receiveK,
			timeofday - acptr->firsttime, timeofday - acptr->since,
			CapTS(acptr) ? "TS" : "Non-TS");

		if (InputRC4(acptr) && OutputRC4(acptr)) {
			send_me_debugNA(cptr, ": - RC4 Encrypted");
		}
		if (InputZIP(acptr)) {
			unsigned long in_b, out_b;
			double ratio;

			zip_in_get_stats(acptr->serv->zip_in, &in_b, &out_b, &ratio);
			if (out_b) {
				send_me_debug(cptr, ": - [I] Zip inbytes %lu, outbytes %lu (%3.2f%%)",
					in_b, out_b, ratio);
			}
		}
		if (OutputZIP(acptr)) {
			unsigned long in_b, out_b;
			double ratio;

			zip_out_get_stats(acptr->serv->zip_out, &in_b, &out_b, &ratio);
			if (in_b) {
				send_me_debug(cptr, ": - [O] Zip inbytes %lu, outbytes %lu (%3.2f%%)",
					in_b, out_b, ratio);
			}
		}
		i++;
	}

	send_me_debug(cptr, ":%u total server%s", i, (i == 1) ? "" : "s");
	send_me_debug(cptr, ":Sent total : %7.2f %s", _GMKv(sendK), _GMKs(sendK));
	send_me_debug(cptr, ":Recv total : %7.2f %s", _GMKv(receiveK), _GMKs(receiveK));

	uptime = (timeofday - me.since);
	send_me_debug(cptr, ":Server send: %7.2f %s (%4.1fK/s total, %4.1fK/s current)",
		_GMKv(me.localClient->sendK), _GMKs(me.localClient->sendK),
		(float)((float)me.localClient->sendK/(float)uptime), Internal.curr_sendK);
	send_me_debug(cptr, ":Server recv: %7.2f %s (%4.1fK/s total, %4.1fK/s current)",
		_GMKv(me.localClient->receiveK), _GMKs(me.localClient->receiveK),
		(float)((float)me.localClient->receiveK/(float)uptime), Internal.curr_recvK);
}
