/*
 * RageIRCd, an advanced Internet Relay Chat daemon (IRCd)
 * src/send.c: sending routenes
 *
 * Copyright (C) 2002 the past and present IRCd developers, and others.
 * Refer to doc/authors/ for full credits and copyright details.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA.
 *
 * $Id: send.c,v 1.129 2004/05/13 18:03:59 rageircd Exp $
 */

#include "struct.h"
#include "common.h"
#include "sys.h"
#include "h.h"
#include "numeric.h"
#include "ssl.h"
#include "zlink.h"
#include "memory.h"
#include "msg.h"
#include "fd.h"
#include "dlink.h"
#include "hook.h"
#include "xmode.h"
#include <stdio.h>
#include <time.h>

#ifdef ALWAYS_SEND_DURING_SPLIT
extern int currently_processing_netsplit;
#endif

static char sendservbuf[4096];
static char sendbuf[2048];
#ifdef USE_OPENSSL
char rc4buf[RC4_BUF_SIZE];
#endif

static int sent_serial = 0;

#define NEXT_SERIAL	if (sent_serial == INT_MAX) sent_serial = 0; else sent_serial++;

#define FlushBuffer(x)	((SBufLength(&x->localClient->sendQ) > 0) || \
				(IsServer(x) && OutputZIP(x) && zip_is_data_out(x->serv->zip_out)))

static int send_queued_buffers(HookData *hdata)
{
	aClient *cptr;
	dlink_node *node;

	DLINK_FOREACH_PREV_DATA(lserver_list.tail, node, cptr, aClient) {
		if (!IsBlocked(cptr) && FlushBuffer(cptr)) {
			engine_send_queued(cptr->localClient->fd, cptr);
		}
	}
	DLINK_FOREACH_PREV_DATA(lclient_list.tail, node, cptr, aClient) {
		if (!IsBlocked(cptr) && FlushBuffer(cptr)) {
			engine_send_queued(cptr->localClient->fd, cptr);
		}
	}
	DLINK_FOREACH_PREV_DATA(lunknown_list.tail, node, cptr, aClient) {
		if (!IsBlocked(cptr) && FlushBuffer(cptr)) {
			engine_send_queued(cptr->localClient->fd, cptr);
		}
	}

	return 0;
}

void flush_queued_buffers(int fd)
{
	aClient *cptr;
	dlink_node *node;

	DLINK_FOREACH_PREV_DATA(lserver_list.tail, node, cptr, aClient) {
		if ((fd == me.localClient->fd || fd == cptr->localClient->fd) && FlushBuffer(cptr)) {
			engine_send_queued(cptr->localClient->fd, cptr);
			if (fd != me.localClient->fd) {
				return;
			}
		}
	}
	DLINK_FOREACH_PREV_DATA(lclient_list.tail, node, cptr, aClient) {
		if ((fd == me.localClient->fd || fd == cptr->localClient->fd) && FlushBuffer(cptr)) {
			engine_send_queued(cptr->localClient->fd, cptr);
			if (fd != me.localClient->fd) {
				return;
			}
		}
	}
	DLINK_FOREACH_PREV_DATA(lunknown_list.tail, node, cptr, aClient) {
		if ((fd == me.localClient->fd || fd == cptr->localClient->fd) && FlushBuffer(cptr)) {
			engine_send_queued(cptr->localClient->fd, cptr);
			if (fd != me.localClient->fd) {
				return;
			}
		}
	}
}

void init_send()
{
	hook_add_event(h_post_netio, send_queued_buffers);
}

static int send_message(aClient *to, char *msg, int len, void *sbuf_ptr)
{
	static int SQinK;
	int flag = 0;

	ASSERT(to != NULL);
	ASSERT(to->localClient != NULL);

	if (to->from != NULL) {
		to = to->from;
	}

	if (sbuf_ptr == NULL || (IsServer(to) && (OutputZIP(to) || OutputRC4(to)))) {
		flag = 1;
	}

	if (flag) { /* FIXME -- remove these, see TODO */
		if (IsServer(to) || DoingDKEY(to)) {
			if (len > 510) {
				msg[511] = '\n';
				msg[512] = '\0';
				len = 512;
			}
			else {
				msg[len++] = '\n';
				msg[len] = '\0';
			}
		}
		else {
			if (len > 509) {
				msg[510] = '\r';
				msg[511] = '\n';
				msg[512] = '\0';
				len = 512;
			}
			else {
				msg[len++] = '\r';
				msg[len++] = '\n';
				msg[len] = '\0';
			}
		}
	}

	if (IsMe(to)) {
		sendto_realops_lev(DEBUG_LEV, "Trying to send to myself! [%s]", msg);
		return 0;
	}
	if (DeadSocket(to)) {
		return 0;
	}

	if (SBufLength(&to->localClient->sendQ) > SENDQLENGTH(to)) {
		if (IsServer(to)) {
			sendto_ops("Max SendQ limit exceeded for %s: %d > %d",
				get_client_name(to, HIDE_IP), SBufLength(&to->localClient->sendQ),
				SENDQLENGTH(to));
		}

		SetSendQEx(to);
		engine_write_error(to, "Max SendQ exceeded for %s, closing link", 0);
		return -1;
	}

	to->localClient->sendM++;
	me.localClient->sendM++;
	if (to->localClient->listener != NULL) {
		to->localClient->listener->sendM++;
	}

	if (IsServer(to)) {
		if (OutputZIP(to)) {
			int ldata = SendingBurst(to);

			msg = zip_output(to->serv->zip_out, msg, &len, 0, &ldata);
			if (len == -1) {
				sendto_realops("Zip output error for %s: (%d) %s", to->name, ldata, msg);
				engine_write_error(to, "Zip output error for %s", IRCERR_ZIP);
				return -1;
			}
			if (!len) {
				return 0;
			}
			flag = 1;
		}
#ifdef USE_OPENSSL
		if (OutputRC4(to)) {
			rc4_process_stream_to_buf(to->serv->rc4_out, msg, rc4buf, len);
			msg = rc4buf;
			flag = 1;
		}
#endif
	}

	if (sbuf_ptr == NULL || flag) {
		if (sbuf_put(&to->localClient->sendQ, msg, len) < 0) {
			engine_write_error(to, "Buffer allocation error for %s, closing link", IRCERR_BUFALLOC);
			return -1;
		}
	}
	else {
		if (sbuf_put_share(&to->localClient->sendQ, sbuf_ptr) < 0) {
			engine_write_error(to, "Buffer allocation error for %s, closing link", IRCERR_BUFALLOC);
			return -1;
		}
	}

	if (IsBlocked(to)) {
		return 0;
	}

#ifdef ALWAYS_SEND_DURING_SPLIT
	if (currently_processing_netsplit) {
		engine_send_queued(to->localClient->fd, to);
		return 0;
	}
#endif

	SQinK = (SBufLength(&to->localClient->sendQ) >> 10);
	if (IsServer(to)) {
		if (SQinK > (to->localClient->last_sendq + 8)) {
			engine_send_queued(to->localClient->fd, to);
		}
	}
	else {
		if (SQinK > (to->localClient->last_sendq + 2)) {
			engine_send_queued(to->localClient->fd, to);
		}
	}
	return 0;
}

static inline int prefix_buffer(int remote, aClient *from, char *prefix, char *buffer, char *pattern, va_list vl)
{
	char *p;
	int msglen, sidx = 1;

	*buffer = ':';
	if (!remote && IsPerson(from)) {
		int flag = 0;
		for (p = from->name; *p != '\0'; p++) {
			buffer[sidx++] = *p;
		}
		if (*from->username != '\0') {
			buffer[sidx++] = '!';
			for (p = from->username; *p != '\0'; p++) {
				buffer[sidx++] = *p;
			}
		}
		if (*from->host != '\0' && !MyConnect(from)) {
			buffer[sidx++] = '@';
			for (p = MaskedHost(from); *p != '\0'; p++) {
				buffer[sidx++] = *p;
			}
			flag = 1;
		}
		if (!flag && MyConnect(from) && (*from->host != '\0')) {
			buffer[sidx++] = '@';
			for (p = MaskedHost(from); *p != '\0'; p++) {
				buffer[sidx++] = *p;
			}
		}
	}
	else {
		for (p = prefix; *p != '\0'; p++) {
			buffer[sidx++] = *p;
		}
	}

	msglen = ircvsprintf(&buffer[sidx], pattern + 3, vl);
	msglen += sidx;

	return msglen;
}

static inline void sendto_dlink_array_preformatted(dlink_array *array, char *buf, int len)
{
	int i;
	void *sbuf = NULL;

	sbuf_begin_share(buf, len, &sbuf);
	for (i = 0; i < dlink_array_count(array); i++) {
		send_message((aClient *)array->nodes[i].data, buf, len, sbuf);
	}
	sbuf_end_share(&sbuf, 1);
}

static inline void vsendto_dlink_array(dlink_array *array, char *pattern, va_list vl)
{
	int i, len = ircvsprintf(sendbuf, pattern, vl);
	void *sbuf = NULL;

	sbuf_begin_share(sendbuf, len, &sbuf);
	for (i = 0; i < dlink_array_count(array); i++) {
		send_message((aClient *)array->nodes[i].data, sendbuf, len, sbuf);
	}
	sbuf_end_share(&sbuf, 1);
}

static void vsendto_one(aClient *to, char *pattern, va_list vl)
{
	int len = ircvsprintf(sendbuf, pattern, vl);

	if (to->from) {
		to = to->from;
	}
	if (IsMe(to)) {
		sendto_realops_lev(DEBUG_LEV, "Trying to send to myself! [%s]", sendbuf);
	}
	else {
		send_message(to, sendbuf, len, NULL);
	}
}

static void vsendto_prefix_one(aClient *to, aClient *from, char *pattern, va_list vl)
{
	static char sender[HOSTLEN + NICKLEN + USERLEN + 5], temp[1024];
	int flag = 0, sidx = 0;
	char *idx, *par;
	va_list vl2;

	VA_COPY(vl2, vl);
	par = va_arg(vl, char *);

	if (from != NULL) {
		if (!MyClient(from) && IsPerson(to) && (to->from == from->from)) {
			if (IsServer(from)) {
				ircvsprintf(temp, pattern, vl2);
				sendto_ops("Send message (%s) to %s[%s] dropped from %s (Fake direction)",
					temp, to->name, to->from->name, from->name);
				return;
			}

			sendto_ops("Ghosted: %s[%s@%s] from %s[%s@%s] (%s)", to->name, to->username,
				MaskedHost(to), from->name, from->username, MaskedHost(from), to->from->name);
			sendto_serv_butone(NULL, ":%s KILL %s :%s (%s[%s@%s] Ghosted %s)", me.name, to->name,
				to->username, MaskedHost(to), to->from->name);
			SetKilled(to);
			exit_client(NULL, to, &me, "Ghosted client");

			if (IsPerson(from)) {
				send_me_numeric(from, ERR_GHOSTEDCLIENT, to->name, to->username,
					MaskedHost(to), to->from);
			}
			return;
		}
		if (MyClient(to) && IsPerson(from) && !mycmp(par, from->name)) {
			for (idx = from->name; *idx != '\0'; idx++) {
				sender[sidx++] = *idx;
			}
			if (*from->username != '\0') {
				sender[sidx++] = '!';
				for (idx = from->username; *idx != '\0'; idx++) {
					sender[sidx++] = *idx;
				}
			}
			if (*from->host != '\0' && !MyConnect(from)) {
				sender[sidx++] = '@';
				for (idx = MaskedHost(from); *idx != '\0'; idx++) {
					sender[sidx++] = *idx;
				}
				flag = 1;
			}
			if (!flag && MyConnect(from) && *from->host != '\0') {
				sender[sidx++] = '@';
				for (idx = MaskedHost(from); *idx != '\0'; idx++) {
					sender[sidx++] = *idx;
				}
			}
			sender[sidx] = '\0';
			par = sender;
		}
	}

	temp[0] = ':';
	sidx = 1;
	for (idx = par; *idx != '\0'; idx++) {
		temp[sidx++] = *idx;
	}
	for (idx = (pattern + 3); *idx != '\0'; idx++) {
		temp[sidx++] = *idx;
	}
	temp[sidx] = '\0';
	vsendto_one(to, temp, vl);
}

static void vsendto_realops(char *pattern, va_list vl)
{
	aClient *cptr;
	char nbuf[1024];
	dlink_node *node;

	DLINK_FOREACH_DATA(oper_list.head, node, cptr, aClient) {
		ircsprintf(nbuf, ":%s NOTICE %s :*** Notice -- %s", me.name, cptr->name, pattern);
		vsendto_one(cptr, nbuf, vl);
	}
}

static inline int check_fake_direction(aClient *from, aClient *to)
{
	if (MyClient(from) || !IsPerson(to) || (to->from != from->from)) {
		return 0;
	}

	if (IsServer(from)) {
		sendto_ops("Message to %s[%s] dropped from %s (Fake direction)",
			to->name, to->from->name, from->name);
		return -1;
	}

	sendto_ops("Ghosted: %s[%s@%s] from %s[%s@%s] (%s)", to->name, to->username,
		MaskedHost(to), from->name, from->username, MaskedHost(from), to->from->name,
		to->from->name);
	sendto_serv_butone(NULL, ":%s KILL %s :%s (%s[%s@%s] Ghosted %s)", me.name, to->name,
		me.name, to->name, to->username, MaskedHost(to), to->from->name);
	SetKilled(to);
	exit_client(NULL, to, &me, "Ghosted client");

	if (IsPerson(from)) {
		send_me_numeric(from, ERR_GHOSTEDCLIENT, to->name, to->username, MaskedHost(to),
			to->from);
	}
	return -1;
}

/*
 * SEND_* FUNCTIONS -- IDENTITY/TOKEN1 framework integration
 *
 * - join_message_buffer
 * - join_prefix_message_buffer
 * Take a message of known length (prepared by ircsprintf), to create a buffer to send
 *
 * - join_message_buffer_copy
 * - join_prefix_message_buffer_copy
 * Take a message of unknown length to create a buffer to send
 *
 * - sendto_one_client (pre & postfix, "from X to ..." send to to)
 * - sendto_one_client_real ("from X to ..." send to realto)
 * - sendto_one_client_prefixed ("from!user@host X to ..." send to to)
 * - sendto_one_client_prefixed_copy ("from!user@host X to string" send to to)
 * - sendto_one_client_nopostfix ("from X ..." send to to)
 * - sendto_one_client_nopostfix_copy ("from X string" send to to)
 * - sendto_one_client_onserver ("from X user@server ..." send to to)
 * - sendto_one_client_onserver_copy ("from X user@server string" send to to)
 *
 * - sendto_client_numeric ("from XXX to [numeric format] ..." send to to)
 * - sendto_serv_msg_butone ("from X to ..." send to all servers but one)
 * - sendto_serv_capab_msg_butone ("from X to ..." send to all servers with capab but one)
 * - sendto_channel_msg_butone ("from!user@host X #chan ..." send to all channels butone)
 * - sendto_channel_msg_butone_copy ("from!user@host X #chan string" send to all channels butone)
 * - sendto_channel_msg_localremote_butone ("... X #channel ..." to all on channel locally butone/remotely)
 * - sendto_channel_msg_localremote_butone_copy ("... X #channel string" to all on channel local butone/remotely)
 * - sendto_chanflag_msg_butone (.. to all on chan with flags butone)
 * - sendto_chanflag_msg_butone_copy (string to all on chan with flags butone)
 *
 * WARNING:
 * All send_*_copy functions prefix the argument with a :, all others do not. The only
 * exception to this is *_nopostfix_copy!
 */

static inline char *join_message_buffer(char *from, char fromprefix, char *to, char toprefix,
  char *cmd, char msgprefix, char *msg, int *msglen)
{
	int plen = 0, len = *msglen;

	if (from != NULL) {
		sendservbuf[plen++] = fromprefix;
		while (*from != '\0') {
			sendservbuf[plen++] = *from++;
		}
		sendservbuf[plen++] = ' ';
	}

	while (*cmd != '\0') {
		sendservbuf[plen++] = *cmd++;
	}
	sendservbuf[plen++] = ' ';

	if (to != NULL) {
		if (toprefix != '\0') {
			sendservbuf[plen++] = toprefix;
		}
		while (*to != '\0') {
			sendservbuf[plen++] = *to++;
		}
		sendservbuf[plen++] = ' ';
	}

	if (msgprefix != 0) {
		sendservbuf[plen++] = msgprefix;
	}
	if (len > 0) {
		memcpy(&sendservbuf[plen], msg, len);
	}
	else {
		if (sendservbuf[plen - 1] == ' ') {
			plen--;
		}
		sendservbuf[plen] = '\0';
	}

	*msglen = (len + plen);
	return sendservbuf;
}

static inline char *join_message_buffer_copy(char *from, char fromprefix, char *to, char toprefix,
  char *cmd, char msgprefix, char *msg, int *msglen)
{
	int plen = 0;

	if (from != NULL) {
		sendservbuf[plen++] = fromprefix;
		while (*from != '\0') {
			sendservbuf[plen++] = *from++;
		}
		sendservbuf[plen++] = ' ';
	}

	while (*cmd != '\0') {
		sendservbuf[plen++] = *cmd++;
	}
	sendservbuf[plen++] = ' ';

	if (to != NULL) {
		if (toprefix != '\0') {
			sendservbuf[plen++] = toprefix;
		}
		while (*to != '\0') {
			sendservbuf[plen++] = *to++;
		}
		sendservbuf[plen++] = ' ';
	}

	if (msg != NULL && *msg != '\0') {
		if (sendservbuf[plen - 1] != ' ') {
			sendservbuf[plen++] = ' ';
		}
		if (msgprefix != 0) {
			sendservbuf[plen++] = msgprefix;
		}
		while (*msg != '\0') {
			sendservbuf[plen++] = *msg++;
		}
	}
	*msglen = plen;
	return sendservbuf;
}

static inline char *join_prefix_message_buffer(char *from, char *fromuser, char *fromhost,
  char fromprefix, char *to, char toprefix, char *cmd, char msgprefix, char *msg, int *msglen)
{
	int plen = 0, len = *msglen;

	if (from != NULL) {
		sendservbuf[plen++] = fromprefix;
		while (*from != '\0') {
			sendservbuf[plen++] = *from++;
		}
		if (fromuser != NULL) {
			sendservbuf[plen++] = '!';
			while (*fromuser != '\0') {
				sendservbuf[plen++] = *fromuser++;
			}
		}
		if (fromhost != NULL) {
			sendservbuf[plen++] = '@';
			while (*fromhost != '\0') {
				sendservbuf[plen++] = *fromhost++;
			}
		}
		sendservbuf[plen++] = ' ';
	}

	while (*cmd != '\0') {
		sendservbuf[plen++] = *cmd++;
	}
	sendservbuf[plen++] = ' ';

	if (to != NULL) {
		if (toprefix != '\0') {
			sendservbuf[plen++] = toprefix;
		}
		while (*to != '\0') {
			sendservbuf[plen++] = *to++;
		}
		sendservbuf[plen++] = ' ';
	}

	if (msgprefix != 0) {
		sendservbuf[plen++] = msgprefix;
	}
	if (len > 0) {
		memcpy(&sendservbuf[plen], msg, len);
	}
	else {
		if (sendservbuf[plen - 1] == ' ') {
			plen--;
		}
		sendservbuf[plen] = '\0';
	}
	*msglen = (len + plen);
	return sendservbuf;
}

static inline char *join_prefix_message_buffer_copy(char *from, char *fromuser, char *fromhost,
  char fromprefix, char *to, char toprefix, char *cmd, char msgprefix, char *msg, int *msglen)
{
	int plen = 0;

	if (from != NULL) {
		sendservbuf[plen++] = fromprefix;
		while (*from != '\0') {
			sendservbuf[plen++] = *from++;
		}
		if (fromuser != NULL) {
			sendservbuf[plen++] = '!';
			while (*fromuser != '\0') {
				sendservbuf[plen++] = *fromuser++;
			}
		}
		if (fromhost != NULL) {
			sendservbuf[plen++] = '@';
			while (*fromhost != '\0') {
				sendservbuf[plen++] = *fromhost++;
			}
		}
		sendservbuf[plen++] = ' ';
	}

	while (*cmd != '\0') {
		sendservbuf[plen++] = *cmd++;
	}
	sendservbuf[plen++] = ' ';

	if (to != NULL) {
		if (toprefix != '\0') {
			sendservbuf[plen++] = toprefix;
		}
		while (*to != '\0') {
			sendservbuf[plen++] = *to++;
		}
		sendservbuf[plen++] = ' ';
	}

	if (msg != NULL && *msg != '\0') {
		if (sendservbuf[plen - 1] != ' ') {
			sendservbuf[plen++] = ' ';
		}
		if (msgprefix != 0) {
			sendservbuf[plen++] = msgprefix;
		}
		while (*msg != '\0') {
			sendservbuf[plen++] = *msg++;
		}
	}
	*msglen = plen;
	return sendservbuf;
}

static inline void get_prefix(aClient *cptr, char **id_ptr, char *prefix_ptr, int pre)
{
	char *id = NULL, prefix = '\0';

	ASSERT(cptr != NULL);

	if (HasSUID(cptr)) {
		id = cptr->id.string;
		prefix = '!';
	}
	else {
		id = cptr->name;
		prefix = (pre) ? ':' : 0;
	}

	*id_ptr = id;
	*prefix_ptr = prefix;
}

void sendto_one_client(aClient *to, aClient *from, msg_ptr *msg, char *pattern, ...)
{
	aClient *realto = (to != NULL && to->from != NULL) ? to->from : to;
	va_list vl;
	int len = 0, tlen;
	char *text;

	if (*pattern != '\0') {
		va_start(vl, pattern);
		len = ircvsprintf(sendbuf, pattern, vl);
		va_end(vl);
	}

	tlen = len;
	if (IsServer(realto) && CapID(realto)) {
		char *msg_str = (msg->tok_str != NULL) ? msg->tok_str : msg->msg_str;
		char *from_id = NULL, *to_id = NULL, fromprefix = 0, toprefix = 0;

		if (from != NULL) {
			get_prefix(from, &from_id, &fromprefix, 1);
		}
		if (to != NULL) {
			get_prefix(to, &to_id, &toprefix, 0);
		}
		text = join_message_buffer(from_id, fromprefix, to_id, toprefix, msg_str, 0,
			sendbuf, &tlen);
	}
	else {
		text = join_message_buffer(from != NULL ? from->name : NULL, ':',
			to != NULL ? to->name : NULL, 0, msg->msg_str, 0, sendbuf, &tlen);
	}
	send_message(realto, text, tlen, NULL);
}

void sendto_one_client_real(aClient *realto, aClient *to, aClient *from, msg_ptr *msg, char *pattern, ...)
{
	va_list vl;
	int len = 0, tlen;
	char *text;

	realto = (realto->from != NULL) ? realto->from : realto;

	if (*pattern != '\0') {
		va_start(vl, pattern);
		len = ircvsprintf(sendbuf, pattern, vl);
		va_end(vl);
	}

	tlen = len;
	if (IsServer(realto) && CapID(realto)) {
		char *msg_str = (msg->tok_str != NULL) ? msg->tok_str : msg->msg_str;
		char *from_id = NULL, *to_id = NULL, fromprefix = 0, toprefix = 0;

		if (from != NULL) {
			get_prefix(from, &from_id, &fromprefix, 1);
		}
		if (to != NULL) {
			get_prefix(to, &to_id, &toprefix, 0);
		}
		text = join_message_buffer(from_id, fromprefix, to_id, toprefix, msg_str, 0,
			sendbuf, &tlen);
	}
	else {
		text = join_message_buffer(from != NULL ? from->name : NULL, ':',
			to != NULL ? to->name : NULL, 0, msg->msg_str, 0, sendbuf, &tlen);
	}
	send_message(realto, text, tlen, NULL);
}

void sendto_one_client_prefixed(aClient *to, aClient *from, msg_ptr *msg, char *pattern, ...)
{
	aClient *realto = (to != NULL && to->from != NULL) ? to->from : to;
	va_list vl;
	int len = 0, tlen;
	char *text;

	if (*pattern != '\0') {
		va_start(vl, pattern);
		len = ircvsprintf(sendbuf, pattern, vl);
		va_end(vl);
	}

	tlen = len;
	if (IsServer(realto) && CapID(realto)) {
		char *msg_str = (msg->tok_str != NULL) ? msg->tok_str : msg->msg_str;
		char *from_id = NULL, *to_id = NULL, fromprefix = 0, toprefix = 0;

		if (from != NULL) {
			get_prefix(from, &from_id, &fromprefix, 1);
		}
		if (to != NULL) {
			get_prefix(to, &to_id, &toprefix, 0);
		}
		text = join_message_buffer(from_id, fromprefix, to_id, toprefix, msg_str, 0,
			sendbuf, &tlen);
	}
	else if (IsServer(realto)) {
		text = join_message_buffer(from != NULL ? from->name : NULL, ':',
			to != NULL ? to->name : NULL, 0, msg->msg_str, 0, sendbuf, &tlen);
	}
	else {
		char *from_user = ((from->user != NULL) && (*from->username != '\0')) ? from->username : NULL;
		char *from_host = NULL;

		if (!MyConnect(from) && (from->user != NULL)) {
			from_host = (*from->host != '\0') ? MaskedHost(from) : NULL;
		}
		else if (MyConnect(from)) {
			from_host = (*from->host != '\0') ? MaskedHost(from) : from->localClient->sockhost;
		}

		text = join_prefix_message_buffer(from->name, from_user, from_host, ':',
			to != NULL ? to->name : NULL, 0, msg->msg_str, 0, sendbuf, &tlen);
	}
	send_message(realto, text, tlen, NULL);
}

void sendto_one_client_prefixed_copy(aClient *to, aClient *from, msg_ptr *msg, int doprefix, char *buf)
{
	aClient *realto = (to != NULL && to->from != NULL) ? to->from : to;
	int tlen;
	char *text;

	if (IsServer(realto) && CapID(realto)) {
		char *msg_str = (msg->tok_str != NULL) ? msg->tok_str : msg->msg_str;
		char *from_id = NULL, *to_id = NULL, fromprefix = 0, toprefix = 0;

		if (from != NULL) {
			get_prefix(from, &from_id, &fromprefix, 1);
		}
		if (to != NULL) {
			get_prefix(to, &to_id, &toprefix, 0);
		}
		text = join_message_buffer_copy(from_id, fromprefix, to_id, toprefix, msg_str, doprefix,
			buf, &tlen);
	}
	else if (IsServer(realto)) {
		text = join_message_buffer_copy(from != NULL ? from->name : NULL, ':',
			to != NULL ? to->name : NULL, 0, msg->msg_str, doprefix, buf, &tlen);
	}
	else {
		char *from_user = ((from->user != NULL) && (*from->username != '\0')) ? from->username : NULL;
		char *from_host = NULL;

		if (!MyConnect(from) && (from->user != NULL)) {
			from_host = (*from->host != '\0') ? MaskedHost(from) : NULL;
		}
		else if (MyConnect(from) && (from->user != NULL)) {
			from_host = (*from->host != '\0') ? MaskedHost(from) : from->localClient->sockhost;
		}

		text = join_prefix_message_buffer_copy(from->name, from_user, from_host, ':',
			to != NULL ? to->name : NULL, 0, msg->msg_str, doprefix, buf, &tlen);
	}
	send_message(realto, text, tlen, NULL);
}

void sendto_one_client_nopostfix(aClient *to, aClient *from, msg_ptr *msg, char *pattern, ...)
{
	aClient *realto = (to != NULL && to->from != NULL) ? to->from : to;
	va_list vl;
	int len = 0, tlen;
	char *text;

	if (*pattern != '\0') {
		va_start(vl, pattern);
		len = ircvsprintf(sendbuf, pattern, vl);
		va_end(vl);
	}

	tlen = len;
	if (IsServer(realto) && CapID(realto)) {
		char *msg_str = (msg->tok_str != NULL) ? msg->tok_str : msg->msg_str;
		char *from_id = NULL, fromprefix = 0;

		if (from != NULL) {
			get_prefix(from, &from_id, &fromprefix, 1);
		}
		text = join_message_buffer(from_id, fromprefix, NULL, 0, msg_str, 0, sendbuf, &tlen);
	}
	else {
		text = join_message_buffer(from != NULL ? from->name : NULL, ':', NULL, 0, msg->msg_str,
			0, sendbuf, &tlen);
	}
	send_message(realto, text, tlen, NULL);
}

void sendto_one_client_nopostfix_copy(aClient *to, aClient *from, msg_ptr *msg, int doprefix, char *buf)
{
	aClient *realto = (to != NULL && to->from != NULL) ? to->from : to;
	int tlen;
	char *text;

	if (IsServer(realto) && CapID(realto)) {
		char *msg_str = (msg->tok_str != NULL) ? msg->tok_str : msg->msg_str;
		char *from_id = NULL, fromprefix = 0;

		if (from != NULL) {
			get_prefix(from, &from_id, &fromprefix, 1);
		}
		text = join_message_buffer_copy(from_id, fromprefix, NULL, 0, msg_str, doprefix, buf,
			&tlen);
	}
	else {
		text = join_message_buffer_copy(from != NULL ? from->name : NULL, ':', NULL, 0,
			msg->msg_str, doprefix, buf, &tlen);
	}
	send_message(realto, text, tlen, NULL);
}

void sendto_one_client_onserver(aClient *to, aClient *from, msg_ptr *msg, char *pattern, ...)
{
	aClient *realto = (to->from != NULL) ? to->from : to;
	va_list vl;
	int len = 0, tlen;
	char *text;

	if (*pattern != '\0') {
		va_start(vl, pattern);
		len = ircvsprintf(sendbuf, pattern, vl);
		va_end(vl);
	}

	tlen = len;
	if (IsServer(realto) && CapID(realto)) {
		char *msg_str = (msg->tok_str != NULL) ? msg->tok_str : msg->msg_str;
		char *from_id = NULL, *to_id = NULL, fromprefix = 0, toprefix = 0;

		if (from != NULL) {
			get_prefix(from, &from_id, &fromprefix, 1);
		}
		if (to != NULL) {
			get_prefix(to, &to_id, &toprefix, 0);
		}
		text = join_message_buffer(from_id, fromprefix, to_id, toprefix, msg_str, 0,
			sendbuf, &tlen);
	}
	else {
		char nickserver[NICKLEN + HOSTLEN + 3];
		ircsprintf(nickserver, "%s@%s", to->name, to->uplink->name);
		text = join_message_buffer(from != NULL ? from->name : NULL, ':', nickserver, 0,
			msg->msg_str, 0, sendbuf, &tlen);
	}
	send_message(realto, text, tlen, NULL);
}

void sendto_one_client_onserver_copy(aClient *to, aClient *from, msg_ptr *msg, int doprefix, char *buf)
{
	aClient *realto = (to->from != NULL) ? to->from : to;
	int tlen;
	char *text;

	if (IsServer(realto) && CapID(realto)) {
		char *msg_str = (msg->tok_str != NULL) ? msg->tok_str : msg->msg_str;
		char *from_id = NULL, *to_id = NULL, fromprefix = 0, toprefix = 0;

		if (from != NULL) {
			get_prefix(from, &from_id, &fromprefix, 1);
		}
		if (to != NULL) {
			get_prefix(to, &to_id, &toprefix, 0);
		}
		text = join_message_buffer_copy(from_id, fromprefix, to_id, toprefix, msg_str,
			doprefix, buf, &tlen);
	}
	else {
		char nickserver[NICKLEN + HOSTLEN + 3];
		ircsprintf(nickserver, "%s@%s", to->name, to->uplink->name);
		text = join_message_buffer_copy(from != NULL ? from->name : NULL, ':', nickserver, 0,
			msg->msg_str, doprefix, buf, &tlen);
	}
	send_message(realto, text, tlen, NULL);
}

void sendto_client_numeric(aClient *to, aClient *from, char *format, int numeric, ...)
{
	aClient *realto = (to != NULL && to->from != NULL) ? to->from : to;
	va_list vl;
	int len = 0, tlen;
	char *pattern = BadPtr(format) ? get_numeric_format(&numeric) : format;
	char numerictext[16], *text;

	make_numeric_string(numerictext, numeric);

	if (*pattern != '\0') {
		va_start(vl, numeric);
		len = ircvsprintf(sendbuf, pattern, vl);
		va_end(vl);
	}

	tlen = len;
	if (IsServer(realto) && CapID(realto)) {
		char *from_id = NULL, *to_id = NULL, fromprefix = 0, toprefix = 0;

		if (from != NULL) {
			get_prefix(from, &from_id, &fromprefix, 1);
		}
		if (to != NULL) {
			get_prefix(to, &to_id, &toprefix, 0);
		}
		text = join_message_buffer(from_id, fromprefix, to_id, toprefix, numerictext,
			0, sendbuf, &tlen);
	}
	else {
		text = join_message_buffer(from != NULL ? from->name : NULL, ':',
			to != NULL ? (to->name != NULL ? to->name : "*") : NULL, 0,
			numerictext, 0, sendbuf, &tlen);
	}
	send_message(realto, text, tlen, NULL);
}

void sendto_serv_msg_butone(aClient *one, aClient *from, aClient *to, msg_ptr *msg, char *pattern, ...)
{
	aClient *cptr;
	va_list vl;
	int len = 0, tlen;
	dlink_array a_id, a_norm;
	char *text;
	dlink_node *node;

	dlink_array_init(&a_id);
	dlink_array_init(&a_norm);

	DLINK_FOREACH_DATA(lserver_list.head, node, cptr, aClient) {
		if (one != NULL && cptr == one->from) {
			continue;
		}
		if (CapID(cptr)) {
			dlink_array_add(&a_id, cptr);
		}
		else {
			dlink_array_add(&a_norm, cptr);
		}
	}
	if (!dlink_array_count(&a_id) && !dlink_array_count(&a_norm)) {
		return;
	}

	if (*pattern != '\0') {
		va_start(vl, pattern);
		len = ircvsprintf(sendbuf, pattern, vl);
		va_end(vl);
	}

	tlen = len;
	if (dlink_array_count(&a_id)) {
		char *msg_str = (msg->tok_str != NULL) ? msg->tok_str : msg->msg_str;
		char *from_id = NULL, *to_id = NULL;
		char toprefix = 0, fromprefix = 0;

		if (from != NULL) {
			get_prefix(from, &from_id, &fromprefix, 1);
		}
		if (to != NULL) {
			get_prefix(to, &to_id, &toprefix, 0);
		}
		text = join_message_buffer(from_id, fromprefix, to_id, toprefix, msg_str,
			0, sendbuf, &tlen);
		sendto_dlink_array_preformatted(&a_id, text, tlen);
	}
	if (dlink_array_count(&a_norm)) {
		text = join_message_buffer(from != NULL ? from->name : NULL, ':',
			to != NULL ? to->name : NULL, 0, msg->msg_str, 0, sendbuf, &tlen);
		sendto_dlink_array_preformatted(&a_norm, text, tlen);
	}
}

void sendto_serv_capab_msg_butone(aClient *one, aClient *from, aClient *to, int include, int exclude,
  msg_ptr *msg, char *pattern, ...)
{
	aClient *cptr;
	va_list vl;
	int len = 0, tlen;
	dlink_array a_id, a_norm;
	char *text;
	dlink_node *node;

	dlink_array_init(&a_id);
	dlink_array_init(&a_norm);

	DLINK_FOREACH_DATA(lserver_list.head, node, cptr, aClient) {
		if (one != NULL && cptr == one->from) {
			continue;
		}
		if ((include && !((cptr->localClient->capabs & include) == include)) ||
		  (exclude && (cptr->localClient->capabs & exclude))) {
			continue;
		}
		if (CapID(cptr)) {
			dlink_array_add(&a_id, cptr);
		}
		else {
			dlink_array_add(&a_norm, cptr);
		}
	}
	if (!dlink_array_count(&a_id) && !dlink_array_count(&a_norm)) {
		return;
	}

	if (*pattern != '\0') {
		va_start(vl, pattern);
		len = ircvsprintf(sendbuf, pattern, vl);
		va_end(vl);
	}

	tlen = len;
	if (dlink_array_count(&a_id)) {
		char *msg_str = (msg->tok_str != NULL) ? msg->tok_str : msg->msg_str;
		char *from_id = NULL, *to_id = NULL;
		char toprefix = 0, fromprefix = 0;

		if (from != NULL) {
			get_prefix(from, &from_id, &fromprefix, 1);
		}
		if (to != NULL) {
			get_prefix(to, &to_id, &toprefix, 0);
		}
		text = join_message_buffer(from_id, fromprefix, to_id, toprefix, msg_str,
			0, sendbuf, &tlen);
		sendto_dlink_array_preformatted(&a_id, text, tlen);
	}
	if (dlink_array_count(&a_norm)) {
		text = join_message_buffer(from != NULL ? from->name : NULL, ':',
			to != NULL ? to->name : NULL, 0, msg->msg_str, 0, sendbuf, &tlen);
		sendto_dlink_array_preformatted(&a_norm, text, tlen);
	}
}

void sendto_channel_msg_butone(aClient *one, aClient *from, aChannel *chptr, char *chname,
  msg_ptr *msg, char *pattern, ...)
{
	chanMember *cm;
	aClient *cptr;
	int len = 0, tlen, didlocal;
	va_list vl;
	char *localtext = NULL, *text;
	dlink_array a_id, a_norm;
	void *sbuf = NULL;

	dlink_array_init(&a_id);
	dlink_array_init(&a_norm);

	NEXT_SERIAL

	if (*pattern != '\0') {
		va_start(vl, pattern);
		len = ircvsprintf(sendbuf, pattern, vl);
		va_end(vl);
	}

	tlen = len;
	for (cm = chptr->members; cm != NULL; cm = cm->nextuser) {
		cptr = cm->cptr;
		if (cptr->from == one || HasMode(cptr, UMODE_DEAF)) {
			continue;
		}
		if (check_fake_direction(from, cptr)) {
			continue;
		}

		if (MyClient(cptr)) {
			if (localtext == NULL) {
				didlocal = len;
				localtext = join_prefix_message_buffer(from->name,
					*from->username != '\0' ? from->username : NULL,
					*from->host != '\0' ? MaskedHost(from) : NULL, ':',
					chname, 0, msg->msg_str, ':', sendbuf, &didlocal);
				sbuf_begin_share(localtext, didlocal, &sbuf);
			}
			send_message(cptr, localtext, didlocal, sbuf);
			cptr->serial = sent_serial;
		}
		else if (cptr->from->serial != sent_serial) {
			cptr->from->serial = sent_serial;
			if (CapID(cptr->from)) {
				dlink_array_add(&a_id, cptr->from);
			}
			else {
				dlink_array_add(&a_norm, cptr->from);
			}
		}
	}

	sbuf_end_share(&sbuf, 1);

	if (dlink_array_count(&a_id)) {
		char *msg_str = (msg->tok_str != NULL) ? msg->tok_str : msg->msg_str;
		char *from_id = NULL, fromprefix = 0;

		if (from != NULL) {
			get_prefix(from, &from_id, &fromprefix, 1);
		}
		text = join_message_buffer(from_id, fromprefix, chname, 0, msg_str, ':',
			sendbuf, &tlen);
		sendto_dlink_array_preformatted(&a_id, text, tlen);
	}
	if (dlink_array_count(&a_norm)) {
		text = join_message_buffer(from != NULL ? from->name : NULL, ':', chname,
			0, msg->msg_str, ':', sendbuf, &tlen);
		sendto_dlink_array_preformatted(&a_norm, text, tlen);
	}
}

void sendto_channel_msg_butone_copy(aClient *one, aClient *from, aChannel *chptr, char *chname,
  msg_ptr *msg, int doprefix, char *buf)
{
	chanMember *cm;
	aClient *cptr;
	int tlen, didlocal;
	char *localtext = NULL, *text;
	dlink_array a_id, a_norm;
	void *sbuf = NULL;

	dlink_array_init(&a_id);
	dlink_array_init(&a_norm);

	NEXT_SERIAL

	for (cm = chptr->members; cm != NULL; cm = cm->nextuser) {
		cptr = cm->cptr;
		if (cptr->from == one || HasMode(cptr, UMODE_DEAF)) {
			continue;
		}
		if (check_fake_direction(from, cptr)) {
			continue;
		}

		if (MyClient(cptr)) {
			if (localtext == NULL) {
				localtext = join_prefix_message_buffer_copy(from->name,
					*from->username != '\0' ? from->username : NULL,
					*from->host != '\0' ? MaskedHost(from) : NULL, ':',
					chname, 0, msg->msg_str, doprefix, buf, &didlocal);
				sbuf_begin_share(localtext, didlocal, &sbuf);
			}
			send_message(cptr, localtext, didlocal, sbuf);
			cptr->serial = sent_serial;
		}
		else if (cptr->from->serial != sent_serial) {
			cptr->from->serial = sent_serial;
			if (CapID(cptr->from)) {
				dlink_array_add(&a_id, cptr->from);
			}
			else {
				dlink_array_add(&a_norm, cptr->from);
			}
		}
	}

	sbuf_end_share(&sbuf, 1);

	if (dlink_array_count(&a_id)) {
		char *msg_str = (msg->tok_str != NULL) ? msg->tok_str : msg->msg_str;
		char *from_id = NULL, fromprefix = 0;

		if (from != NULL) {
			get_prefix(from, &from_id, &fromprefix, 1);
		}
		text = join_message_buffer_copy(from_id, fromprefix, chname, 0, msg_str,
			doprefix, buf, &tlen);
		sendto_dlink_array_preformatted(&a_id, text, tlen);
	}
	if (dlink_array_count(&a_norm)) {
		text = join_message_buffer_copy(from != NULL ? from->name : NULL, ':',
			chname, 0, msg->msg_str, doprefix, buf, &tlen);
		sendto_dlink_array_preformatted(&a_norm, text, tlen);
	}
}

void sendto_channel_msg_localremote_butone(int dolocal, aClient *one, aClient *from, aChannel *chptr,
  char *chname, msg_ptr *msg, char *pattern, ...)
{
	chanMember *cm;
	aClient *cptr;
	va_list vl;
	int didlocal, len = 0, tlen;
	char *localtext = NULL, *text;
	dlink_array a_id, a_norm;
	void *sbuf = NULL;

	dlink_array_init(&a_id);
	dlink_array_init(&a_norm);

	NEXT_SERIAL

	if (*pattern != '\0') {
		va_start(vl, pattern);
		len = ircvsprintf(sendbuf, pattern, vl);
		va_end(vl);
	}

	tlen = len;
	for (cm = chptr->members; cm != NULL; cm = cm->nextuser) {
		cptr = cm->cptr;
		if (cptr->from == one || HasMode(cptr, UMODE_DEAF)) {
			continue;
		}

		if (dolocal && MyClient(cptr)) {
			if (check_fake_direction(from, cptr)) {
				continue;
			}
			if (localtext == NULL) {
				didlocal = len;
				localtext = join_prefix_message_buffer(from->name,
					*from->username != '\0' ? from->username : NULL,
					*from->host != '\0' ? MaskedHost(from) : NULL, ':',
					chname, 0, msg->msg_str, 0, sendbuf, &didlocal);
				sbuf_begin_share(localtext, didlocal, &sbuf);
			}
			send_message(cptr, localtext, didlocal, sbuf);
			cptr->serial = sent_serial;
		}
		else if (!dolocal) {
			if (check_fake_direction(from, cptr)) {
				continue;
			}
			if (cptr->from->serial != sent_serial) {
				cptr->from->serial = sent_serial;
				if (CapID(cptr->from)) {
					dlink_array_add(&a_id, cptr->from);
				}
				else {
					dlink_array_add(&a_norm, cptr->from);
				}
			}
		}
	}

	sbuf_end_share(&sbuf, 1);

	if (dlink_array_count(&a_id)) {
		char *msg_str = (msg->tok_str != NULL) ? msg->tok_str : msg->msg_str;
		char *from_id = NULL, fromprefix = 0;

		if (from != NULL) {
			get_prefix(from, &from_id, &fromprefix, 1);
		}
		text = join_message_buffer(from_id, fromprefix, chname, 0, msg_str, 0,
			sendbuf, &tlen);
		sendto_dlink_array_preformatted(&a_id, text, tlen);
	}
	if (dlink_array_count(&a_norm)) {
		text = join_message_buffer(from != NULL ? from->name : NULL, ':',
			chname, 0, msg->msg_str, 0, sendbuf, &tlen);
		sendto_dlink_array_preformatted(&a_norm, text, tlen);
	}
}

void sendto_channel_msg_localremote_butone_copy(int dolocal, aClient *one, aClient *from, aChannel *chptr,
  char *chname, msg_ptr *msg, int doprefix, char *buf)
{
	chanMember *cm;
	aClient *cptr;
	int didlocal, tlen;
	char *localtext = NULL, *text;
	dlink_array a_id, a_norm;
	void *sbuf = NULL;

	dlink_array_init(&a_id);
	dlink_array_init(&a_norm);

	NEXT_SERIAL

	for (cm = chptr->members; cm != NULL; cm = cm->nextuser) {
		cptr = cm->cptr;
		if (cptr->from == one || HasMode(cptr, UMODE_DEAF)) {
			continue;
		}

		if (dolocal && MyClient(cptr)) {
			if (check_fake_direction(from, cptr)) {
				continue;
			}
			if (localtext == NULL) {
				localtext = join_prefix_message_buffer_copy(from->name,
					*from->username != '\0' ? from->username : NULL,
					*from->host != '\0' ? MaskedHost(from) : NULL, ':',
					chname, 0, msg->msg_str, doprefix, buf, &didlocal);
				sbuf_begin_share(localtext, didlocal, &sbuf);
			}
			send_message(cptr, localtext, didlocal, sbuf);
			cptr->serial = sent_serial;
		}
		else if (!dolocal) {
			if (check_fake_direction(from, cptr)) {
				continue;
			}
			if (cptr->from->serial != sent_serial) {
				cptr->from->serial = sent_serial;
				if (CapID(cptr->from)) {
					dlink_array_add(&a_id, cptr->from);
				}
				else {
					dlink_array_add(&a_norm, cptr->from);
				}
			}
		}
	}

	sbuf_end_share(&sbuf, 1);

	if (dlink_array_count(&a_id)) {
		char *msg_str = (msg->tok_str != NULL) ? msg->tok_str : msg->msg_str;
		char *from_id = NULL, fromprefix = 0;

		if (from != NULL) {
			get_prefix(from, &from_id, &fromprefix, 1);
		}
		text = join_message_buffer_copy(from_id, fromprefix, chname, 0, msg_str,
			doprefix, buf, &tlen);
		sendto_dlink_array_preformatted(&a_id, text, tlen);
	}
	if (dlink_array_count(&a_norm)) {
		text = join_message_buffer_copy(from != NULL ? from->name : NULL, ':',
			chname, 0, msg->msg_str, doprefix, buf, &tlen);
		sendto_dlink_array_preformatted(&a_norm, text, tlen);
	}
}

void sendto_chanflag_msg_butone(aClient *one, aClient *from, aChannel *chptr, char *chname, int flags,
  msg_ptr *msg, char *pattern, ...)
{
	chanMember *cm;
	aClient *cptr;
	int len = 0, tlen, didlocal;
	va_list vl;
	char *localtext = NULL, *text;
	dlink_array a_id, a_norm;
	void *sbuf = NULL;

	dlink_array_init(&a_id);
	dlink_array_init(&a_norm);

	NEXT_SERIAL

	if (*pattern != '\0') {
		va_start(vl, pattern);
		len = ircvsprintf(sendbuf, pattern, vl);
		va_end(vl);
	}

	tlen = len;
	for (cm = chptr->members; cm != NULL; cm = cm->nextuser) {
		cptr = cm->cptr;
		if (!(cm->flags & flags) || (cptr->from == one)) {
			continue;
		}
		if (check_fake_direction(from, cptr)) {
			continue;
		}

		if (MyClient(cptr)) {
			if (localtext == NULL) {
				didlocal = len;
				localtext = join_prefix_message_buffer(from->name,
					*from->username != '\0' ? from->username : NULL,
					*from->host != '\0' ? MaskedHost(from) : NULL, ':',
					chname, 0, msg->msg_str, ':', sendbuf, &didlocal);
				sbuf_begin_share(localtext, didlocal, &sbuf);
			}
			send_message(cptr, localtext, didlocal, sbuf);
			cptr->serial = sent_serial;
		}
		else if (cptr->from->serial != sent_serial) {
			cptr->from->serial = sent_serial;
			if (CapID(cptr->from)) {
				dlink_array_add(&a_id, cptr->from);
			}
			else {
				dlink_array_add(&a_norm, cptr->from);
			}
		}
	}

	sbuf_end_share(&sbuf, 1);

	if (dlink_array_count(&a_id)) {
		char *msg_str = (msg->tok_str != NULL) ? msg->tok_str : msg->msg_str;
		char *from_id = NULL, fromprefix = 0;

		if (from != NULL) {
			get_prefix(from, &from_id, &fromprefix, 1);
		}
		text = join_message_buffer(from_id, fromprefix, chname, 0, msg_str, ':',
			sendbuf, &tlen);
		sendto_dlink_array_preformatted(&a_id, text, tlen);
	}
	if (dlink_array_count(&a_norm)) {
		text = join_message_buffer(from != NULL ? from->name : NULL, ':',
			chname, 0, msg->msg_str, ':', sendbuf, &tlen);
		sendto_dlink_array_preformatted(&a_norm, text, tlen);
	}
}

void sendto_chanflag_msg_butone_copy(aClient *one, aClient *from, aChannel *chptr, char *chname,
  int flags, msg_ptr *msg, int doprefix, char *buf)
{
	chanMember *cm;
	aClient *cptr;
	int tlen, didlocal;
	char *localtext = NULL, *text;
	dlink_array a_id, a_norm;
	void *sbuf = NULL;

	dlink_array_init(&a_id);
	dlink_array_init(&a_norm);

	NEXT_SERIAL

	for (cm = chptr->members; cm != NULL; cm = cm->nextuser) {
		cptr = cm->cptr;
		if (!(cm->flags & flags) || (cptr->from == one)) {
			continue;
		}
		if (check_fake_direction(from, cptr)) {
			continue;
		}

		if (MyClient(cptr)) {
			if (localtext == NULL) {
				localtext = join_prefix_message_buffer_copy(from->name,
					*from->username != '\0' ? from->username : NULL,
					*from->host != '\0' ? MaskedHost(from) : NULL, ':',
					chname, 0, msg->msg_str, doprefix, buf, &didlocal);
				sbuf_begin_share(localtext, didlocal, &sbuf);
			}
			send_message(cptr, localtext, didlocal, sbuf);
			cptr->serial = sent_serial;
		}
		else if (cptr->from->serial != sent_serial) {
			cptr->from->serial = sent_serial;
			if (CapID(cptr->from)) {
				dlink_array_add(&a_id, cptr->from);
			}
			else {
				dlink_array_add(&a_norm, cptr->from);
			}
		}
	}

	sbuf_end_share(&sbuf, 1);

	if (dlink_array_count(&a_id)) {
		char *msg_str = (msg->tok_str != NULL) ? msg->tok_str : msg->msg_str;
		char *from_id = NULL, fromprefix = 0;

		if (from != NULL) {
			get_prefix(from, &from_id, &fromprefix, 1);
		}
		text = join_message_buffer_copy(from_id, fromprefix, chname, 0, msg_str, doprefix,
			buf, &tlen);
		sendto_dlink_array_preformatted(&a_id, text, tlen);
	}
	if (dlink_array_count(&a_norm)) {
		text = join_message_buffer_copy(from != NULL ? from->name : NULL, ':',
			chname, 0, msg->msg_str, doprefix, buf, &tlen);
		sendto_dlink_array_preformatted(&a_norm, text, tlen);
	}
}

static int match_it(aClient *one, char *mask, int what)
{
	if (what == MATCH_HOST) {
		return (!match(mask, one->host));
	}
	return (!match(mask, one->user->server));
}

void sendto_match_butone(aClient *one, aClient *from, char *mask, int what, char *pattern, ...)
{
	aClient *cptr, *acptr;
	int cansendlocal, cansendglobal;
	va_list vl;
	dlink_node *node, *next = NULL;

	if (MyConnect(from)) {
		cansendlocal = OPHasFlag(from, OFLAG_LNOTICE) ? 1 : 0;
		cansendglobal = OPHasFlag(from, OFLAG_GNOTICE) ? 1 : 0;
	}
	else {
		cansendlocal = cansendglobal = 1;
	}

	if (!cansendlocal && !cansendglobal) {
		return;
	}

	va_start(vl, pattern);

	if (cansendlocal) {
		DLINK_FOREACH_SAFE_DATA(lclient_list.head, node, next, cptr, aClient) {
			if (cptr == one) {
				continue;
			}
			if (IsRegisteredUser(cptr) && match_it(cptr, mask, what)) {
				vsendto_prefix_one(cptr, from, pattern, vl);
			}
		}
	}

	if (cansendglobal) {
		DLINK_FOREACH_SAFE_DATA(lserver_list.head, node, next, cptr, aClient) {
			if (cptr == one) {
				continue;
			}
			for (acptr = client; acptr != NULL; acptr = acptr->next) {
				if (IsRegisteredUser(acptr) && match_it(acptr, mask, what) && (acptr->from == cptr)) {
					break;
				}
			}
			if (acptr != NULL) {
				vsendto_prefix_one(cptr, from, pattern, vl);
			}
		}
	}

	va_end(vl);
}

void sendto_serv_butone(aClient *one, char *pattern, ...)
{
	aClient *cptr;
	va_list vl;
	dlink_node *node;
	dlink_array a_send;

	dlink_array_init(&a_send);

	va_start(vl, pattern);
	DLINK_FOREACH_DATA(lserver_list.head, node, cptr, aClient) {
		if (one && cptr != one->from) {
			dlink_array_add(&a_send, cptr);
		}
	}
	if (dlink_array_count(&a_send)) {
		vsendto_dlink_array(&a_send, pattern, vl);
	}
	va_end(vl);
}

void sendto_common_channels(aClient *from, char *pattern, ...)
{
	chanMember *chans, *users;
	aClient *cptr;
	va_list vl;
	char *pfix;
	int msglen = 0;
	void *sbuf = NULL;

	va_start(vl, pattern);
	pfix = va_arg(vl, char *);

	NEXT_SERIAL
	from->serial = sent_serial;

	if (from->user != NULL) {
		for (chans = from->user->channel; chans != NULL; chans = chans->nextchan) {
			for (users = chans->chptr->members; users != NULL; users = users->nextuser) {
				cptr = users->cptr;

				if (!MyConnect(cptr) || (cptr->serial == sent_serial)) {
					continue;
				}

				if (!msglen) {
					msglen = prefix_buffer(0, from, pfix, sendbuf, pattern, vl);
					sbuf_begin_share(sendbuf, msglen, &sbuf);
				}
				if (check_fake_direction(from, cptr)) {
					continue;
				}
				send_message(cptr, sendbuf, msglen, sbuf);
				cptr->serial = sent_serial;
			}
		}
	}
	if (MyConnect(from)) {
		if (!msglen) {
			msglen = prefix_buffer(0, from, pfix, sendbuf, pattern, vl);
		}
		send_message(from, sendbuf, msglen, sbuf);
		from->serial = sent_serial;
	}
	sbuf_end_share(&sbuf, 1);
	va_end(vl);
}

void sendto_wallops(aClient *from, char *pattern, ...)
{
	int len;
	aClient *cptr;
	va_list vl;
	char *prefix = NULL;
	dlink_node *node;

	va_start(vl, pattern);
	prefix = va_arg(vl, char *);
	len = prefix_buffer(0, from, prefix, sendbuf, pattern, vl);
	va_end(vl);

	DLINK_FOREACH_DATA(oper_list.head, node, cptr, aClient) {
		if (HasMode(cptr, UMODE_WALLOPS)) {
			send_message(cptr, sendbuf, len, NULL);
		}
	}
}

void sendto_gnotice(char *pattern, ...)
{
	aClient *cptr;
	int len = 0;
	char nbuf[1024];
	va_list vl;
	dlink_node *node;

	va_start(vl, pattern);
	DLINK_FOREACH_DATA(oper_list.head, node, cptr, aClient) {
		if (HasMode(cptr, UMODE_RSTAFF)) {
			len = ircsprintf(nbuf, ":%s NOTICE %s :*** Routing -- ", me.name, cptr->name);
			strncat(nbuf, pattern, sizeof(nbuf) - len);
			vsendto_one(cptr, nbuf, vl);
		}
	}
	va_end(vl);
}

void sendto_one(aClient *to, char *pattern, ...)
{
	va_list vl;
	int len;

	va_start(vl, pattern);
	len = ircvsprintf(sendbuf, pattern, vl);
	va_end(vl);

	if (to->from) {
		to = to->from;
	}
	if (IsMe(to)) {
		sendto_realops_lev(DEBUG_LEV, "Trying to send to myself! [%s]", sendbuf);
	}
	else {
		send_message(to, sendbuf, len, NULL);
	}
}

void sendto_prefix_one(aClient *to, aClient *from, char *pattern, ...)
{
	static char sender[HOSTLEN + NICKLEN + USERLEN + 5], temp[1024];
	char *idx, *par;
	int flag = 0, sidx = 0;
	va_list vl, vl2;

	va_start(vl, pattern);
	VA_COPY(vl2, vl);
	par = va_arg(vl, char *);

	if (from != NULL) {
		if (!MyClient(from) && IsPerson(to) && (to->from == from->from)) {
			if (IsServer(from)) {
				ircvsprintf(temp, pattern, vl2);
				sendto_ops("Send message (%s) to %s[%s] dropped from %s (Fake direction)",
					temp, to->name, to->from->name, from->name);
				va_end(vl);
				return;
			}

			sendto_ops("Ghosted: %s[%s@%s] from %s[%s@%s] (%s)", to->name, to->username,
				MaskedHost(to), from->name, from->username, MaskedHost(from), to->from->name);
			sendto_serv_butone(NULL, ":%s KILL %s :%s (%s[%s@%s] Ghosted %s)", me.name,
				to->name, me.name, to->name, to->username, MaskedHost(to), to->from->name);
			SetKilled(to);
			exit_client(NULL, to, &me, "Ghosted client");

			if (IsPerson(from)) {
				send_me_numeric(from, ERR_GHOSTEDCLIENT, to->name, to->username,
					MaskedHost(to), to->from);
			}
			va_end(vl);
			return;
		}
		if (MyClient(to) && IsPerson(from) && !mycmp(par, from->name)) {
			for (idx = from->name; *idx != '\0'; idx++) {
				sender[sidx++] = *idx;
			}
			if (*from->username != '\0') {
				sender[sidx++] = '!';
				for (idx = from->username; *idx != '\0'; idx++) {
					sender[sidx++] = *idx;
				}
			}
			if (*from->host != '\0' && !MyConnect(from)) {
				sender[sidx++] = '@';
				for (idx = MaskedHost(from); *idx != '\0'; idx++) {
					sender[sidx++] = *idx;
				}
				flag = 1;
			}
			if (!flag && MyConnect(from) && *from->host != '\0') {
				sender[sidx++] = '@';
				for (idx = MaskedHost(from); *idx != '\0'; idx++) {
					sender[sidx++] = *idx;
				}
			}
			sender[sidx] = '\0';
			par = sender;
		}
	}

	temp[0] = ':';
	sidx = 1;
	for (idx = par; *idx != '\0'; idx++) {
		temp[sidx++] = *idx;
	}
	for (idx = (pattern + 3); *idx != '\0'; idx++) {
		temp[sidx++] = *idx;
	}
	temp[sidx] = '\0';
	vsendto_one(to, temp, vl);
	va_end(vl);
}

void sendto_ops(char *pattern, ...)
{
	va_list vl;
	va_start(vl, pattern);
	vsendto_realops(pattern, vl);
	va_end(vl);
}

void sendto_realops(char *pattern, ...)
{
	aClient *cptr;
	char nbuf[1024];
	va_list vl;
	dlink_node *node;

	va_start(vl, pattern);
	DLINK_FOREACH_DATA(oper_list.head, node, cptr, aClient) {
		if (HasMode(cptr, UMODE_OPER)) {
			ircsprintf(nbuf, ":%s NOTICE %s :*** Notice -- %s", me.name, cptr->name, pattern);
			vsendto_one(cptr, nbuf, vl);
		}
	}
	va_end(vl);
}

void sendto_realops_lev(int lev, char *pattern, ...)
{
	aClient *cptr;
	int len = 0;
	char nbuf[1024];
	va_list vl;
	dlink_node *node;
	char *notice;

	switch (lev) {
		case CCONN_LEV:
		case GCCONN_LEV:
			notice = "Client";
			break;
		case REJ_LEV:
			notice = "Reject";
			break;
		case SPY_LEV:
			notice = "Spy";
			break;
		case DCCSEND_LEV:
			notice = "DCC Allow";
			break;
		case FLOOD_LEV:
			notice = "Flood";
			break;
		case SPAM_LEV:
			notice = "Spam";
			break;
		case DEBUG_LEV:
			notice = "Debug";
			break;
		default:
			notice = "Notice";
	}

	va_start(vl, pattern);
	DLINK_FOREACH_DATA(oper_list.head, node, cptr, aClient) {
		switch (lev) {
			case CCONN_LEV:
				if (!HasMode(cptr, UMODE_LCLICONN) || !HasMode(cptr, UMODE_OPER)) {
					continue;
				}
				break;
			case GCCONN_LEV:
				if (!HasMode(cptr, UMODE_GCLICONN) || !HasMode(cptr, UMODE_OPER)) {
					continue;
				}
				break;
			case REJ_LEV:
				if (!HasMode(cptr, UMODE_REJNOTICES) || !HasMode(cptr, UMODE_OPER)) {
					continue;
				}
				break;
			case SKILL_LEV:
				if (!HasMode(cptr, UMODE_KNOTICE)) {
					continue;
				}
				break;
			case SPY_LEV:
				if (!HasMode(cptr, UMODE_SPYING) || !HasMode(cptr, UMODE_OPER)) {
					continue;
				}
				break;
			case DCCSEND_LEV:
				if (!HasMode(cptr, UMODE_DCCNOTICE) || !HasMode(cptr, UMODE_OPER)) {
					continue;
				}
				break;
			case FLOOD_LEV:
				if (!HasMode(cptr, UMODE_FNOTICE) || !HasMode(cptr, UMODE_OPER)) {
					continue;
				}
				break;
			case SPAM_LEV:
				if (!HasMode(cptr, UMODE_SPAMBOT) || !HasMode(cptr, UMODE_OPER)) {
					continue;
				}
				break;
			case DEBUG_LEV:
				if (!HasMode(cptr, UMODE_DNOTICE) || !HasMode(cptr, UMODE_OPER)) {
					continue;
				}
				break;
		}
		len = ircsprintf(nbuf, ":%s NOTICE %s :*** %s -- ", me.name, cptr->name, notice);
		strncat(nbuf, pattern, sizeof(nbuf) - len);
		vsendto_one(cptr, nbuf, vl);
	}
	va_end(vl);
}

void send_nick_to_serv_butone(aClient *one, aClient *sptr, char *ubuf)
{
	/* NICK: Non-ID/TOKEN */
	sendto_serv_capab_msg_butone(one, NULL, NULL, 0, CAP_SN2|CAP_ID, &CMD_NICK,
		"%s %d %ld %s %s %s %s %lu %lu :%s", sptr->name, sptr->hopcount + 1,
		sptr->tsinfo, ubuf, sptr->username, sptr->host, sptr->user->server,
		sptr->user->servicestamp, htonl(sptr->ip.s_addr), sptr->info);

	/* SNICK: Non-ID/TOKEN */
	sendto_serv_capab_msg_butone(one, NULL, NULL, CAP_SN2, CAP_ID, &CMD_SNICK,
		"%s %ld %d %s %s %lu %s %s %lu %s :%s", sptr->name, sptr->tsinfo,
		sptr->hopcount + 1, sptr->username, sptr->host, htonl(sptr->ip.s_addr),
		HasMode(sptr, UMODE_MASKED) ? sptr->user->maskedhost : "*",
		sptr->user->server, sptr->user->servicestamp, ubuf, sptr->info);

	if (HasSUID(sptr)) {
		/* NICK: ID/TOKEN, HasID */
		sendto_serv_capab_msg_butone(one, NULL, NULL, CAP_ID, CAP_SN2, &CMD_NICK,
			"%s %d %B %s %s %s !%s %B %B :%s", sptr->name, sptr->hopcount + 1,
			sptr->tsinfo, ubuf, sptr->username, sptr->host, sptr->id.string,
			sptr->user->servicestamp, sptr->ip.s_addr, sptr->info);

		/* SNICK: ID/TOKEN, HasID */
		sendto_serv_capab_msg_butone(one, NULL, NULL, CAP_ID|CAP_SN2, 0, &CMD_SNICK,
			"%s %B %d %s %s %B %s !%s %B %s :%s", sptr->name, sptr->tsinfo,
			sptr->hopcount + 1, sptr->username, sptr->host,
			sptr->ip.s_addr, HasMode(sptr, UMODE_MASKED) ? sptr->user->maskedhost : "*",
			sptr->id.string, sptr->user->servicestamp, ubuf, sptr->info);
	}
	else {
		/* NICK: ID/TOKEN, NoID */
		sendto_serv_capab_msg_butone(one, NULL, NULL, CAP_ID, CAP_SN2, &CMD_NICK,
			"%s %d %B %s %s %s %s %B %B :%s", sptr->name, sptr->hopcount + 1,
			sptr->tsinfo, ubuf, sptr->username, sptr->host, sptr->user->server,
			sptr->user->servicestamp, sptr->ip.s_addr, sptr->info);

		/* SNICK: ID/TOKEN, NoID */
		sendto_serv_capab_msg_butone(one, NULL, NULL, CAP_ID|CAP_SN2, 0, &CMD_SNICK,
			"%s %B %d %s %s %B %s %s %B %s :%s", sptr->name, sptr->tsinfo,
			sptr->hopcount + 1, sptr->username, sptr->host, sptr->ip.s_addr,
			HasMode(sptr, UMODE_MASKED) ? sptr->user->maskedhost : "*",
			sptr->user->server, sptr->user->servicestamp, ubuf, sptr->info);
	}
}

void send_quit_to_common_channels(aClient *from, char *reason)
{
	chanMember *chans, *users;
	aClient *cptr;
	int msglen;
	void *sbuf = NULL;

	msglen = ircsprintf(sendbuf, ":%s!%s@%s QUIT :%s", from->name, from->username, MaskedHost(from), reason);
	sbuf_begin_share(sendbuf, msglen, &sbuf);

	NEXT_SERIAL
	from->serial = sent_serial;

	for (chans = from->user->channel; chans != NULL; chans = chans->nextchan) {
		if (can_send(from, chans->chptr, reason)) {
			continue;
		}
		for (users = chans->chptr->members; users != NULL; users = users->nextuser) {
			cptr = users->cptr;

			if (!MyConnect(cptr) || (cptr->serial == sent_serial)) {
				continue;
			}
			if (check_fake_direction(from, cptr)) {
				continue;
			}

			send_message(cptr, sendbuf, msglen, sbuf);
			cptr->serial = sent_serial;
		}
	}

	sbuf_end_share(&sbuf, 1);
}

void send_part_to_common_channels(aClient *from, char *reason)
{
	chanMember *chans, *users;
	aClient *cptr;
	int msglen = 0;
	void *sbuf = NULL;

	for (chans = from->user->channel; chans != NULL; chans = chans->nextchan) {
		if (!can_send(from, chans->chptr, reason)) {
			continue;
		}

		msglen = ircsprintf(sendbuf, ":%s!%s@%s PART %s", from->name, from->username,
			MaskedHost(from), chans->chptr->chname);
		sbuf_begin_share(sendbuf, msglen, &sbuf);

		NEXT_SERIAL
		from->serial = sent_serial;

		for (users = chans->chptr->members; users != NULL; users = users->nextuser) {
			cptr = users->cptr;

			if (!MyConnect(cptr) || (cptr->serial == sent_serial)) {
				continue;
			}
			if (check_fake_direction(from, cptr)) {
				continue;
			}

			send_message(cptr, sendbuf, msglen, sbuf);
			cptr->serial = sent_serial;
		}

		sbuf_end_share(&sbuf, 1);
	}
}

void send_chatops(char *pattern, ...)
{
	aClient *cptr;
	char nbuf[1024];
	va_list vl;
	dlink_node *node;

	va_start(vl, pattern);
	DLINK_FOREACH_DATA(oper_list.head, node, cptr, aClient) {
		if (HasMode(cptr, UMODE_CHATOPS)) {
			ircsprintf(nbuf, ":%s NOTICE %s :*** ChatOps -- %s", me.name, cptr->name, pattern);
			vsendto_one(cptr, nbuf, vl);
		}
	}
	va_end(vl);
}

void send_globops(char *pattern, ...)
{
	aClient *cptr;
	char nbuf[1024];
	va_list vl;
	dlink_node *node;

	va_start(vl, pattern);
	DLINK_FOREACH_DATA(oper_list.head, node, cptr, aClient) {
		if (OPHasFlag(cptr, OFLAG_GLOBOPS)) {
			ircsprintf(nbuf, ":%s NOTICE %s :*** Global -- %s", me.name, cptr->name, pattern);
			vsendto_one(cptr, nbuf, vl);
		}
	}
	va_end(vl);
}

void send_locops(char *pattern, ...)
{
	aClient *cptr;
	int len = 0;
	char nbuf[1024];
	va_list vl;
	dlink_node *node;

	va_start(vl, pattern);
	DLINK_FOREACH_DATA(oper_list.head, node, cptr, aClient) {
		if (OPHasFlag(cptr, OFLAG_LOCOPS)) {
			len = ircsprintf(nbuf, ":%s NOTICE %s :*** LocOps -- ", me.name, cptr->name);
			strncat(nbuf, pattern, sizeof(nbuf) - len);
			vsendto_one(cptr, nbuf, vl);
		}
	}
	va_end(vl);
}

void ts_warn(char *pattern, ...)
{
	static time_t last = 0;
	static int warnings = 0;
	time_t now;
	va_list vl;

	va_start(vl, pattern);

	now = timeofday;
	if (now - last < 5) {
		if (++warnings > 5) {
			return;
		}
	}
	else {
		last = now;
		warnings = 0;
	}
	vsendto_realops(pattern, vl);
	va_end(vl);
}
